package app;

import java.io.Serializable;
import sassy.monitoring_steering.NameValuePair;
import sassy.*;
import java.util.*;
import java.util.Random;

/**This class is the java-version of the Phold.c from GTW. It is used for testing the sim kernel performance. 
  *The original Vaucher Duval hold mode was introduced solely to approximate a
  *simulation application interaction with its event calendar.
  *
  *<p> An application is assumed to operate with a fixed number of initial events in the calendar. 
  *In operation, it deletes the earliest event from the calendar, synchronizes the simulation clock with
  *this event scheduled time of occurrence, processes the event, and then inserts a
  *newly scheduled event into the calendar. 
  *
  *<p>begin
  * <ul>
  *   <li>initialize event calendar with msgPop events
  *   <li>done =false; hold_count=0;
  *   <li>begin titak_time measure
  *   <li>while (not done) do {
  *     <ul>
  *        <li>Delete highest priority event from calendar
  *        <li>Synchronize clock with this event's time of occurence
  *        <li>Define a scheduled time to be clock time plus a random variate from a 
  *            given scheduling distributioin
  *        <li>Encode scheduled time in new_event;
  *        <li>Insert new_event in calendar;
  *        <li>hold_count:=hold_count + 1
  *        <li>if(hold_count=n) then done:=true}
  *     </ul >
  *   <li>End total_time measure
  *   <li>hold_time=hold_time/n;
  *   <li>end hold algorithm
  * </ul>     
  *@author Yin Xiong
  *@version 2.0 (03/2007)
  */
  

public class Phold extends sassy.LogicalProcess implements Serializable
{

	/**the flag for displaying debugging info*/
	private static final boolean  STRACE = true;

	/**counter for cache hit; for testing purpose only*/
	int cacheHits=0;

	/**class member for keeping track of how many messages have been sent.*/
	int holdCount;

	/**message content; updated when receive a new message.*/
	public int count;

	/**initial message population;default to be 10 */	
	int msgPop=10;
	
        /** total of messages to be sent; default to be 50*/
	int totalSend=50;

	/**the number of LPs*/
	int numLPs=0;

	/**the time increment for messages sent by this LP; the message will be sent at currentTime+timeIncrement*/
	int timeIncrement;

	/**random number generator*/
	Random random;

	/**PEid and LPid of this LP, for debugging only*/
	String alias;

	/**the flag indicating whether a "done" message has already been sent*/
	boolean doneFlagSent=false;

	/**the vector to store endOfSim flags from LPs*/
	List<String> flagVector=new ArrayList<String>();

	final double PREFER_NEIGHBORS = 0.0;

	/**Constructs a new Phold instance.*/
	public Phold()
	{
   		super();
   		count=0;  
   		timeIncrement=0;
   		random=new Random();
	}


/**Initiates the LP and generates the initial message. It's the counterpart of "IProc"
  *in GTW.
  */
	@Override
public void appInitializeLP()
{
	alias=getPID()+"-"+getLID();
	//System.out.println(alias + " initializing...\n");

	int destLP = 0;

	//send initial messages
  	count=0;

  	for(int i=getLID();i<msgPop;i=i+numLPs){

		if(random.nextDouble() < PREFER_NEIGHBORS) {
			if(random.nextDouble() < 0.5) {
				destLP = getLID() + 1;
			}
			else {
				destLP = getLID() + numLPs - 1;
			}
			destLP %= numLPs;
		}
		else {
			destLP=random.nextInt(numLPs);
		}
	
	//destLP=getLID();

	//get timestamp
	double u=random.nextDouble();
	double lambda = 1.0;
//	if(destLP != getLID())
//		lambda = 1.0 / Math.abs(destLP - getLID());
	double ts=nextExponential(u, lambda);
	timeIncrement=(int)(ts*10);

	//content is the number of holds
	String newcontent=String.valueOf(count);
	if(STRACE)
		System.out.println(alias + " initial send " + newcontent + " to " + destLP + " at " + getCurrentTime() 
					 + " plus " + timeIncrement);
	sendSimMessage(newcontent, timeIncrement, destLP);
	count++;
	tSleep(500);
	holdCount++;
  }//end for
}


/**Handles events for Phold.
  *It is the counterpart of "Proc" in GTW.
  *@param content an Object representing the content of the message to be sent
  *@param sender String the application-assigned name of the sender of this message
  */
	@Override
public void appRunLP(Object content, String sender)
{
	System.out.println(alias + " run " + content);
	//if the message content is "done", then phold0 store a flag for the sender
	//String cont=(String)content;
	count=Integer.parseInt((String)content);
	String newcontent;
	int destLP;

	//only send totalSend of messages
	if(holdCount<totalSend)
	{

		//do dummy-cache
		//int a=random.nextInt();
		//int result=cacheManager.dummy1(this, a);	
		//System.out.println("\n" + alias + " get result: "  + result);		

		//get timestamp from exponential distribution
		double u=random.nextDouble();
		double lambda=1;
		double ts=nextExponential(u,lambda);
		timeIncrement=(int)(ts*10);
   
		//get destination from uniform distribution

		if(random.nextDouble() < PREFER_NEIGHBORS) {
			if(random.nextDouble() < 0.5) {
				destLP = getLID() + 1;
			}
			else {
				destLP = getLID() + numLPs - 1;
			}
			destLP %= numLPs;
		}
		else {
			destLP=random.nextInt(numLPs);
		}

		count++;
		holdCount++;
		newcontent=String.valueOf(count);

		//time stamp is current time plus time increment
		//int tim=getCurrentTime()+timeIncrement;

		if(STRACE)
			System.out.println(alias + " send " + newcontent + " to " + destLP + " at " 
						 + getCurrentTime() + " plus " + timeIncrement);
     
		sendSimMessage(newcontent, timeIncrement, destLP);
	}//if count<totalSend
	
	else
		setEndOfSimFlag();
}//end appRunLP

/**Sets application-specific configuration data.
 * @param cdata String representing configuration data, tab-delimilated with: 
 * numLPs msgPop totalSend
 */
	@Override
public void appSetConfigData(Serializable cdata)
{
	String args[]=((String)cdata).split("\t");
        numLPs=Integer.parseInt(args[0]); 
        msgPop=Integer.parseInt(args[1]); 
        totalSend=Integer.parseInt(args[2]);
        System.out.println("\n numLPs, msgPop, totalSend: "
                          + numLPs + " " + msgPop + " " + totalSend);
}//end appSetConfigData


/**Dummy method for testing caching mechanism*/
public int dummy1(int a)
{
	double da=(double)a;
	double db=(double)count;
	double p =Math.pow(da,db);
	int result=(int)p;
	return result;
}//end dummy1 

/**Dummy method for testing caching mechanism*/
public double dummy2(int a, int b)
{
	double da=(double)a;
	double db=(double)b;
	double p =Math.pow(da,db);
	return p; 
}//end dummy2 

/**Dummy method for testing caching mechanism*/
public String dummy3()
{
	return "test cache-aware functions";
}//end dummy3 

/**Terminates LP and displays simulation statistics. It is optionally called at the end of the
  *simulation. It is the counterpart of "FProc" in GTW.
  */
	@Override
public void appFinalizeLP()
{
	//System.out.println(alias + " is finalizing...\n");
}

void tSleep(int tim)
{
   try{
     Thread t=Thread.currentThread();
     Thread.sleep(tim);
   }
   catch(Exception e){
     e.printStackTrace();
  }
}//end tSleep

/**Returns a random number from exponential distributioin
  *@param u double the random number from a uniform distribution
  *@param lambda double the arrival rate
  */
double nextExponential(double u, double lambda)
{
   double d=Math.log(1-u)*(-1/lambda);
   return d;
}   


/**Implements abstract method to save LP state*/
	@Override
public Serializable appSaveLPState()
{
   Phold p=new Phold();
   p.count=count;
   return p;
}

/**Implements abstract method to set the state of this LP to the state of a specified Phold*/
public void appSetLPState(Serializable lp)
{
   Phold p=(Phold)lp;
   count=p.count;
}//end setState



	@Override
public String toString()
{
    return(alias+ " " + count);
}

/** Returns a Vector of NameValuePair describing all application-specific steerable
         * parameters of this LP, each NameValuePair is in the form "paramName, paramType".
         * @since 1.5
         */
	@Override
        public Vector<NameValuePair> appGetSteerable()
        {
                Vector<NameValuePair> nvVector=new Vector<NameValuePair>();
                NameValuePair nv;
                nv=new NameValuePair("count", "int");
                nvVector.add(nv);
                return nvVector;
        }//end appGetSteerable


        /** Returns a Vector of NameValuePair describing all application-specific monitorable
         * parameters of this LP, each NameValuePair is in the form "paramName, paramType".
         * @since 1.5
         */
	@Override
        public Vector<NameValuePair> appGetMonitorable()
        {
                Vector<NameValuePair> nvVector=new Vector<NameValuePair>();
                NameValuePair nv;
                nv=new NameValuePair("count", "int");
                nvVector.add(nv);
                return nvVector;
        }//end appGetMonitorable


}//end Phold
