package app;

import java.io.Serializable;
import java.util.*;
import sassy.*;

 /* This class can be used to test whether a Time Warp engine can handle simulataneous events.
  * Idea: sender s2 and s3 each send out 5 A-events or 5 B-events at time 1,2,3,4,5. So there
  * are 5 pairs of simulttaneous events, namely, A1, B1; A2, B2; ,,,.
  * If the kernel can handle simulataneous events, then the effective events received by s1 will
  * follow a unified order, with either A-events go first, or B-events go first. If not, then the
  * kernel does not implemented any mechanism to handle simultaneous events.
  * @author Yin Xiong
  * @version 1.0 (02/2007)
  */



public class SimulEvent extends sassy.LogicalProcess implements Serializable
{
	/**class member for the total of messages each sender will send*/
	int total=5;
	
	/**class member for keeping track of how many messages have been sent.*/
	int count;
	
	/**class member for keeping track of how many messages from A have been received.*/
	int aCount;
	
	/**class member for keeping track of how many messages from B have been received*/
	int bCount;
	
	/**variable to store time increments*/
	int[] incs;
	/**class member to store messages to send*/
	String[] outMsgs;
	
	/** class member for keeping received messages; only for s1, the receiver*/
	Vector<String> inMsgs;
	
	/**the application-specified name of the destination to which this LP will send messages to*/
	String dest;
	
	/**the alias of this LP: PEid + LPid*/
	String alias;
	
	/**Constructs a new SimulEvent instance.*/
	public SimulEvent()
	{
		super();
		count=0;
		aCount=0;
		bCount=0;
		inMsgs=new Vector<String>();
	}
	
	/**Sets application-specific configuration data.
	 * @param cdata String representing configuration data.
	 */
	@Override
	public void appSetConfigData(Serializable cdata)
	{
		alias=getPID()+ "-" + getLID();
		System.out.println(alias + " processing cdata...");
		String[] s=((String)cdata).split(" ");
		int slength=s.length;
		incs=new int[slength];
		outMsgs=new String[slength];
		
		for(int i=0;i<slength;i++)
		{
			String timeMsg=s[i];
			String[] tokens=timeMsg.split("/");
			incs[i]=Integer.parseInt(tokens[0]);
			outMsgs[i]=tokens[1];
		}
		for(int i=0;i<slength;i++)
		{
			System.out.println(incs[i] + " " + outMsgs[i]);
		}
		
		System.out.println();
	}//end appSetConfigData
	
	
	/**Initiates the LP and generates the initial message. It's the counterpart of "IProc"
	 *in GTW.
	 */
	@Override
	public void appInitializeLP()
	{
		System.out.println(alias + " is initializing...\n");
		
		//sender LP sends msgs to receiver LP
		if(!getAPPid().equalsIgnoreCase("s1"))
		{
			dest="s1";
			int timeIncrement=incs[count];
			sendSimMessage(outMsgs[count], timeIncrement, dest);
			System.out.println("app " + alias + " sending " + outMsgs[count] +
					" to "+  dest);
			count++;
			//send a message to oneself to propel forward
			sendSimMessage("tick", timeIncrement, getAPPid());
			//System.out.println("app " + alias + " sending msg to " + getAPPid());
		}//end if
	}
	
	
	/**Handles events for SimulEvent.
	 *It is the counterpart of "Proc" in GTW.
	 *@param content an Object representing the content of the message to be sent
	 *@param sender String the application-assigned name of the sender of this message
	 */
	public void appRunLP(Object content, String sender)
	{
		if(!getAPPid().equalsIgnoreCase("s1"))
		{
			if(count<total)
			{
				dest="s1";
				sendSimMessage(outMsgs[count],incs[count], dest);
				System.out.println("app " + alias + " sending " + outMsgs[count] +
						" to "+  dest);
				int timeIncrement=1;
				sendSimMessage("tick", timeIncrement, getAPPid());
				//System.out.println("app " + alias + " sending msg to self");
				count++;
			}//end if count<total
		}//end if it's sender
		
		//if it is receiver
		else
		{
			if(aCount<total || bCount<total)
			{
				String c=(String)content;
				inMsgs.add(c);
				if(sender.equalsIgnoreCase("s2"))
				{
					System.out.printf("A%d processed.\n", aCount);
					aCount++;
				}
				else
				{
					System.out.printf("B%d processed.\n", bCount);
					bCount++;
				}
				//System.out.println("A" +aCount + " " + "B"+bCount);
				
				if(aCount>=total && bCount>=total)
				{
					setEndOfSimFlag();
				}//end if
			}//end else
		}//end if it's receiver
	}
	
	
	/**Terminates LP and displays simulation statistics. It is optionally called at the end of the
	 *simulation. It is the counterpart of "FProc" in GTW.
	 */
	@Override
	public void appFinalizeLP()
	{
		System.out.println(alias + " is finalizing...\n");
		
	}
	
	
	@Override
	public String toString()
	{
		return(alias+ " " + count);
	}
	
	/**Implements abstract method to save LP state*/
	@Override
	public Serializable appSaveLPState()
	{
		SimulEvent p = new SimulEvent();
		p.count=count;
		p.aCount=aCount;
		p.bCount=bCount;
		return p;
	}
	
	
	/**Implements abstract method to set the state of this LP to the state of a specified Ping*/
	@Override
	public void appSetLPState(Serializable lp)
	{
		SimulEvent p=(SimulEvent)lp;
		count=p.count;
		aCount=p.aCount;
		bCount=p.bCount;
		
	}//end setState
	
	
	/*****************to be implemented*****************************/
	
	
	/**Method to get the value of a given application-level monitorable variable.
	 * @since 1.5
	 * @param name The name of the variable to monitor
	 * @return The value of the variable in question
	 */
	public Object appGetState(String name)
	{
		return null;
	}
	
	
	/**Method to set the value of a given application-level monitorable variable.
	 * @since 1.5
	 * @param name The name of the variable to steer
	 * @param val The value to assign to the variable
	 * @return Null if the steering was successful, else a String describing the
	 * error that occurred.
	 */
	public String appSetState(String name, Object val)
	{
		return "under programming";
	}
	
	
	
}//end SimulEvent
