package app.antsim;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import sassy.*;
import sassy.agent.*;

/**
 *
 * @author George Vulov <georgevulov@hotmail.com>
 */
public class AntAgent extends Agent implements Serializable
{
	/** Class storing the entire state of the head of an ant. */
	private static final class State implements Serializable
	{
		/** Current state of the pheromones in the world. */
		public Map<String, PheromoneAgent.UpdateMessage> pheromones = new HashMap<String, PheromoneAgent.UpdateMessage>();
		
		/** The current goal of the ant. */
		public double currGoalX, currGoalY;
		
		/** If the ant is carrying food, it returns to nest. Else, it seeks food. */
		public boolean isCarryingFood = false;
		
		/** Default constructor. */
		public State()
		{
		}
		
		/** Copy constructor. */
		public State(State other)
		{
			pheromones = new HashMap<String, PheromoneAgent.UpdateMessage>(other.pheromones);
			currGoalX = other.currGoalX;
			currGoalY = other.currGoalY;
			isCarryingFood = other.isCarryingFood;
		}
	}
	
	/** The current state of the ant. */
	private State state = new State();
	
	
	public AntAgent()
	{
	}


	@Override
	public void init()
	{
	}


	@Override
	public void readConfig(Serializable config)
	{
		AntEnvironmentCreator.InitInfo antConfig = (AntEnvironmentCreator.InitInfo)config;

		body = new Body2D(antConfig.x, antConfig.y, antConfig.theta);
		
		// The agent will have one body segment that is circular
		body.addSegment(new CircleSegment(0.0, 0.0, 20));
		
		//Set the initial goal at our first location, so we have to pick a new goal
		state.currGoalX = antConfig.x;
		state.currGoalY = antConfig.y;
	}


	@Override
	public void rcvMessage(Object content, String senderId)
	{
		if (content instanceof PheromoneAgent.UpdateMessage)
		{
			//New value of a pheromone cell
			state.pheromones.put(senderId, (PheromoneAgent.UpdateMessage)content);
		}
	}


	@Override
	public void think()
	{
		//Wait to get input from pheromones first.
		if (getCurrentTime() < 7)
			return;
		
		double currX = ((Body2D) body).getX();
		double currY = ((Body2D) body).getY();
		double currTheta = ((Body2D) body).getTheta();
		double radius = ((Body2D) body).getRadius();
		Map<String, PheromoneAgent.UpdateMessage> visibleLocations = getVisibleLocations();
		
		//If food (or nest) are within sight, they become our goals
		double ultimateGoalX, ultimateGoalY;
		if (state.isCarryingFood)
		{
			ultimateGoalX = SimParams.NEST_X;
			ultimateGoalY = SimParams.NEST_Y;
		}
		else
		{
			ultimateGoalX = SimParams.FOOD_X;
			ultimateGoalY = SimParams.FOOD_Y;
		}
		
		double distX = (currX - ultimateGoalX);
		double distY = (currY - ultimateGoalY);
		double distToUltimateGoal = Math.sqrt(distX * distX + distY * distY);
		if (distToUltimateGoal <= SimParams.ANT_SIGHT_RADIUS)
		{
			state.currGoalX = ultimateGoalX;
			state.currGoalY = ultimateGoalY;
		}
		
		//Ok, check if we have reached our current goal. If not, we just continue in the same direction
		//If we've reached the current goal, we much choose a new one.
		if (hasReachedCurrentGoal())
		{
			//If the current goal was the nest/food, change state!
			if (state.currGoalX == ultimateGoalX && state.currGoalY == ultimateGoalY)
			{
				state.isCarryingFood = !state.isCarryingFood;
			}
			
			//Find a new goal based on pheromone gradients
			chooseNewGoal(visibleLocations.values());
		}
		
		//Whatever the (updated) current goal is, move towards it
		double goalDistX = state.currGoalX - currX;
		double goalDistY = state.currGoalY - currY;
		double desiredAngle = Math.atan2(goalDistY, goalDistX);
		int speed = SimParams.ANT_SPEED; //Can move faster!
		action = new AgentAction(speed, desiredAngle - currTheta, 0);
		
		//
		// Top off the pheromone at the current location
		//
		for (Map.Entry<String, PheromoneAgent.UpdateMessage> idLocationPair : visibleLocations.entrySet())
		{
			PheromoneAgent.UpdateMessage location = idLocationPair.getValue();
			if (location.x <= currX && currX <= location.x + location.width &&
					location.y <= currY && currY <= location.y + location.width)
			{
				if (state.isCarryingFood)
					sendMessage(new PheromoneAgent.TopOffFoodPheromoneMessage(), senseThinkActDelay, idLocationPair.getKey());
				else
					sendMessage(new PheromoneAgent.TopOffHomePheromoneMessage(), senseThinkActDelay, idLocationPair.getKey());
			}
		}
		
		//
		//Send message to the GUI notifying it of our location and orientation
		//
		sendMessage(new AntGUI.AntInfo(currX, currY, currTheta, radius, state.isCarryingFood), senseThinkActDelay, "gui");
	}

	/** Returns true if we're seeking food and we've found a food item, or if we're returning
	 * to the nest and we've reached the nest. 
	 */
	private boolean hasReachedCurrentGoal()
	{
		double currX = ((Body2D) body).getX();
		double currY = ((Body2D) body).getY();
		
		double distSquared = (currX - state.currGoalX) * (currX - state.currGoalX) + 
				(currY - state.currGoalY) * (currY - state.currGoalY);
		
		return distSquared <= SimParams.ANT_DESTINATION_THRESHOLD * SimParams.ANT_DESTINATION_THRESHOLD;
	}

	
	/** We choose a new point to which we should travel. Locations with higher desired pheromone concentrations are
	 * more likely, while locations with lower pheromone concentrations are less likely. 
	 * In the absence of pheromones, all locations are equally likely.
	 */
	private void chooseNewGoal(Collection<PheromoneAgent.UpdateMessage> visibleLocations)
	{		
		//Just choose the max
//		double max = 0;
//		PheromoneAgent.UpdateMessage maxLocation = null;
//		for (PheromoneAgent.UpdateMessage location : visibleLocations)
//		{
//			double visiblePheromone;
//			if (state.isCarryingFood)
//				visiblePheromone = location.homePheromone;
//			else
//				visiblePheromone =  location.foodPheromone;
//			
//			if (visiblePheromone > max)
//			{
//				maxLocation = location;
//				max = visiblePheromone;
//			}
//		}
//		
//		//If no pheromones around, choose randomly
//		if (max == 0)
//		{
//			int index = (int)(Math.random() * (visibleLocations.size() - 1) + 0.5);
//			int i = 0;
//			for (PheromoneAgent.UpdateMessage location : visibleLocations)
//			{
//				if (i++ == index)
//				{
//					maxLocation = location;
//					break;
//				}
//			}
//		}
//		
//		state.currGoalX = maxLocation.x + maxLocation.width / 2;
//		state.currGoalY = maxLocation.y + maxLocation.width / 2;
//		
//		if (maxLocation != null)
//			return;
		
		//There's always some probability of entering a suboptimal cell
		//We give each cell some nonzero pheromone content, and increase the total pheromone levels
		//accordingly, so all the probabilities sum to 1
		final double MIN_PHEROMONE = SimParams.MAX_PHEROMONE * 0.00003;
		
		double totalVisiblePheromone = 0;
		double[] adjustedPheromoneLevels = new double[visibleLocations.size()];
		int i = 0;
		for (PheromoneAgent.UpdateMessage location : visibleLocations)
		{
			double visiblePheromone;
			if (state.isCarryingFood)
				visiblePheromone = location.homePheromone;
			else
				visiblePheromone =  location.foodPheromone;
			
			visiblePheromone += MIN_PHEROMONE;
			visiblePheromone = Math.pow(visiblePheromone, SimParams.PHEROMONE_GREEDINESS);
			adjustedPheromoneLevels[i++] = visiblePheromone; 
			
			totalVisiblePheromone += visiblePheromone;
		}
		
		//Pick where we'll go. We generate a number between 0 and 1 and see in which segment it falls
		double rand = Math.random();
		double sumSoFar = 0;
		i = 0;
		for (PheromoneAgent.UpdateMessage location : visibleLocations)
		{
			double probability = adjustedPheromoneLevels[i++] / totalVisiblePheromone;
			
			sumSoFar += probability;
			
			if (rand <= sumSoFar)
			{
				state.currGoalX = location.x + location.width / 2;
				state.currGoalY = location.y + location.width / 2;
				return;
			}
		}
		
		assert false : "Did not choose any neighboring location to go to! Probabilities should have summed to one";
	}
	
	/** Returns only the pheromone locations currently visible to the ant. */
	private Map<String, PheromoneAgent.UpdateMessage> getVisibleLocations()
	{
		double currX = ((Body2D) body).getX();
		double currY = ((Body2D) body).getY();
		
		Map<String, PheromoneAgent.UpdateMessage> visible = new HashMap<String, PheromoneAgent.UpdateMessage>();
		
		for (Map.Entry<String, PheromoneAgent.UpdateMessage> idPheromonePair : state.pheromones.entrySet())
		{
			PheromoneAgent.UpdateMessage pheromone = idPheromonePair.getValue();
			
			//Find minimum x distance to the box
			double xDist = Math.abs(pheromone.x - currX);
			xDist = Math.min(xDist, Math.abs(pheromone.x + pheromone.width - currX));
			
			//Minimum y distance to the box
			double yDist = Math.abs(pheromone.y - currY);
			yDist = Math.min(yDist, Math.abs(pheromone.y + pheromone.width - currY));
			
			//If the total distance is less than the visibility radius, return this location
			double distSq = xDist * xDist + yDist * yDist;
			if (distSq <= SimParams.ANT_SIGHT_RADIUS * SimParams.ANT_SIGHT_RADIUS)
			{
				visible.put(idPheromonePair.getKey(), pheromone);
			}
		}
		
		return visible;
	}
	

	@Override
	public Serializable saveHeadState()
	{
		return new State(state);
	}


	@Override
	public void restoreHeadState(Serializable savedThoughts)
	{
		state = (State)savedThoughts;
	}
}