package app.antsim;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import sassy.agent.*;
import sassy.agent.AgentMessage.MessageType;

/**
 * Pheromone agents are square agents responsible for managing pheromone levels in the environment
 * 
 * @author George Vulov (georgevulov@hotmail.com)
 */
public class PheromoneAgent extends Agent implements Serializable
{
	/** Represents the state of the thoughts of one pheromone agent. */
	final public static class PheromoneState implements Serializable
	{
		/** The level of the pheromone that leads to the nest. */
		public double homePheromone;
		
		/** Level of food pheromone. */
		public double foodPheromone;
		
		/** The current state of all the neighbors. */
		public Map<String, UpdateMessage> neighborStates = new HashMap<String, UpdateMessage>();
		
		public List<String> currentSubscribers = Collections.<String>emptyList();
		
		/** Copy constructor. */
		public PheromoneState(PheromoneState other)
		{
			homePheromone = other.homePheromone;
			foodPheromone = other.foodPheromone;
			neighborStates = new HashMap<String, UpdateMessage>(neighborStates);
			
			if (other.currentSubscribers != Collections.<String>emptyList())
			{
				currentSubscribers = new ArrayList<String>(other.currentSubscribers);
			}
			else
			{
				currentSubscribers = Collections.<String>emptyList();
			}
		}
		
		/** Default constructor. */
		public PheromoneState()
		{
		}
	}
	
	
	/** Initialization data for this agent. */
	public static class BodyInfo implements Serializable
	{
		/** The ID's of all the neighbors of this agent. */
		public List<String> neighborIds = new ArrayList<String>();
		
		public List<String> initialSubscribers = Collections.<String>emptyList();
		
		/** Top left corner of the square. */
		public int xMin;
		public int yMin;
		
		/** Width of the square. */
		public int width;
	}
	
	
	/** Class uses to send update messages to the GUI and to other agents. */
	public static class UpdateMessage implements Serializable
	{
		public final double homePheromone, foodPheromone;
		public final int x, y, width;
		
		public UpdateMessage(PheromoneAgent agent)
		{
			homePheromone = agent.state.homePheromone;
			foodPheromone = agent.state.foodPheromone;
			x = agent.bodyInfo.xMin;
			y = agent.bodyInfo.yMin;
			width = agent.bodyInfo.width;
		}
		
		@Override
		public String toString()
		{
			return String.format("[r=%d, c=%d, food=%.1f, home=%.1f]", y / SimParams.PHEROMONE_AGENT_WIDTH, 
					x / SimParams.PHEROMONE_AGENT_WIDTH, foodPheromone, homePheromone);
		}
	}
	
	public static class TopOffFoodPheromoneMessage implements Serializable
	{
	}
	
	public static class TopOffHomePheromoneMessage implements Serializable
	{
	}
	
	
	/** Current state of the pheromone agent. */
	private PheromoneState state = new PheromoneState();
	
	/** The data that this agent was initialized with. */
	private BodyInfo bodyInfo;
	
	
	@Override
	public void init()
	{
		state.foodPheromone = 0; //Math.random() * SimParams.MAX_PHEROMONE;
		state.homePheromone = 0; //Math.random() * SimParams.MAX_PHEROMONE;
		senseThinkActDelay = 3;
	}
	
	
	@Override
	public void readConfig(Serializable configData)
	{
		BodyInfo config = (BodyInfo)configData;
		bodyInfo = config;
		
		//Initialize the body
		Body2D myBody = new Body2D(config.xMin, config.yMin, 0);
		myBody.addSegment(new RectangleSegment(0, 0, config.width, config.width, 0));
		body = myBody;
		
		//Initialize the subscribers
		state.currentSubscribers = config.initialSubscribers;
		config.initialSubscribers = null;
	}

	private static boolean boxContains(int xMin, int yMin, int width, int x, int y)
	{
		return (xMin <= x && x <= xMin + width && yMin <= y && y <= yMin + width);
	}

	@Override
	public void think()
	{
		//
		//Apply diffusion operator. Diffusion only takes into account immediate neighbors (not diagonal neighbors)
		//
		if (getCurrentTime() > senseThinkActDelay * 2)
		{
			int neighborCount = 0;
			double foodPheromoneSum = 0, homePheromoneSum = 0;
			for (UpdateMessage neighborInfo : state.neighborStates.values())
			{
				if (neighborInfo.x == bodyInfo.xMin || neighborInfo.y == bodyInfo.yMin)
				{
					//We don't perform diffusion from the home or from the food source
					//Otherwise, we get an automatic gradient
//					if (boxContains(neighborInfo.x, neighborInfo.y, neighborInfo.width, SimParams.FOOD_X, SimParams.FOOD_Y) ||
//							boxContains(neighborInfo.x, neighborInfo.y, neighborInfo.width, SimParams.NEST_X, SimParams.NEST_Y))
//					{
//						continue;
//					}
					
					neighborCount++;
					foodPheromoneSum += neighborInfo.foodPheromone;
					homePheromoneSum += neighborInfo.homePheromone;
				}
			}
			assert neighborCount >= 2 && neighborCount <= 4 : "Too many or too few immediate neighbors!";

			foodPheromoneSum -= neighborCount * state.foodPheromone;
			homePheromoneSum -= neighborCount * state.homePheromone;

			//Now we have applied the diffusion stencil. Scale by the diffusion rate
			foodPheromoneSum *= SimParams.DIFFUSION_RATE;
			homePheromoneSum *= SimParams.DIFFUSION_RATE;

			//Appy the diffusion change to the pheromone values
			state.foodPheromone += foodPheromoneSum;
			state.homePheromone += homePheromoneSum;

		
			//
			//Evaporation
			//
			state.foodPheromone *= (1 - SimParams.EVAPORATION_RATE);
			state.homePheromone *= (1 - SimParams.EVAPORATION_RATE);			
		}
		
		//The food location and the nest location are always at maximum pheromone levels
		if (bodyInfo.xMin <= SimParams.FOOD_X && SimParams.FOOD_X <= bodyInfo.xMin + bodyInfo.width &&
				bodyInfo.yMin <= SimParams.FOOD_Y && SimParams.FOOD_Y <= bodyInfo.yMin + bodyInfo.width)
		{
			state.foodPheromone = SimParams.MAX_PHEROMONE;
		}
		if (bodyInfo.xMin <= SimParams.NEST_X && SimParams.NEST_X <= bodyInfo.xMin + bodyInfo.width &&
				bodyInfo.yMin <= SimParams.NEST_Y && SimParams.NEST_Y <= bodyInfo.yMin + bodyInfo.width)
		{
			state.homePheromone = SimParams.MAX_PHEROMONE;
		}
		
		//
		//Send a message to the GUI letting it know our updated state.
		//
		UpdateMessage message = new UpdateMessage(this);
		sendMessage(message, senseThinkActDelay, "gui");
		
		//
		//Send a message to all our neighbors informing them of the new state
		//
		for (String neighborId : bodyInfo.neighborIds)
		{
			sendMessage(message, senseThinkActDelay, neighborId);
		}
		
		//Send a message to all the subscribers informing them of the new state
		for (String subscriber : state.currentSubscribers)
		{
			sendMessage(message, senseThinkActDelay, subscriber);
		}
	}
	

	@Override
	public void rcvMessage(Object content, String senderId)
	{
		if (content instanceof UpdateMessage)
		{
			//We got an update from a neighbor notifying us if its state
			UpdateMessage update = (UpdateMessage)content;

			assert bodyInfo.neighborIds.indexOf(senderId) != -1 : "Pheromone cell receing message from unknown cell";

			state.neighborStates.put(senderId, update);
		}
		else if (content instanceof TopOffFoodPheromoneMessage) //Dummy comparison for now
		{
			double maxFood = 0;
			for (UpdateMessage neighborInfo : state.neighborStates.values())
			{
				maxFood = Math.max(maxFood, neighborInfo.foodPheromone);
			}
			
			double desiredLevel = Math.max(0, maxFood - SimParams.PHEROMONE_DESIRABILITY_DECREASE);
			state.foodPheromone = desiredLevel;
		}
		else if (content instanceof TopOffHomePheromoneMessage)
		{
			double maxHome = 0;
			for (UpdateMessage neighborInfo : state.neighborStates.values())
			{
				maxHome = Math.max(maxHome, neighborInfo.homePheromone);
			}
			
			double desiredLevel = Math.max(0, maxHome - SimParams.PHEROMONE_DESIRABILITY_DECREASE);
			state.homePheromone = desiredLevel;
		}
	}


	/** Support for rollback. */
	@Override
	public Serializable saveHeadState()
	{
		return null;
		//return new PheromoneState(state);
	}


	/** Support for rollback. */
	@Override
	public void restoreHeadState(Serializable savedHead)
	{
		//state = (PheromoneState)savedHead;
	}	
}
