package im;

import im.AgentMsg.MessageType;

import java.awt.Point;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Map;

import sassy.LogicalProcess;

/**
 * A ball is just a dummy agent which can bounce with environment wall or other balls.
 * The ball agent is a application for testing and evaluating interest management middleware for 
 * SASSY.
 * @author hth
 * @version general
 */
public class Ball extends LogicalProcess implements Serializable
{
	//	private transient CacheManager;
	private static final long serialVersionUID = 1234L;

	/**total number of the ball's movement throughout the simulation*/
	private int totalMove;

	/**the time increment for messages sent by this LP; the message will be sent at currentTime+timeIncrement*/
	private final int timeIncrement = 0;

	private final int moveIncrement = 1;

	private final int collisionIncrement = 0;

	/**the application-specified name of the monitor LP*/
	private String moniDest;

	/**ball's radius*/
	private int radius;

	/**enviroment size*/
	private Point envi;

	/**total number of interest manager is: numberGrid*numberGrid*/
	private int numberGrid;

	/**ball's coordinate Position */
	private Point currPosition = new Point();

	/**ball's coordinate Position */
	private Point lastPosition = new Point();

	/**ball's next coordinate Position*/
	private Point nextPosition = new Point();

	/**ball's two dimension velocity*/
	private Point currVelocity = new Point();

	/**ball's next velocity*/
	private Point nextVelocity = new Point();

	/**local world view of this ball, all balls around.*/
	private transient HashMap<String, MsgBall> ballAround = new HashMap<String, MsgBall>();

	private transient HashMap<String, List<String>> ballLocations = new HashMap<String, List<String>>();

	//terminator address
	private String terAddr;

	//communication commType
	private String commType;

	private final boolean symmetric = true;

	private int ratio = 0;


	@Override
	public void appSetConfigData(Serializable cdata)
	{
		String[] s = ((String)cdata).split(" ");
		/*radius = Double.parseDouble(s[0]);*/
		radius = Integer.parseInt(s[0]);
		envi = new Point(Integer.parseInt(s[1]), Integer.parseInt(s[2]));
		numberGrid = Integer.parseInt(s[3]);
		lastPosition.x = Integer.parseInt(s[4]);
		lastPosition.y = Integer.parseInt(s[5]);
		currPosition.x = Integer.parseInt(s[4]);
		currPosition.y = Integer.parseInt(s[5]);
		nextPosition.x = Integer.parseInt(s[4]);
		nextPosition.y = Integer.parseInt(s[5]);
		currVelocity.x = Integer.parseInt(s[6]);
		currVelocity.y = Integer.parseInt(s[7]);
		nextVelocity.x = Integer.parseInt(s[6]);
		nextVelocity.y = Integer.parseInt(s[7]);

		//updateIntervel = Integer.parseInt(s[8]);
		totalMove = Integer.parseInt(s[9]);
		//numBalls = Integer.parseInt(s[10]);
		terAddr = s[11];
		commType = s[12];

		//check whether a GUI is used.
		if (s.length > 13)
		{
			moniDest = s[13];
		}
	}


	@Override
	public void appInitializeLP()
	{
		//If the simulation is asymmetric, make odd balls see further
		int ballID = Integer.parseInt(getAPPid().substring(4));
		if (!symmetric)
		{
			if (ballID % 2 != 0)
				ratio++;
		}

		//get overlap IMLPs
		ArrayList<Integer> overlapIMLPs = getOverlapIMLPsFromPosition(currPosition);

		//get subscribe IMLPs
		ArrayList<Integer> subIMLPs = getLargerOverlapIMLPsFromPosition(currPosition);

		//sends subscribe messages
		for (int i = 0; i < subIMLPs.size(); i++)
			sendSimMessage(new AgentMsg(MessageType.SUBSCRIBE, new MsgBall(getAPPid(), null, null)), timeIncrement, subIMLPs.get(i).toString());

		//sends enter messages
		for (int i = 0; i < overlapIMLPs.size(); i++)
			sendSimMessage(new AgentMsg(MessageType.ENTER, new MsgBall(getAPPid(), currPosition, currVelocity)), timeIncrement, overlapIMLPs.get(i).toString());

		//schedules next move 1/14
		AgentMsg moveMess = new AgentMsg(AgentMsg.MessageType.SENSE, new MsgBall(getAPPid(), null, null));
		sendSimMessage(moveMess, moveIncrement, getAPPid());
	}


	@Override
	public void appRunLP(Object content, String sender)
	{
		System.out.println("Got a message from: " + sender);

		AgentMsg agentMsg = (AgentMsg)content;
		AgentMsg.MessageType msgType = agentMsg.getType();
		MsgBall msgBall = agentMsg.getMsgBall();
		
		switch (msgType)
		{
			//direct communication: receives update message directly from agents
			case UPDATE:

				if (ballAround.containsKey(msgBall.getAPPid()))
				{
					ballAround.put(msgBall.getAPPid(), msgBall);
				}
				else
				{
					printRollbackSafe(getAPPid() + " update: the incoming update from a unknown ball " + msgBall.getAPPid());
				}
				break;
				
			//Direct communication: another ball has entered the region
			case SUBSCRIBE:
				ballAround.put(msgBall.getAPPid(), msgBall);
				if (!ballLocations.containsKey(msgBall.getAPPid()))
					ballLocations.put(msgBall.getAPPid(), new ArrayList<String>());
				if (!ballLocations.get(msgBall.getAPPid()).contains(sender))
					ballLocations.get(msgBall.getAPPid()).add(sender);
				
				break;
			
			//Direct communication: another ball has left the region
			case UNSUBSCRIBE:
				if (ballAround.containsKey(msgBall.getAPPid()))
				{
					ballLocations.get(msgBall.getAPPid()).remove(sender);

					if (ballLocations.get(msgBall.getAPPid()).isEmpty())
					{
						ballLocations.remove(msgBall.getAPPid());
						ballAround.remove(msgBall.getAPPid());
					}
				}
				break;

			//relay communication: receive update message from IMLP
			case REFLECTUPDATE:

				if (ballAround.containsKey(msgBall.getAPPid()))
				{
					ballAround.put(msgBall.getAPPid(), msgBall);
				}
				else
				{
					printRollbackSafe(getAPPid() + " update: the incoming update from a unknown ball " + msgBall.getAPPid());
				}
				break;

			case REFLECTENTER:

				printRollbackSafe(getAPPid() + " reflectEnter: " + msgBall.getAPPid() + " through " + sender);

				//first enter
				if (!ballAround.containsKey(msgBall.getAPPid()))
				{
					ballLocations.put(msgBall.getAPPid(), new ArrayList<String>());
					ballAround.put(msgBall.getAPPid(), msgBall);
				}

				//Add new location to the locations of the ball
				if (!ballLocations.get(msgBall.getAPPid()).contains(sender))
					ballLocations.get(msgBall.getAPPid()).add(sender);

				break;

			case REFLECTLEAVE:

				printRollbackSafe(getAPPid() + " reflectLeave: " + msgBall.getAPPid() + " through " + sender);

				if (ballAround.containsKey(msgBall.getAPPid()))
				{
					ballLocations.get(msgBall.getAPPid()).remove(sender);

					if (ballLocations.get(msgBall.getAPPid()).isEmpty())
					{
						ballLocations.remove(msgBall.getAPPid());
						ballAround.remove(msgBall.getAPPid());
					}
				}
				else
				{
					printRollbackSafe(getAPPid() + " reflectLeave: the incoming leave message from a unknown ball " + msgBall.getAPPid());
				}

				break;


			case COLLISION:

				if (ballAround.containsKey(msgBall.getAPPid()))
				{
					nextVelocity.setLocation(msgBall.getCurrVelocity());
					nextPosition.setLocation(currPosition.x + nextVelocity.x, currPosition.y + nextVelocity.y);
				}
				else
				{
					printRollbackSafe(getAPPid() + " Collision Msg: the incoming collision message from a unknown ball " + msgBall.getAPPid());
				}
				break;

			//receives move message from itself, compute next Position and schedule next move to itself
			case SENSE:

				//think for a while
				try
				{
					Thread.sleep((long)(Math.random() * 100));
				}
				catch (InterruptedException e)
				{
				}

//			if(this.getCurrentTime()%5==0)	System.out.println(this.getName()+": at time: "+this.getCurrentTime()+" y: "+this.currPosition.y);

				//We have reached the end of the simulation; don't move anymore
				if (this.getCurrentTime() > totalMove)
				{
					sendSimMessage(new AgentMsg(MessageType.UPDATE, new MsgBall(getAPPid(), null, null)), timeIncrement, terAddr);
					return;
				}

				//updates the new states
				lastPosition.setLocation(currPosition);
				currPosition.setLocation(nextPosition);
				currVelocity.setLocation(nextVelocity);

				nextPosition.setLocation(currPosition.x + currVelocity.x, currPosition.y + currVelocity.y);//0225

				//overlap with other balls, realize overlaps with other ball, exchange velocity
				for (MsgBall ball : ballAround.values())
				{
					if (ball.getCurrPosition() == null)
					{
						continue;
					}

					//distance between two ball center
					double dis = currPosition.distance(ball.getCurrPosition());

					//if two ball overlap 
					if (dis <= 2 * radius)
					{

						//	printRollbackSafe("*********"+this.getName()+ " this velocity "+currVelocity.x+" "+currVelocity.y+ " time is : "+this.getCurrentTime());

						//Exchange two ball's velocity, it is just a simple reaction of bouncing ball
						nextVelocity.setLocation(-ball.getCurrVelocity().x, -ball.getCurrVelocity().y);
						nextPosition.setLocation(currPosition.x + nextVelocity.x, currPosition.y + nextVelocity.y);//0225

						//send a collision message to the other ball
						//sendSimMessage(new AgentMsg(MessageType.COLLISION,new MsgBall(getAPPid(),currPosition,currVelocity)), collisionIncrement, ball.getAPPid());
						break;
					}
				}

				//bounces with the left side wall
				if (nextPosition.x - radius < 0)
				{
					nextPosition.x = radius;
					nextVelocity.x = -currVelocity.x;
				}
				//bounces with the right side wall   
				else if (nextPosition.x > envi.x - radius)
				{
					nextPosition.x = envi.x - radius;
					nextVelocity.x = -currVelocity.x;
				}
				//bounces with the ceiling
				if (nextPosition.y - radius < 0)
				{
					nextPosition.y = radius;
					nextVelocity.y = -currVelocity.y;
				}
				//bounces with the ground
				else if (nextPosition.y > envi.y - radius)
				{
					nextPosition.y = envi.y - radius;
					nextVelocity.y = -currVelocity.y;
				}


				//Update the region in which the ball publishes
				ArrayList<Integer> newOverlapIMLPs = getOverlapIMLPsFromPosition(currPosition);
				ArrayList<Integer> oldOverlapIMLPs = getOverlapIMLPsFromPosition(lastPosition);

				if (!newOverlapIMLPs.containsAll(oldOverlapIMLPs) || !oldOverlapIMLPs.containsAll(newOverlapIMLPs))
				{
					ArrayList<Integer> newSubIMLPs = getLargerOverlapIMLPsFromPosition(currPosition);
					ArrayList<Integer> oldSubIMLPs = getLargerOverlapIMLPsFromPosition(lastPosition);

					//sends leave message
					for (int i = 0; i < oldOverlapIMLPs.size(); i++)
					{
						//sends leave message to cells in old list but not the new one
						if (!newOverlapIMLPs.contains(oldOverlapIMLPs.get(i)))
						{
							sendSimMessage(new AgentMsg(MessageType.LEAVE, new MsgBall(getAPPid(), null, null)), timeIncrement, oldOverlapIMLPs.get(i).toString());
						}
					}

					//sends subscribe message
					for (int i = 0; i < newSubIMLPs.size(); i++)
					{
						//subscribes to cells in new list but not the old one
						if (!oldSubIMLPs.contains(newSubIMLPs.get(i)))
						{
							sendSimMessage(new AgentMsg(MessageType.SUBSCRIBE, new MsgBall(getAPPid(), null, null)), timeIncrement, newSubIMLPs.get(i).toString());
						}
					}

					//sends unsubscribe message 
					for (int i = 0; i < oldSubIMLPs.size(); i++)
					{
						//sends unsub message to cells in old list but not the new one
						if (!newSubIMLPs.contains(oldSubIMLPs.get(i)))
						{
							unsubscribeIMLP(oldSubIMLPs.get(i).toString());
						}
					}

					//sends enter message to new IMLP
					for (int i = 0; i < newOverlapIMLPs.size(); i++)
					{
						//send enter message to cells in new list but not the old one
						if (!oldOverlapIMLPs.contains(newOverlapIMLPs.get(i)))
						{
							sendSimMessage(new AgentMsg(MessageType.ENTER, new MsgBall(getAPPid(), currPosition, currVelocity)), timeIncrement, newOverlapIMLPs.get(i).toString());
						}
					}
				}

				//No new subscriptions/unsubscriptions
				if (commType.equals("relay"))
				{
					//sends update message 
					for (int i = 0; i < newOverlapIMLPs.size(); i++)
						sendSimMessage(new AgentMsg(MessageType.UPDATE, new MsgBall(getAPPid(), currPosition, currVelocity)), timeIncrement, newOverlapIMLPs.get(i).toString());
				}
				else
				{
					for (String id : ballAround.keySet())
					{
						sendSimMessage(new AgentMsg(MessageType.UPDATE, new MsgBall(getAPPid(), currPosition, currVelocity)), moveIncrement, id);
					}
				}

				//Publish to monitor LP for demonstration
				if (moniDest != null)
				{
					sendSimMessage(new AgentMsg(MessageType.UPDATE, new MsgBall(getAPPid(), currPosition, currVelocity, ballAround.size())), timeIncrement, moniDest);
				}

				//Schedule next move
				sendSimMessage(new AgentMsg(AgentMsg.MessageType.SENSE, new MsgBall(getAPPid(), null, null)), moveIncrement, getAPPid());
				break;

			default:
				printRollbackSafe("!!!!!!!!Unknown message commType in ball");
				break;
		}
	}


	/**
	 * @param position the position of ball's center point
	 * @return all grid covered in the rectangle region
	 */
	private ArrayList<Integer> getOverlapIMLPsFromPosition(Point position)
	{
		final float SAFETY_FACTOR = 1.1f;
		float gridSizeX = (float) envi.x / numberGrid;
		float gridSizeY = (float) envi.y / numberGrid;

		int xLeft = (int)((position.x - radius * SAFETY_FACTOR) / gridSizeX);
		int xRight = (int)((position.x + radius * SAFETY_FACTOR) / gridSizeX);
		int yBottom = (int)((position.y + radius * SAFETY_FACTOR) / gridSizeY);
		int yTop = (int)((position.y - radius * SAFETY_FACTOR) / gridSizeY);

		if (xLeft < 0)
			xLeft = 0;
		if (xRight >= numberGrid)
			xRight = numberGrid - 1;
		if (yBottom >= numberGrid)
			yBottom = numberGrid - 1;
		if (yTop < 0)
			yTop = 0;

		ArrayList<Integer> overlapIMLPs = new ArrayList<Integer>();

		for (int y = yTop; y <= yBottom; y++)
		{
			for (int x = xLeft; x <= xRight; x++)
			{
				overlapIMLPs.add(y * numberGrid + x);
			}
		}

		return overlapIMLPs;
	}


	/**
	 * @param position the position of ball's center point
	 * @return enlarged grid region
	 */
	private ArrayList<Integer> getLargerOverlapIMLPsFromPosition(Point position)
	{
		final float SAFETY_FACTOR = 1.1f;
		float gridSizeX = (float) envi.x / numberGrid;
		float gridSizeY = (float) envi.y / numberGrid;

		int xLeft = (int)((position.x - radius * SAFETY_FACTOR) / gridSizeX) - ratio;
		int xRight = (int)((position.x + radius * SAFETY_FACTOR) / gridSizeX) + ratio;
		int yBottom = (int)((position.y + radius * SAFETY_FACTOR) / gridSizeY) + ratio;
		int yTop = (int)((position.y - radius * SAFETY_FACTOR) / gridSizeY) - ratio;

		if (xLeft < 0)
			xLeft = 0;
		if (xRight >= numberGrid)
			xRight = numberGrid - 1;
		if (yBottom >= numberGrid)
			yBottom = numberGrid - 1;
		if (yTop < 0)
			yTop = 0;

		ArrayList<Integer> overlapIMLPs = new ArrayList<Integer>();

		for (int y = yTop; y <= yBottom; y++)
		{
			for (int x = xLeft; x <= xRight; x++)
			{
				overlapIMLPs.add(y * numberGrid + x);
			}
		}

		return overlapIMLPs;
	}


	@Override
	public Serializable appSaveLPState()
	{
		Ball ball = new Ball();
		ball.currPosition = new Point(currPosition);
		ball.currVelocity = new Point(currVelocity);
		ball.lastPosition = new Point(lastPosition);
		ball.nextPosition = new Point(nextPosition);
		ball.nextVelocity = new Point(nextVelocity);
		ball.ballAround = new HashMap<String, MsgBall>(ballAround);

		//Do a deep copy of the ball locations
		ball.ballLocations = new HashMap<String, List<String>>();
		for (Map.Entry<String, List<String>> keyValuePair : ballLocations.entrySet())
		{
			List<String> copyList = new ArrayList<String>(keyValuePair.getValue());
			ball.ballLocations.put(keyValuePair.getKey(), copyList);
		}

		return ball;
	}


	@Override
	public void appSetLPState(Serializable lp)
	{
		Ball ball = (Ball)lp;
		currPosition = ball.currPosition;
		currVelocity = ball.currVelocity;
		lastPosition = ball.lastPosition;
		nextPosition = ball.nextPosition;
		nextVelocity = ball.nextVelocity;
		ballAround = ball.ballAround;
		ballLocations = ball.ballLocations;
	}


	/**
	 * Send the unsubscribe message to a certain IMLP and update the local
	 * data structures.
	 * 
	 * @param imlpId appId of the IMLP from which to unsubscribe
	 */
	private void unsubscribeIMLP(String imlpId)
	{
		//If we unsubscribe from an IMLP, remove our knowledge of any balls there
		Set<String> keys = new HashSet<String>(ballLocations.keySet());
		for (String ballId : keys)
		{
			List<String> imlps = ballLocations.get(ballId);

			imlps.remove(imlpId);

			if (imlps.isEmpty())
			{
				ballLocations.remove(ballId);
				ballAround.remove(ballId);
			}
		}

		//Now send the unsubscribe message
		sendSimMessage(new AgentMsg(MessageType.UNSUBSCRIBE, new MsgBall(getAPPid(), null, null)), timeIncrement, imlpId);
	}


	private void printRollbackSafe(final String msg)
	{
//		submitIrreversibleAction(new Runnable()
//		{
//			public void run()
//			{
//				System.out.println(msg);
//			}
//		});
	}


	@Override
	public void appFinalizeLP()
	{
	}
//	/**
//	 * computes the cell ID base on the ball's Position
//	 * @param Position: the coordinate position for the ball
//	 * @param envior: the size of the envioronment in coordinate format
//	 * @param numGrid: number of grid base, total number of cell is numGrid * numGrid
//	 * @return: the cell ID in the formate (1.2.3)etc.
//	 */
//	private int getIMLPIDFromPosition(Point Position, Point envior, int numGrid)
//	{
//		if (Position.getX() < 0 || Position.getY() < 0)
//		{
//			return -1;
//		}
//		double gradSizeX = envior.getX() / numGrid;
//		double gradSizeY = envior.getY() / numGrid;
//		return (int)(Position.getY() / gradSizeY) * numGrid + (int)(Position.getX() / gradSizeX);
//	}
//
//
//	/**
//	 * converts coordinate ID to integer ID
//	 * @param gridIDCoordinate: cell ID in coordinate format
//	 * @return integer cell ID
//	 */
//	private int getGridIDFromCoordinate(Point gridIDCoordinate)
//	{
//		if ((int)gridIDCoordinate.getY() >= numberGrid || (int)gridIDCoordinate.getX() >= numberGrid)
//		{
//			return -1;
//		}
//		else
//		{
//			return (int)gridIDCoordinate.getY() * numberGrid + (int)gridIDCoordinate.getX();
//		}
//	}
//
//
//	/**
//	 * converts integer ID to coordinate ID
//	 * @param gridID: cell ID in integer format
//	 * @return cell's coordinate ID
//	 */
//	private Point getCoordinateFromID(int gridID)
//	{
//		Point gridPair = new Point();
//		gridPair.y = gridID / numberGrid;
//		gridPair.x = gridID % numberGrid;
//		return gridPair;
//	}
//	
//	/**
//	 * computes the neighbor list around a given cell
//	 * @param gridX: the X dimension index of the cell
//	 * @param gridY: the Y dimension index of the cell
//	 * @param numGrid: the number of grid base
//	 * @return neighbor list around the given cell
//	 */
//	private LinkedList<Integer> getNeighborIMID(int gridX, int gridY, int numGrid)
//	{
//
//		if (numGrid <= gridX || numGrid <= gridY)
//		{
//			return null;
//		}
//
//		LinkedList<Integer> neighbor = new LinkedList<Integer>();
//		int gridID = gridY * numGrid + gridX;
//		neighbor.add(gridID);
//		if (gridX != 0 && gridY != 0)
//		{
//			neighbor.add(gridID - numGrid - 1);
//		}
//		if (gridX != 0 && gridY != numGrid - 1)
//		{
//			neighbor.add(gridID + numGrid - 1);
//		}
//		if (gridX != numGrid - 1 && gridY != 0)
//		{
//			neighbor.add(gridID - numGrid + 1);
//		}
//		if (gridX != numGrid - 1 && gridY != numGrid - 1)
//		{
//			neighbor.add(gridID + numGrid + 1);
//		}
//
//		if (gridX != 0)
//		{
//			neighbor.add(gridID - 1);
//		}
//		if (gridX != numGrid - 1)
//		{
//			neighbor.add(gridID + 1);
//		}
//		if (gridY != 0)
//		{
//			neighbor.add(gridID - numGrid);
//		}
//		if (gridY != numGrid - 1)
//		{
//			neighbor.add(gridID + numGrid);
//		}
//		return neighbor;
//	}
}
