package im;

import im.AgentMsg.MessageType;
import java.io.Serializable;
import java.util.HashMap;
import sassy.LogicalProcess;

/**
 * Grid based interest manager cells representing world state for agent based simulator.
 * Each IMLP manages a publisher list and a subscriber list. The interest manager will 
 * receives messages from publisher and reflects these messages to their subscribers.
 * @author hth
 * @version general
 */
public class InterestManager extends LogicalProcess implements Serializable
{
	private static final long serialVersionUID = 1L;

	/**subscriber list*/
	private HashMap<String, MsgBall> subscriberList;

	/**publisher list*/
	private HashMap<String, MsgBall> publisherList;

	/**the time increment for messages sent by this LP; the message will be sent at currentTime+timeIncrement*/
	private int timeIncrement = 0;

	/**type of the communication supported*/
	private String communicationType;


	public InterestManager()
	{
//		super(true);
		super(false);
		publisherList = new HashMap<String, MsgBall>();
		subscriberList = new HashMap<String, MsgBall>();
	}


	/**
	 * <p>Invoked by an subscirbe message from the subscribe agent. Adds the agent to the subscriber list</p>
	 * <p>Version Direct: Reflects a HELLO message indicating the subscription with 0 lookahead.</p>
	 * @param ball the agent who subscribe to this IMLP
	 */
	private void subscribe(MsgBall ball)
	{
		printRollbackSafe(this.getName() + ": sub: " + ball.getAPPid() + " : " + this.getCurrentTime());

		//add to subscriber list
		subscriberList.put(ball.getAPPid(), ball);

		//If the communication is relayed, then the IMLP has a copy of the current state in the region
		//This cached copy is sent to the new subscriber
		if (communicationType.equals("relay"))
		{
			//Notify the ball of all the other balls in the region
			for (String id : publisherList.keySet())
			{
				if (!id.equals(ball.getAPPid()))
				{
					sendSimMessage(new AgentMsg(MessageType.REFLECTENTER, publisherList.get(id)), timeIncrement, ball.getAPPid());
				}
			}
		}
		//If the communication is direct, this means that this ball is now a subscriber to all the other balls
		//in the region
		else if (communicationType.equals("direct"))
		{
			for (String id : publisherList.keySet())
			{
				if (!id.equals(ball.getAPPid()))
				{
					sendSimMessage(new AgentMsg(MessageType.SUBSCRIBE, ball), timeIncrement, id);
				}
			}
		}
	}


	/**
	 * <p>Invoked by an upsubscirbe message from the unsubscribe agent. Removes the agent from the subscriber list</p>
	 * <p>Version Direct: Reflects a BYE message indicating the unsubscription with 0 lookahead.</p>
	 * @param ball the agent who unsubscribe from this IMLP
	 */
	private void unsubscribe(MsgBall ball)
	{
		printRollbackSafe(this.getName() + ": unsub: " + ball.getAPPid() + " : " + this.getCurrentTime());

		//remove from subscriber list
		if (subscriberList.containsKey(ball.getAPPid()))
		{
			subscriberList.remove(ball.getAPPid());
		}
		else
		{
			printRollbackSafe(this.getName() + ": the unsub ball is not in the sub list: " + ball.getAPPid() + " : " + this.getCurrentTime());
		}

		//If the communication is direct, unsubscribe the ball from all publishers
		if (communicationType.equals("direct"))
		{
			for (String id : publisherList.keySet())
			{
				if (!id.equals(ball.getAPPid()))
				{
					sendSimMessage(new AgentMsg(MessageType.UNSUBSCRIBE, ball), timeIncrement, id);
				}
			}
		}
	}


	/**
	 * <p>Invoked by an update message from the publisher. </p>
	 * <p>version Relay: The IMLP Reflects the update message to all subscribers of this IMLP with 0 lookahead.</p>
	 * @param ball the agent who sends the update message
	 */
	private void update(MsgBall ball)
	{
		//Update our local copy of the environment state
		publisherList.put(ball.getAPPid(), ball);

		//Relay the update message to all subscribers
		for (String id : subscriberList.keySet())
		{
			if (!id.equals(ball.getAPPid()))
				sendSimMessage(new AgentMsg(MessageType.REFLECTUPDATE, ball), timeIncrement, id);
		}
	}


	/**
	 * Invoked by an enter message from the agent just entered this IMLP, the agent becomes a publisher of this IMLP. 
	 * Adds the agent to the publisher list and inform subscribers by send a REFLECTUPDATE message.
	 * @param ball the agent enters this IMLP
	 */
	private void enter(MsgBall ball)
	{
		//Add the entry to the list of publishers
		publisherList.put(ball.getAPPid(), ball);

		if (communicationType.equals("relay"))
		{
			//Notify all subscribers of the new entry
			for (String id : subscriberList.keySet())
			{
				if (!id.equals(ball.getAPPid()))
					sendSimMessage(new AgentMsg(MessageType.REFLECTENTER, ball), timeIncrement, id);
			}
		}
		//In direct communication, when a new ball enters the region, all the
		//subscribers must receive updates from it
		else if (communicationType.equals("direct"))
		{
			for (MsgBall subscriber : subscriberList.values())
			{
				if (!subscriber.getAPPid().equals(ball.getAPPid()))
					sendSimMessage(new AgentMsg(MessageType.SUBSCRIBE, subscriber), timeIncrement, ball.getAPPid());
			}
		}
	}


	/**
	 * Invoked by a leave message from the left agent, which means the agent no long be a publisher of this IMLP. Removes the agent from the publisher list.
	 * @param ball the agent leaves this IMLP
	 */
	private void leave(MsgBall ball)
	{
		if (!publisherList.containsKey(ball.getAPPid()))
		{
			printRollbackSafe(this.getName() + ": the leave ball is not in the publisher list" + ball.getAPPid() + " : " + this.getCurrentTime());
			return;
		}

		//Leave the cell
		publisherList.remove(ball.getAPPid());


		if (communicationType.equals("relay"))
		{
			for (String id : subscriberList.keySet())
			{
				if (!id.equals(ball.getAPPid()))
					sendSimMessage(new AgentMsg(MessageType.REFLECTLEAVE, ball), timeIncrement, id);
			}
		}
		
		//In a direct communication, when the ball leaves the region it must no longer send
		//updates to all the subscribers of the region
		else if (communicationType.equals("direct"))
		{
			for (MsgBall subscriber : subscriberList.values())
			{
				if (!subscriber.getAPPid().equals(ball.getAPPid()))
					sendSimMessage(new AgentMsg(MessageType.UNSUBSCRIBE, subscriber), timeIncrement, ball.getAPPid());
			}
		}
	}


	@Override
	public void appFinalizeLP()
	{
	}


	@Override
	public void appInitializeLP()
	{
	}


	@Override
	public void appRunLP(Object content, String sender)
	{
		System.out.println(sender + " sent to IMLP: " + getAPPid());

		/**get message content and check the type of the message, type: <publish, subscribe, unsubscribe>*/
		AgentMsg agentMess = (AgentMsg)content;
		AgentMsg.MessageType type = agentMess.getType();
		MsgBall ball = agentMess.getMsgBall();

		switch (type)
		{
			case UPDATE:
				update(ball);
				break;
			case SUBSCRIBE:
				subscribe(ball);
				break;
			case UNSUBSCRIBE:
				unsubscribe(ball);
				break;
			case ENTER:
				enter(ball);
				break;
			case LEAVE:
				leave(ball);
				break;
			default:
				printRollbackSafe("Unknown message type in InterestManager");
				break;
		}
	}


	@Override
	public void appSetConfigData(Serializable cdata)
	{
		String[] s = ((String)cdata).split(" ");
		communicationType = s[1];
	}


	@Override
	public Serializable appSaveLPState()
	{
		InterestManager imlp = new InterestManager();
		imlp.subscriberList = new HashMap<String, MsgBall>(subscriberList);
		imlp.publisherList = new HashMap<String, MsgBall>(publisherList);
		return imlp;
	}


	@Override
	public void appSetLPState(Serializable lp)
	{
		InterestManager imlp = (InterestManager)lp;
		this.publisherList = imlp.publisherList;
		this.subscriberList = imlp.subscriberList;
	}


	private void printRollbackSafe(final String msg)
	{
//		submitIrreversibleAction(new Runnable()
//		{
//			public void run()
//			{
//				System.out.println(msg);
//			}
//		});
	}
}
