package im;

import im.AgentMsg.MessageType;
import sassy.Message;

import java.util.Arrays;
import java.util.Comparator;
import java.util.List;


public class TieBreakingComparator implements Comparator<Message>
{
	/**
	 * When an IMLP receives two messages from the same agent, they should be processed
	 * in this order.
	 */
	private static final MessageType[] priorityArrayToIMLP =
	{
		MessageType.LEAVE,
		MessageType.UNSUBSCRIBE,
		MessageType.SUBSCRIBE,
		MessageType.ENTER,
		MessageType.UPDATE,
	};

	/**
	 * When an agent A sends messages to an agent B, B should process
	 * A's messages in this order.
	 */
	private static final MessageType[] priorityArrayToAgent =
	{
		MessageType.REFLECTLEAVE,
		MessageType.REFLECTENTER,
		MessageType.COLLISION,
		MessageType.UPDATE, //for direct communication
		MessageType.REFLECTUPDATE,
	};

	private static final List<MessageType> prioritiesToIMLP = Arrays.asList(priorityArrayToIMLP);

	private static final List<MessageType> prioritiesAgentToAgent = Arrays.asList(priorityArrayToAgent);


	/**
	 * Break simultaneous events in bouncing ball agent based simulator.
	 *
	 * @param m1 The first message
	 * @param m2 The second message, with equal time stamp to the first
	 * @return -1 if m1 should go first, 1 if m2 should go first, 0 if can't decide
	 */
	public int compare(Message m1, Message m2)
	{
		boolean m1ReceiverIsAgent = m1.getReceiverAPPid().contains("ball");
		boolean m2receiverIsAgent = m2.getReceiverAPPid().contains("ball");

		if (m1ReceiverIsAgent && m2receiverIsAgent)
			return compareMessagesToAgent(m1, m2);
		else if (!m1ReceiverIsAgent && !m2receiverIsAgent)
			return compareMessagesToIMLP(m1, m2);
		
		//Now, we have 1 message to an agent and 1 message to an IMLP
		//For correctness, it does not matter at all which one is processed first
		//But it will probably reduce rollbacks if IMLP messages are processed first
		if (m1ReceiverIsAgent && !m2receiverIsAgent)
			return 1;
		else
			return -1;
	}


	/**
	 * Compare two agent middleware messages when they are both destined to an agent LP.
	 * Agents have a moving order. An agent processes messages in the order
	 * 1. Messages from agents of higher order
	 * 2. the agent's own SENSE message
	 * 3. Messages from agents of lower order
	 * 
	 * @param m1 first mesage sent
	 * @param m2 second message sent, with an equal timestamp
	 * @return less than 0 if m1 should be processed first, more than 0 if m2 goes first
	 */
	private static int compareMessagesToAgent(Message m1, Message m2)
	{
		//If the two messages are destined for different agents, first
		//process the message of the agent that goes first
		// *** This has no semantic significance since agents can process out of order.
		//     However, it is an optimization which reduces the number of rollbacks ***
		if (m1.getReceiverLP() != m2.getReceiverLP())
			return compareAgentPriorities(m1.getReceiverAPPid(), m2.getReceiverAPPid());

		//Now, we handle the case where the two message are destined for the same agent
		AgentMsg agentMsg1 = (AgentMsg)m1.getContent();
		AgentMsg agentMsg2 = (AgentMsg)m2.getContent();

		String senderMsg1 = agentMsg1.getMsgBall().getAPPid();
		String senderMsg2 = agentMsg2.getMsgBall().getAPPid();

		//If an agent receives messages from other agents, it should process first the ones
		//by agents which move first.
		//Note that this takes care of processing the agent's own SENSE event in the correct order
		if (!senderMsg1.equals(senderMsg2))
			return compareAgentPriorities(senderMsg1, senderMsg2);

		//Now we have two messages originating from the agent A, both destined for agent B.
		//They should be processed in the order { REFLECTLEAVE, REFLECTENTER, REFLECTUPDATE }
		if (m1.getReceiverAPPid().equals(senderMsg1))
		{
			//These are two SENSE events from an agent to itself. The order doesn't matter
			return 0;
		}

		MessageType typeMsg1 = agentMsg1.getType();
		MessageType typeMsg2 = agentMsg2.getType();

		int priorityMsg1 = prioritiesAgentToAgent.indexOf(typeMsg1);
		int priorityMsg2 = prioritiesAgentToAgent.indexOf(typeMsg2);
		assert priorityMsg1 != -1 && priorityMsg2 != -1;

		return priorityMsg1 - priorityMsg2;
	}


	/**
	 * Compares two messages both destined for IMLP's. An IMLP processes messages from
	 * agents based on the priority of the agents. Messages from the same agent are
	 * processed in the order { LEAVE, UNSUBSCRIBE, SUBSCRIBE, ENTER, UPDATE }
	 * 
	 * @param m1 message 1
	 * @param m2 message 2
	 * @return less than 0 if m1 should be processed first, more than 0 if m2 goes first
	 */
	private static int compareMessagesToIMLP(Message m1, Message m2)
	{
		//If the messages are destined for different IMLP's, we have no preference
		//over which one should be processed first. 
		if (m1.getReceiverLP() != m2.getReceiverLP())
			return 0;

		//Now, if the two messages are destined for the same IMLP, we have to check
		//the priority of the sender agents
		AgentMsg agentMsg1 = (AgentMsg)m1.getContent();
		AgentMsg agentMsg2 = (AgentMsg)m2.getContent();

		String senderMsg1 = agentMsg1.getMsgBall().getAPPid();
		String senderMsg2 = agentMsg2.getMsgBall().getAPPid();

		//First process messages from the agent with the higher priority
		if (!senderMsg1.equals(senderMsg2))
			return compareAgentPriorities(senderMsg1, senderMsg2);

		//If both messages are from the same agent, process them based on the message type
		MessageType typeMsg1 = agentMsg1.getType();
		MessageType typeMsg2 = agentMsg2.getType();

		int priorityMsg1 = prioritiesToIMLP.indexOf(typeMsg1);
		int priorityMsg2 = prioritiesToIMLP.indexOf(typeMsg2);
		assert priorityMsg1 != -1 && priorityMsg2 != -1;

		return priorityMsg1 - priorityMsg2;
	}


	/**
	 * Given the names of two agents, determines which one of them should move first.
	 * 
	 * @param agentName1 name of the first agent
	 * @param agentName2 name of the second agent
	 * @return less than 0 if agent 1 should move first, more than 0 if agent 2 should move first
	 */
	private static int compareAgentPriorities(String agentName1, String agentName2)
	{
		//For now, balli should execute before ballj if i < j
		agentName1 = agentName1.replaceAll("ball", "");
		agentName2 = agentName2.replaceAll("ball", "");

		int agent1Priority = Integer.parseInt(agentName1);
		int agent2Priority = Integer.parseInt(agentName2);

		return agent1Priority - agent2Priority;
	}
}

