package sassy;

import java.io.*;
import java.util.*;

/**
 * This class provides a default creator to the simulation. The format of the configuration file is:
 * appID appName className appConfigData
 * 
 * @author Yin Xiong
 * @author George Vulov
 */
public class DefaultCreator implements Creator
{
	/** Class name for breaking simultaneous events tie*/
	private String simulEventComparator = null;

	/**Distributes the LP among the available PEs
	 *@param numPEs int the number of PEs in this simulation
	 *@param numLPs int the number of LPs in this simulation
	 */
	@Override
	public List<PEInfo> appSetPEtable(int numPEs, List<LPConfigInfo> lpConfigInfoVector)
	{
		int numLPs = lpConfigInfoVector.size();
		List<PEInfo> peTable = new ArrayList<PEInfo>();
		int rmder = numLPs % numPEs;
		int basicQuota = numLPs / numPEs;
		int begin = 0;
		int end = 0;
		int lpForEach = 0;
		for (int i = 0; i < numPEs; i++)
		{
			lpForEach = basicQuota;
			if (i < rmder)
				lpForEach++;

			//Calculate the range of LP's which will be assigned to this PE
			begin = i * lpForEach;
			end = begin + lpForEach - 1;
			System.out.println("PE " + i + " LPs from " + begin + " to " + end);

			PEInfo peinfo = new PEInfo();

			for (int j = begin; j <= end; j++)
			{
				LPConfigInfo lpc = lpConfigInfoVector.get(j);
				System.out.println("PE " + i + " take LP " + lpc.LPid);
				peinfo.lpcVector.add(lpc);
			} //end for j

			peTable.add(peinfo);
		} //end for

		return peTable;
	} //end appAllocateLPs


	/**Processes LP configuration information.
	 *@param fileName String the name of the configuration file
	 *@return a vector of LPConfigInfo object, each being the configuration information for an LP
	 */
	@Override
	public List<LPConfigInfo> appCreateLPs(String fileName)
	{
		List<LPConfigInfo> lpConfigInfoVector = new ArrayList<LPConfigInfo>();
		
		//Read configuration file to decide how many instance for each class need to be created
		try
		{
			String newLine = null;
			BufferedReader bf = new BufferedReader(new FileReader(fileName));

			int numLPs = 0;
			//Each line is of the format: "<line type>:type-specific info"
			//LP's are specified as "LP:<id> <name> <class name> <application data>"
			while ((newLine = bf.readLine()) != null)
			{
				//If the first character is a '#', this is a comment. Also, ignore blank lines
				if (newLine.trim().length() == 0 || newLine.charAt(0) == '#')
					continue;

				//Extract the type of line
				int pos = newLine.indexOf(':');
				if (pos == -1)
				{
					bf.close();
					System.err.println("Error1: Malformated configration file.");
					System.exit(1);
				}
				String lineType = newLine.substring(0, pos).trim();
				newLine = newLine.substring(pos + 1).trim();

				//Now, process the line depending on what type it is
				if (lineType.equalsIgnoreCase("LP"))
				{
					//Only split on the first 3 fields because the application data can have
					String[] tokens = newLine.split("\\s+", 4);

					if (tokens.length < 3)
					{
						bf.close();
						System.err.println("Error: Malformated configration file.");
						System.exit(1);
					}

					String id = tokens[0].trim();
					String name = tokens[1].trim();
					String className = tokens[2].trim();
					String dat = tokens.length > 3 ? tokens[3] : "";
					LPConfigInfo lpcInfo = new LPConfigInfo(numLPs, id, name, className, dat);
					lpConfigInfoVector.add(lpcInfo);
					numLPs++;
				}
				else if (lineType.equalsIgnoreCase("SimultenousMessageComparator"))
				{
					simulEventComparator = newLine;
				}
			} //end while

			bf.close();
		}
		catch (IOException e)
		{
			System.err.println("Error: Could not read configuration file:\n\t" + e.getMessage());
		}

		return lpConfigInfoVector;
	} 


	@Override
	public String getTieBreakerClass()
	{
		return simulEventComparator;
	}
}

