package sassy;

import sassy.monitoring_steering.SimSocket;
import sassy.monitoring_steering.ReportNode;
import java.rmi.*;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.*;
import java.io.*;
import java.util.*;

/**
 * This class represents the Global Name Server that create the simulation world
 * and maintains the routing information for all the PEs.
 *
 * @author Yin Xiong
 * @author George Vulov <georgevulov@hotmail.com>
 */
public class GNS implements GNSInterface
{
	/** The flag for displaying debugging MS messages.*/
	private static final boolean MTRACE = true;

	/** The flag for displaying debugging statistics messages.*/
	private static final boolean STRACE = true;

	/** The flag for collecting statistics: 1=LP; 2=PE; 3=both */
	private int collectStat = 0;

	/** The queue that store simulation statistics*/
	private PriorityQueue<LPStat> statisticsQ;

	/** The RMI registry which is currently running. */
	private Registry rmiRegistry;

	/** The time duration of the simulation*/
	private long simTime = 0;

	/** The beginning time of the simulation*/
	private long beginTime;

	/** The ending time of the simulation*/
	private long endTime;

	/**
	 * The socket server that communicate between this GNS and the monitoring and
	 * steering clients.
	 */
	private SimSocket simSocket;

	/** True if the simulation has been terminated. */
	private volatile boolean simulationEnded;

	//Lock to prevent re-entrancy to endSimulation method
	private final Object endSimLock = new Object();

	/** Current number of PEs registered with the GNS. */
	private int registeredPEcount;

	/** Table that contains the information about the PEs. */
	private List<PEInfo> PEtable;

	/** Routing table. */
	private final List<RoutingInfo> routingTable;

	/** Total number of PEs for this simulation. */
	private int numPEs;

	/** Total number of LPs for this simulation. */
	private int numLPs;

	/** Vector that contains the LP configuration information obtained from a file*/
	private List<LPConfigInfo> lpConfigInfoVector;

	/** Full class name of the Comparator used to break ties with simultaneous events.*/
	private String simulEventComparator;

	/** Is the simulation currently calculating GVT? */
	private boolean isCalculatingGVT;

	//Lock to prevent re-entrancy to GVT function
	private final Object gvtLock = new Object();


	/**
	 * Constructor.
	 *
	 * @param stats true if stat collection should be on, false otherwise
	 */
	public GNS(int stats)
	{
		collectStat = stats;

		statisticsQ = new PriorityQueue<LPStat>();
		simulationEnded = false;
		isCalculatingGVT = false;
		registeredPEcount = 0;
		PEtable = new ArrayList<PEInfo>();
		lpConfigInfoVector = new ArrayList<LPConfigInfo>();
		routingTable = new ArrayList<RoutingInfo>();
	}


	/**main function*/
	public static void main(String[] args)
	{
		int collectStat = 0;
		String fileName = null;
		String appCreator = null;
		int numPEs = -1;
		long simTime = 0;

		if (args.length == 0)
		{
			System.out.println("\nExpecting more args...");
			System.out.println("\nUsage: java sassy.GNS file= numPEs= stat= creator=");
			System.exit(1);
		}

		for (int i = 0; i < args.length; i++)
		{
			String s[] = args[i].split("=");
			if (s[0].equalsIgnoreCase("file"))
			{
				fileName = s[1];
				System.out.println("\nConfig file: " + fileName);
			}
			else if (s[0].equalsIgnoreCase("numPEs"))
			{
				numPEs = Integer.parseInt(s[1]);
				System.out.println("\nNumber of PEs: " + numPEs);
			}
			else if (s[0].equalsIgnoreCase("stat"))
			{
				collectStat = Integer.parseInt(s[1]);
				System.out.println("\ncollect stat: 1=LP, 2=PE, 3=both :" + collectStat);
			}
			else if (s[0].equalsIgnoreCase("creator"))
			{
				appCreator = s[1];
				System.out.println("\nCreator: " + appCreator);
			}
			else if (s[0].equalsIgnoreCase("time"))
			{
				simTime = Long.parseLong(s[1]);
				System.out.println("\nSim time: " + simTime);
			}
		}
		
		if (numPEs <= 0)
		{
			System.err.println("Invalid number of PE's. There must be at least one processing executive.");
			System.exit(1);
		}

		//Start the GNS
		final GNS gns = new GNS(collectStat);

		gns.numPEs = numPEs;
		gns.simTime = simTime;
		gns.procConfigFile(fileName, numPEs, appCreator);
		gns.startGNS();

		//wait until all PEs get registered, then send begin-to-run command
		while (gns.registeredPEcount < gns.numPEs)
		{
			try
			{
				Thread.sleep(100);
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
		}

		gns.beginSim();
		if (simTime > 0.0)
		{
			Timer simTimer = new Timer();
			simTimer.schedule(new TimerTask()
			{
				@Override
				public void run()
				{
					System.out.println("\nTime is up. Ending simulation...");
					try
					{
						gns.endSimulation();
					}
					catch (Exception e)
					{
						System.out.println("\nError ending sim...");
					}
				}
			}, simTime);
		}
		System.out.println("\nSimulation started successfuly!");
	}


	/**Gets statistics information from PEs. */
	void getStatistics()
	{
		for (int i = 0; i < PEtable.size(); i++)
		{
			PEInfo pinfo = PEtable.get(i);
			Peer peerStub = pinfo.getRmiPeerStub();
			for (int j = 0; j < pinfo.lpcVector.size(); j++)
			{
				try
				{
					LPConfigInfo lpc = pinfo.lpcVector.get(j);
					LPStat lpstat = peerStub.getLPStat(lpc.LPid);
					statisticsQ.add(lpstat);
				}
				catch (Exception e)
				{
					e.printStackTrace();
				}
			}
		}
	}


	/**Displays statistics*/
	private void displayStats()
	{
		int evProcessed, evCanceled, evRollbacked, evCommitted, evTotal, evNet, numRollbacks, numRollbacksIM, numRollbacksAgent;
		evProcessed = evCanceled = evRollbacked = evCommitted = evTotal = evNet = numRollbacks = numRollbacksIM = numRollbacksAgent = 0;
		int antiMsgsPrevented = 0;

		int msgAntiIn, msgAntiOut, msgMonitoring, msgSteering;
		msgAntiIn = msgAntiOut = msgMonitoring = msgSteering = 0;

		double avgRBLength, eventRate;
		avgRBLength = eventRate = 0.0;

		int fossilCollect, longestRollback;
		fossilCollect = longestRollback = 0;

		long totalLPTime = 0;
		numLPs = statisticsQ.size();

		System.out.println("statistics queue size: " + statisticsQ.size());
		if (statisticsQ.size() == 0)
			return;

		Iterator<LPStat> iterator = statisticsQ.iterator();

		System.out.println("\n\tStatistics by LP\t\t\n\n");

		//legend
		if (collectStat == 1 || collectStat == 3)
		{
			System.out.println("ID: PE id, LP id");
			System.out.println("Time: beginning time, endding time, time duration");
			System.out.println("Events: processed, rollback, committed, aborted, canceled");
			System.out.println("Messages: Anti-in, Anti-out, monitoring, steering");
			System.out.println("Others: fossil collected, avg rollback length, longest rollback");
			System.out.println("Caching:  cache hits, cache misses, hit rate");
		}

		while (iterator.hasNext())
		{
			LPStat lpstat = iterator.next();

			evProcessed += lpstat.numProcEvents;
			evRollbacked += lpstat.numEventsRolledBack;
			numRollbacks += lpstat.numRollbacks;

			//tianhao added: count number of rollback for IM and agents
			if (lpstat.getAPPid() != null && lpstat.getAPPid().contains("ball"))
			{
				numRollbacksAgent += lpstat.numRollbacks;
			}
			else
			{
				numRollbacksIM += lpstat.numRollbacks;
			}//tianhao

			evCommitted += lpstat.numCommEvents;
			evCanceled += lpstat.numCanceledEvents;
			antiMsgsPrevented += lpstat.numAntiMessagesLazilyCancelled;

			msgAntiIn += lpstat.numAntiMsgsIn;
			msgAntiOut += lpstat.numAntiMsgsOut;
			msgMonitoring += lpstat.numMonitoringMsgs;
			msgSteering += lpstat.numSteeringMsgs;

			fossilCollect += lpstat.numFossils;
			avgRBLength += lpstat.avgRBLength;
			totalLPTime += lpstat.simTime;
			if (longestRollback < lpstat.longestRollback)
				longestRollback = lpstat.longestRollback;


		} //end while

		if (collectStat == 2 || collectStat == 3)
		{
			System.out.println("\n\t\tSummary Statistics\t\t\n\n");

			System.out.println("Execusion time:		" + (endTime - beginTime) + "(millisedonds)");
			System.out.println("Rollbacks:		" + numRollbacks);
			System.out.println("Agent Rollbacks: 	" + numRollbacksAgent);
			System.out.println("IM Rollbacks:		" + numRollbacksIM);
			System.out.println("Rolled-back events:	" + evRollbacked);
			System.out.println("");
			System.out.println("Number of PEs: " + numPEs);
			System.out.println("Number of LPs: " + numLPs);
			evTotal = evProcessed + evRollbacked + evCanceled;
			evNet = evProcessed - evRollbacked;

			System.out.println("Total events: " + evTotal);
			System.out.println("Net events: " + evNet);
			System.out.println("Processed events: " + evProcessed);

			System.out.println("*Anti-messages not sent out due to lazy cancellation: " + antiMsgsPrevented);
			System.out.println("Canceled events: " + evCanceled);
			System.out.println("Committed events: (to be implemented)");

			System.out.println("Incoming anti-messages: " + msgAntiIn);
			System.out.println("Outgoing anti-messages: " + msgAntiOut);
			System.out.println("Monitoring messages: " + msgMonitoring);
			System.out.println("Steering messages: " + msgSteering);

			eventRate = (double) evNet / ((double)simTime / 1000.0);

			System.out.println("Fossil collections: " + fossilCollect);
			System.out.println("Average length of rollback: " + (avgRBLength / numLPs));
			System.out.println("Event rate (event/second): " + eventRate);
			System.out.println("longegst rollback: " + longestRollback);
		}
	} 


	/** Remote method for PEs to get the flag for collection statistics.
	 * @return true if collectStat is greater than 0
	 */
	@Override
	public boolean isCollectingStats() throws RemoteException
	{
		if (collectStat > 0)
			return true;
		else
			return false;
	}


	/**
	 * Receives a Monitoring/Steering/Statistics message from a PE. 
	 * @param msg a Message instance to be received.
	 */
	@Override
	public void receive_MS_Msg(Message m) throws RemoteException
	{
		System.out.println("gns received stat msg: " + m.getContent());
		if (m.getMsgType() == Message.MsgType.SYSTEM_STAT)
		{
			LPStat lpstat = (LPStat) m.getContent();
			statisticsQ.offer(lpstat);
			if (STRACE)
			{
				String source = lpstat.PEid + "-" + lpstat.LPid;
				System.out.println("gns received stat info from " + source);
			//lpstat.display();
			} //end if
		} //end if it's a statistics msg
		else if (m.getMsgType() == Message.MsgType.SYSTEM_REPORT)
		{

			System.out.println("gns received MS meg...");

			ReportNode rnode = (ReportNode) m.getContent();
			if (MTRACE)
				System.out.println("gns received ms report: " + rnode.stringValue());
			simSocket.sendReport(rnode);
		} //end if it's a MS message
	}


	/**
	 * Sends the begin-to-run  message
	 */
	public void beginSim()
	{
		beginTime = System.currentTimeMillis();
		for (int i = 0; i < PEtable.size(); i++)
		{
			PEInfo pinfo = PEtable.get(i);
			Peer peerStub = pinfo.getRmiPeerStub();
			try
			{
				peerStub.beginSim();
				System.out.println("\nGNS sets PE " + pinfo.getPEid() + " to run");
			}
			catch (Exception e)
			{
				e.printStackTrace();
				System.exit(1);
			}
		}
	} 


	/**
	 * Sends monitoring and steering message
	 * @param msg Message object - the message to be sent
	 */
	public int sendMSmessage(Message msg)
	{

		int pid = msg.getReceiverPE();
		int success = 0;
		System.out.println("GNS get MS msg pid: " + pid);

		for (int i = 0; i < PEtable.size(); i++)
		{
			PEInfo pinfo = PEtable.get(i);
			System.out.println("in-pid, pid: " + pid + " " + pinfo.getPEid());
			if (pinfo.getPEid() == pid)
			{
				Peer peerStub = pinfo.getRmiPeerStub();
				try
				{
					peerStub.receiveMsg(msg);
					System.out.println("GNS send MS msg to " + pid);
				}
				catch (Exception e)
				{
					success = -1;
				}

			} //end if
		} //end for

		return success;
	} //end sendMSmessge


	/**
	 * Returns wether the simulation is still active.
	 *
	 * @return false if the simulation is being shut down
	 */
	public boolean isSimulationRunning() throws RemoteException
	{
		return !simulationEnded;
	} //end getEOFflag


	/**
	 * Sets the end of simulation flag, sends the flag to each PE and shuts down
	 * the RMI server for each PE.
	 */
	public synchronized void endSimulation()
			throws RemoteException
	{
		synchronized (endSimLock)
		{
			if (simulationEnded)
				return;
			simulationEnded = true;
		}

		//Stop the monitoring/steering socket server
		simSocket.stopMS();

		//This is an RMI method so we must return from it before
		//shutting down RMI. Thus, we'll set up a thread to do the work and
		//then return
		new Thread("GNS - Shutdown")
		{
            @Override
			public void run()
			{
				//Give time for the RMI method to return
				try
				{
					Thread.sleep(2000);
				}
				catch (InterruptedException ex)
				{
				}

				//Shut down the RMI registry
				try
				{
					UnicastRemoteObject.unexportObject(rmiRegistry, true);
				}
				catch (NoSuchObjectException ex)
				{
					ex.printStackTrace();
				}


				System.out.println("\nGNS is collecting statistics...");
				getStatistics();

				System.out.println("\nGNS is terminating the simulation...");

				//Tell peer servers to exit
				try
				{
					for (int i = 0; i < PEtable.size(); i++)
					{
						PEInfo pinfo = PEtable.get(i);
						System.out.println("Sending STOP to " + pinfo.getMachineName());
						pinfo.getRmiPeerStub().shutDown();
					} //end for

					endTime = System.currentTimeMillis();
				}
				catch (Exception e)
				{
					e.printStackTrace();
				}

				//if asked to collect stat, then display stats
				if (collectStat > 0)
					displayStats();

				//Unexport the GNS object so this process can shut down
				try
				{
					UnicastRemoteObject.unexportObject(GNS.this, true);
				}
				catch (NoSuchObjectException ex)
				{
					ex.printStackTrace();
				}
			}
		}.start();
	} //endSimulation()


	/** Starts the GNS as a RMI server offering remote methods for the PEs. */
	public void startGNS()
	{
		try
		{
			//Start up an RMI registry
			rmiRegistry = LocateRegistry.createRegistry(1099);

			//Bind to the registry
			String name = "GNS";
			GNSInterface stub = (GNSInterface) UnicastRemoteObject.exportObject(this, 0);
			Registry registry = LocateRegistry.getRegistry();
			registry.rebind(name, stub);
			System.out.println("\nGNS started successfully. Waiting for PEs...");
		}
		catch (RemoteException e)
		{
			System.err.println("\nError binding GNS server: \n" + e);
			System.exit(1);
		}

		//Turn off monitoring/steering code
		if (false)
		{
			simSocket = new SimSocket(this);
			simSocket.start();
		}
	}


	/**
	 * RMI method which registers a new PE and returns the LP configuration
	 * information to the new PE.
	 *
	 * @param ip the ip of the PE
	 * @param machineName the machine name of the PE
	 * @param peerStub the RMI remote reference for the PE
	 * @return the LP's that are assigned to the PE
	 */
	@Override
	public PEConfig registerPE(String ip, String machineName, Peer peerStub) throws RemoteException
	{
		boolean isRegistered = false;
		PEInfo peInfo = null;

		//Check if the PE is already registered
		for (int i = 0; i < PEtable.size(); i++)
		{
			peInfo = PEtable.get(i);
			if (peInfo.getRmiPeerStub() == peerStub)
			{
				assert false; //The same PE shouldn't register twice
				isRegistered = true;
				break;
			}
		} //end for

		if (!isRegistered)
		{
			assert registeredPEcount < PEtable.size();
			peInfo = PEtable.get(registeredPEcount);
			peInfo.setIP(ip);
			peInfo.setMachineName(machineName);
			peInfo.setPeerStub(peerStub);
			peInfo.setPEid(registeredPEcount);
			PEtable.set(registeredPEcount, peInfo);

			//Print each LP this PE is getting
			for (LPConfigInfo lpinfo : peInfo.lpcVector)
			{
				lpinfo.PEid = peInfo.getPEid();
				System.out.println(machineName + " get LP " + lpinfo.PEid + " " + lpinfo.LPid);
				RoutingInfo rinfo = new RoutingInfo(peInfo, lpinfo.LPid, lpinfo.getAPPid());
				routingTable.add(rinfo);
			}
			
			//Notify anyone waiting that a new PE has registered
			synchronized (routingTable)
			{
				routingTable.notifyAll();
			}

			System.out.println("New Client accepted: " + registeredPEcount + " " + machineName + " " + ip);
			registeredPEcount++;
		}
		
		return new PEConfig(peInfo.lpcVector, simulEventComparator);
	}


	/**
	 * Processes the configuration file. 
	 *
	 * @param configFile String indicating the name of the configuration file.
	 * @param size int the number of PEs.
	 * @param appCreator String the class name of the application-provided creator.
	 */
	public void procConfigFile(String configFile, int size, String appCreator)
	{
		try
		{
			String creatorName = "sassy.DefaultCreator";
			if (appCreator != null)
				creatorName = appCreator;

			Creator creator = (Creator) (Class.forName(creatorName).newInstance());

			//creator returns a vector of LPConfigInfo objects
			lpConfigInfoVector = creator.appCreateLPs(configFile);

			simulEventComparator = creator.getTieBreakerClass();

			//creator returns a vector of PEInfo object
			PEtable = creator.appSetPEtable(size, Collections.unmodifiableList(lpConfigInfoVector));

		}
		catch (Exception e)
		{
			e.printStackTrace();
			System.exit(1);
		}
	}


	/**
	 * Looks up the routing info by an LP's application id.
	 * @param appid String representing the application id of the LP.
	 */
	@Override
	public RoutingInfo getRoutingInfoByAid(String appid) throws RemoteException
	{
		//Don't return any routing information until all the PE's are available
		synchronized (routingTable)
		{
			while (routingTable.size() < lpConfigInfoVector.size())
			{
				try
				{
					routingTable.wait();
				}
				catch (InterruptedException ex)
				{
				}
			}
		}

		for (RoutingInfo rinfo : routingTable)
		{
			if (appid.equalsIgnoreCase(rinfo.getAPPid()))
				return rinfo;
		}
		return null;
	}


	/**
	 * Looks up the routing info by an LP id.
	 * @param lid int representing the LP id.
	 */
	@Override
	public RoutingInfo getRoutingInfoByLid(int lid) throws RemoteException
	{
		//Don't return any routing information until all the PE's are available
		synchronized (routingTable)
		{
			while (routingTable.size() < lpConfigInfoVector.size())
			{
				try
				{
					routingTable.wait();
				}
				catch (InterruptedException ex)
				{
				}
			}
		}

		for (RoutingInfo rinfo : routingTable)
		{
			if (lid == rinfo.getLPid())
				return rinfo;
		}
		return null;
	}


	/**
	 * PE Requests that the GNS initiate a calculation of Global Virtual Time (GVT).
	 * GVT guarantees that no messages with timestamp < GVT will occur, thus
	 * allowing logical processes to free rollback history for such messages.
	 */
	@Override
	public void requestGVTCalculation() throws RemoteException
	{
		synchronized (gvtLock)
		{
			if (isCalculatingGVT || simulationEnded)
				return;

			//Make sure no GVT calculation is initiated until every LP has been assigned to
			//			a PE! Such calculation would most likely be incorrect
			if (registeredPEcount < numPEs)
				return;

			isCalculatingGVT = true;
		}

		new Thread("GVT Calculation")
		{
			@Override
			public void run()
			{
				//Note: We implement GVT calculation using Samadi's algorithm
				int GVT = Integer.MAX_VALUE;

				try
				{
					//Go around and ask every Process Executive to calculate its minimum local timestamp
					for (PEInfo peRecord : PEtable)
					{
						int localMin = peRecord.getRmiPeerStub().GVTCalculationStarted();

						GVT = Math.min(GVT, localMin);
					}

					//Now, announce the new GVT value to all the PE's
					for (PEInfo peRecord : PEtable)
						peRecord.getRmiPeerStub().GVTCalculationEnded(GVT);
				}
				catch (RemoteException ex)
				{
					ex.printStackTrace(System.err);
				}

				isCalculatingGVT = false;
			}
		}.start();
	}

	/**
	 * Helper class which contains all parameters for running a PE,
	 * such as the LP's which will run on it, the tie-breaking comparator,
	 * etc.
	 */
	public static class PEConfig implements Serializable
	{
		/**
		 * Vector of the configuration data for all the LP's which will run
		 * on the specified PE.
		 */
		public List<LPConfigInfo> LPConfigs;

		/**
		 * Fully-qualified class name of the Comparator which will be used
		 * to break ties between events with equal time stamps.
		 */
		public String simultEventComparator;


		/** Initialize a PEConfig instance. */
		PEConfig(List<LPConfigInfo> LPConfigs, String simultEventComparator)
		{
			this.LPConfigs = LPConfigs;
			this.simultEventComparator = simultEventComparator;
		}
	}
} //end class

