package sassy;

import sassy.monitoring_steering.NameValuePair;
import java.io.*;
import java.util.*;

/**
 * This class represents a logical process that communicates with other logical
 * processes  by exchanging time-stamped messages. LPs cannot "spontaneously" 
 * begin new computations without first receiving a message. 
 *
 * <p>If an LP receives a "straggler" message with timestamps smaller than that 
 * its LVT(local virtual time), it rolls back to its LVT and reprocesses (after the straggler) 
 * in timestamp order. Each message sent by a rolled back computation is cancelled by 
 * sending an anti-message that annihilates the original message. A received anti-message 
 * will also cause a rollback if the cancelled message has already been processed.</p>
 *
 * <p>This is an abstract class. The simulation application programmer needs 
 * to implement the abstract methods specified in this class. All such methods
 * are prefixed "app" for ease of recognition.</p>
 *
 * @author George Vulov <georgevulov@hotmail.com>
 * @author Yin Xiong
 * @author Glenn Matthews
 */
public abstract class LogicalProcess
{
	/** If true, display debugging output. */
	private static final boolean TRACE = false;

	/** The flag for displaying debugging msgs for collecting statistics*/
	private static final boolean STRACE = false;

	/** The flag indicating whether simulation statistics should be collected*/
	private boolean collectStat = false;

	/** An LPStat object that stores simulation statistics for this LP. */
	transient LPStat lpstat = new LPStat();

	/** Unique numeric identifier for the PE where this LP resides; for debugging purposes. */
	private int PEid;

	/** Unique numeric identifier assigned to this LP by the system. */
	private int LPid;

	/** Unique application-assigned identifier of this LP. */
	private String APPid;

	/** Unique application-assigned full name of this LP. */
	private String myname;

	/** Local virtual time(LVT) for this LP. */
	private transient int currentTime;

	/** Value of {@link #currentTime} the previous time {@link #runLP} was called.  */
	private transient int lastTime;

	/**
	 * A counter which increments by 1 each time this LP sends a message.
	 * It is used in breaking ties of simultaneous events.
	 * NOTE: This counter is not saved and restored during rollback; it always increases.
	 */
	private transient long messageCounter;

	/** SchedulingLoop that this LP can access. */
	private transient SchedulingLoop scheduler;

	/**
	 * The list that stores the saved LP states and processed messages. Note that this list 
	 * will always be sorted in order of increasing timestamp, since this is how events are processed,
	 * and state saving occurs right before processing an event.
	 */
	private transient final LinkedList<ProcessingRecord> processedMessages = new LinkedList<ProcessingRecord>();

	/**
	 * This list stores all the anti messages for which the positive messages have not
	 * come in yet. When the positive message comes in & matches an anti-message,
	 * both are deleted.
	 */
	private transient final List<Message> unmatchedAntiMessages = new LinkedList<Message>();

	/** True if the LP should not immediately send-out anti-messages after a rollback. */
	private final boolean useLazyCancellation;

	/**
	 * For each rolled back message, we keep a list of the anti-messages which had to
	 * be sent out to cancel it.
	 */
	private transient final Map<Message, List<Message>> lazilyCancelledMessages;


	/** 
	 * Default constructor, which creates an LP which does not
	 * use lazy cancellation. 
	 */
	public LogicalProcess()
	{
		this(false);
	}


	/**
	 * Create a logical process, specifying whether it uses lazy cancellation or not.
	 * 
	 * @param useLazyCancellation true if anti-messages should not be sent immediately after a rollback
	 */
	public LogicalProcess(boolean useLazyCancellation)
	{
		this.useLazyCancellation = useLazyCancellation;
		if (useLazyCancellation)
			lazilyCancelledMessages = new HashMap<Message, List<Message>>();
		else
			lazilyCancelledMessages = null;
	}


	/**
	 * Invoked by the kernel at the beginning of the simulation when the LP
	 * should initialize itself and send any initial messages.
	 */
	public abstract void appInitializeLP();


	/**
	 * Invoked when a new message destined for this logical process
	 * is received.
	 * @param content The object associated with the new event.
	 * @param sender The application id of the LP the event came from.
	 */
	public abstract void appRunLP(Object content, String sender);


	/**
	 * Invoked by the kernel when the LP is being initialized.
	 * The application-specific text from the configuration file is passed in.
	 * The LP should parse it and set up its internal state accordingly.
	 *
	 * @param cdata  Configuration data. This object is a string when the default creator is used.
	 */
	public abstract void appSetConfigData(Serializable cdata);


	/**
	 * Invoked by the kernel when the LP should save its current state. The application should
	 * allocate a <b>new</b> instance of the proper subclass of <code>LogicalProcess</code>, then
	 * copy all state data to that object and return it. 
	 * 
	 * @return a <code>LogicalProcess</code> object containing the current state
	 */
	public abstract Serializable appSaveLPState();


	/**
	 * Invoked by the kernel when the LP's state is being reverted back to a previous state.
	 * The application should cast the parameter to the proper subclass of <code>LogicalProcess</code>.
	 * It should transfer all the appropriate application-specific data from the passed LP object
	 * to the current one.
	 * 
	 * @param lp LogicalProcess containing the state information; the application
	 *         program decides what state information need to be restored
	 */
	public abstract void appSetLPState(Serializable lp);


	/**
	 * Invoked by the kernel at the end of the simulation. 
	 */
	public abstract void appFinalizeLP();


	/**
	 * Returns a Vector of NameValuePair describing all application-specific monitorable
	 * parameters of this LP, each in the form "paramName, paramType". Overriding this
	 * is optional.
	 */
	public Vector<NameValuePair> appGetMonitorable()
	{
		return null;
	}


	/**
	 * Returns a Vector of NameValuePair describing all application-specific steerable
	 * parameters of this LP, each in the form "paramName, paramType". Overriding this is
	 * optional.
	 */
	public Vector<NameValuePair> appGetSteerable()
	{
		return null;
	}


	/**
	 * Called by the scheduling loop when a simulation message destined for this
	 * LP should be processed. Note that this method is package-private since
	 * it should only be called from within the SASSY kernel.
	 *
	 * @param msg the new message which should be processed
	 */
	synchronized void processPositiveMessage(Message msg)
	{
		//Check if this message has a corresponding anti-message
		ListIterator<Message> antiMsgIter = unmatchedAntiMessages.listIterator();
		while (antiMsgIter.hasNext())
		{
			Message antiMsg = antiMsgIter.next();
			if (msg.compareTo(antiMsg) == 0)
			{
				//The message & the anti-message match, so annihilate them
				antiMsgIter.remove();

				//If this message had any unsent anti-messages assciated with it, send them
				//since there is no way those same messages will be re-generated
				if (useLazyCancellation && lazilyCancelledMessages.containsKey(msg))
				{
					List<Message> unsentAntiMsgs = lazilyCancelledMessages.remove(msg);
					for (Message unsentAntiMsg : unsentAntiMsgs)
					{
						scheduler.sendMessage(unsentAntiMsg);
						if (collectStat)
							lpstat.numAntiMsgsOut++;
					}
				}
				return;
			}
		}

		//First, check if a rollback is necessary & perform one if so
		if (!processedMessages.isEmpty())
		{
			Message lastMsg = processedMessages.getLast().incomingMessage;
			if (lastMsg.compareTo(msg) >= 0)
				rollBack(msg);
		}

		if (TRACE)
			System.out.printf("(LPid=%d, APPid=%s) PROCESSING message from (LPid=%d, APPid=%s) with TS=%d\n",
					LPid, APPid, msg.getSenderLP(), msg.getSenderAPPid(), msg.getTimeStamp());

		//Save the LP state
		Serializable savedLP = appSaveLPState();

		//Create a new processing record and add it to the list
		ProcessingRecord newRecord = new ProcessingRecord(msg, savedLP, lastTime, currentTime);
		processedMessages.add(newRecord);

		//update statistics
		if (collectStat)
			lpstat.numProcEvents++;

		//Update the simulation time
		lastTime = currentTime;
		currentTime = msg.getTimeStamp();

		//Let the application handle the message
		appRunLP(msg.getContent(), msg.getSenderAPPid());

		//If we're using lazy cancellation and re-processing an event which was previously
		//rolled back, we must now send any anti-messages which we didn't send during
		//the original rollback and which did not get re-generated during the second execution of the event
		if (useLazyCancellation && lazilyCancelledMessages.containsKey(msg))
		{
			List<Message> antiMessages = lazilyCancelledMessages.remove(msg);
			for (Message antiMessage : antiMessages)
			{
				scheduler.sendMessage(antiMessage);
				if (collectStat)
					lpstat.numAntiMsgsOut++;
			}
		}
	}


	/**
	 * Called by the scheduling loop when an anti-message destined for this LP has been 
	 * received. Note that this class is package-private since it should only be called
	 * from within the SASSY kernel.
	 *
	 * @param antiMsg the anti-message destined for this LP
	 */
	synchronized void processAntiMessage(Message antiMsg)
	{
		if (TRACE)
			System.out.printf("(LPid=%d, APPid=%s) PROCESSING anti-message from (LPid=%d, APPid=%s) with TS=%d\n",
					LPid, APPid, antiMsg.getSenderLP(), antiMsg.getSenderAPPid(), antiMsg.getTimeStamp());

		//When an anti-message comes in, there are 3 cases
		//1. The positive message has not yet been processed
		//2. The positive message has been processed
		//3. The positive message has not yet arrived.
		//
		//In cases 1 and 3, we store the anti-message and when the postive message comes in, we
		//cancel both. In case 2, we roll back to before the message was processed, then we have case 1.

		//See if the message has been processed
		Message positiveMsg = null;
		for (ProcessingRecord record : processedMessages)
		{
			if (record.incomingMessage.compareTo(antiMsg) == 0)
			{
				positiveMsg = record.incomingMessage;
				break;
			}
		}

		//The message has already been processed, do a rollback (Case 2)
		//This will roll back the positive message and everything after it
		if (positiveMsg != null)
		{
			rollBack(antiMsg);
		}

		//Add the anti-msg to the unmatched list so when the positive message comes in it is annihilated
		unmatchedAntiMessages.add(antiMsg);

		//Update statistics
		if (collectStat)
			lpstat.numAntiMsgsIn++;
	}


	/**
	 * <p>Rolls back the state of this LP to a time before the given message.
	 * All messages which are rolled back are returned to the scheduling loop.</p>
	 *
	 * <p>IMPORTANT: Messages which compareTo 0 with the given message are also
	 * rolled back.</p>
	 *
	 * @param msg the message which initiates the rollback
	 * @return the total number of messages rolled back
	 */
	private int rollBack(Message msg)
	{
		int rollbackCount = 0;
		ProcessingRecord rollbackState = null;

		while (!processedMessages.isEmpty())
		{
			Message lastMsg = processedMessages.getLast().incomingMessage;
			if (lastMsg.compareTo(msg) >= 0)
			{
				//Get the state before the above message was processed
				rollbackState = processedMessages.getLast();

				//Return the message to the scheduling loop
				scheduler.returnMessage(lastMsg);

				//Send anti-messages out, or save them for later if we're lazily cancelling
				List<Message> antiMessages = processedMessages.getLast().getAntiMessages();
				if (useLazyCancellation)
				{
					lazilyCancelledMessages.put(lastMsg, antiMessages);
				}
				else
				{
					for (Message antiMsg : antiMessages)
					{
						scheduler.sendMessage(antiMsg);
						if (collectStat)
							lpstat.numAntiMsgsOut++;
					}
				}

				//Remove the processing record of processing the message
				processedMessages.removeLast();

				rollbackCount++;
			}
			//Rollback has reached a message which should not be rolled back
			else
			{
				//Ok, so we should restore the state of the LP to this point
				if (rollbackState != null)
				{
					lastTime = rollbackState.lastTime;
					currentTime = rollbackState.currentTime;
					appSetLPState(rollbackState.savedState);
				}

				//Exit the rollback loop
				break;
			}
		}

		if (TRACE)
		{
			String msgType = (msg.getMsgType() == Message.MsgType.SIM_REGULAR) ? "straggler " : "anti-msg";
			System.out.printf("(LPid=%d, APPid=%s) ROLLED BACK %d messages.\n", LPid, APPid, rollbackCount);
			System.out.printf("\t\tRollback was triggered by %s from (LPid=%d, APPid=%s) with TS=%d\n",
					msgType, msg.getSenderLP(), msg.getSenderAPPid(), msg.getTimeStamp());
		}

		//Update statistics
		if (collectStat && rollbackCount > 0)
		{
			lpstat.numEventsRolledBack += rollbackCount;
			lpstat.numRollbacks++;
		}

		return rollbackCount;
	}


	/**
	 * Remove from memory any processed messages whose time stamp is less than the
	 * global virtual time. Such messages are guaranteed to never be rolled back.
	 * Also, perform non-reversible actions whose time stamp is less than the GVT.
	 */
	synchronized void collectFossils()
	{
		int GVT = scheduler.getGVT();

		ListIterator<ProcessingRecord> iter = processedMessages.listIterator();

		int count = 0;
		while (iter.hasNext())
		{
			ProcessingRecord record = iter.next();

			//Processing records are always in time stamp order. So, TS >= GVT for all subsequent record
			if (record.incomingMessage.getTimeStamp() >= GVT)
				break;

			//Commit any irreversible changes associated with this event, since it won't be rolled back
			record.commitIrreversibleActions();

			//Delete the record - it is no longer necessary
			iter.remove();
			++count;
		}

		if (TRACE && count > 0)
			System.out.printf("(LPid=%d, APPid=%s) fossil collected %d records.\n", LPid, APPid, count);
	}


	/**
	 * Sends a new simulation message to a logical process, possibly the current one.
	 *
	 * @param content The contents of the simulation message.
	 * @param timeInc the time increment assigned by the sender. The send time of the
	 *					message is LVT + time increment
	 * @param targetAPPid the {@link #APPid} of the LP the message should go to.
	 */
	public final void sendSimMessage(Serializable content, int timeInc, String targetAPPid)
	{
		doSendSimMessages(content, timeInc, -1, targetAPPid);
	}


	/**
	 * Sends a new simulation message to a logical process, possibly the current one.
	 * 
	 * @param content The contents of the simulation message.
	 * @param timeInc the time increment assigned by the sender. The send time of the
	 *					message is LVT + time increment
	 * @param targetLPid int the LPid of the receiver
	 */
	public final void sendSimMessage(Serializable content, int timeInc, int targetLPid)
	{
		doSendSimMessages(content, timeInc, targetLPid, null);
	}


	/**
	 * Actually hand off a simulation message to the scheduler for transport.
	 * 
	 * @param content message content
	 * @param timeInc time increment of the message, in simulation time
	 * @param targetLPid ID of the destination LP, can be -1
	 * @param targetAPPid Application ID of the destination LP, can be null
	 */
	private void doSendSimMessages(Serializable content, int timeInc, int targetLPid, String targetAPPid)
	{
		if (timeInc < 0)
		{
			String error = "LP attempted to send message in the past (time increment < 0).";
			throw new IllegalArgumentException(error);
		}

		//Timestamp of the outgoing message
		int sendtime = currentTime + timeInc;

		//Get the message whose processing generated this message
		Message ancestor = null;
		if (!processedMessages.isEmpty())
			ancestor = processedMessages.getLast().incomingMessage;

		Message m = new Message(sendtime, Message.MsgType.SIM_REGULAR, content, APPid, targetAPPid, ancestor, messageCounter++);
		m.setSenderLP(LPid);
		m.setReceiverLP(targetLPid);


		//If this message was generated by an event previously rolled back and now
		//re-processed, we might have lazily cancelled it, so it doesn't need to be resent.
		if (useLazyCancellation && lazilyCancelledMessages.containsKey(ancestor))
		{
			List<Message> unsentAntiMsgs = lazilyCancelledMessages.get(ancestor);
			ListIterator<Message> iter = unsentAntiMsgs.listIterator();
			while (iter.hasNext())
			{
				Message antiMsg = iter.next();
				if (antiMsg.contentEquals(m))
				{
					//So, we have re-generated the same message as the one previosly cancelled
					//No need to send this message out
					iter.remove();

					//Store the anti message in case of a later rollback
					if (!processedMessages.isEmpty())
						processedMessages.getLast().getAntiMessages().add(antiMsg);

					if (collectStat)
						lpstat.numAntiMessagesLazilyCancelled++;
					return;
				}
			}
		}

		if (TRACE)
			System.out.printf("(LPid=%d, APPid=%s) SENT a positive message with TS=%d to (APPid=%s).\n",
					LPid, APPid, sendtime, targetAPPid);

		scheduler.sendMessage(m);

		//Store an anti-message copy of this message
		if (!processedMessages.isEmpty())
			processedMessages.getLast().getAntiMessages().add(m.makeAntiMsg());
	}


	/**
	 * Schedules the execution of an action which cannot be rolled back.
	 * An example of such action is standard output. The Runnable passed
	 * to this method will be executed after the Global Virtual Time
	 * has passed the time of the request, and thus the request is guaranteed
	 * to not be rolled back.
	 *
	 * @param action the permanent action which is to be queued up
	 */
	public void submitIrreversibleAction(Runnable action)
	{
		processedMessages.getLast().addIrreversibleAction(action);
	}


	/**
	 * Sets the end-of-sim flag so the simulation will stop. The first time an LP calls this method,
	 * the whole simulation will stop, even if there are pending events to be processed.
	 */
	synchronized public void setEndOfSimFlag()
	{
		scheduler.endSimulation();
	}


	/** Record the beginning time of this LP. */
	void recordBeginTime()
	{
		lpstat.beginTime = System.currentTimeMillis();
	}


	/** Record the ending time of this LP. */
	void recordEndTime()
	{
		lpstat.endTime = System.currentTimeMillis();
	}


	/**
	 * Assigns the {@link SchedulingLoop} for this LP.
	 * @param sl SchedulingLoop to control this LP. 
	 */
	void setScheduler(SchedulingLoop sl)
	{
		scheduler = sl;
	}


	/** Returns PE id where this LP resides. */
	public int getPID()
	{
		return PEid;
	}


	/**
	 * Sets the PE id where the LP resides. Can be used in a potential LP migration.
	 * 
	 * @param pid the Process Executive which owns this PE
	 */
	void setPID(int pid)
	{
		PEid = pid;
	}



	/** Returns the LP id. */
	synchronized public int getLID()
	{
		return LPid;
	}


	/**
	 * Sets the LP id for this LP. LP id is unique and system-assigned.
	 * @param lid the LP id.
	 */
	synchronized void setLID(int lid)
	{
		LPid = lid;
	}



	/** Returns the application-assigned id of this LP. */
	public String getAPPid()
	{
		return APPid;
	} //ends getAppid



	/**
	 * Sets the application-specific short id for this LP.
	 * @param aid the application-specific id for this LP.
	 */
	void setAppID(String aid)
	{
		APPid = aid;
	} //end setAppID



	/** Returns the name of this LP. */
	public String getName()
	{
		return myname;
	} //ends getName



	/**
	 * Sets the long name of this LP.
	 * @param n String, the long name of this LP.
	 */
	void setName(String n)
	{
		myname = n;
	} //end setName



	/** Returns the current simulation time of this LogicalProcess. */
	public int getCurrentTime()
	{
		return currentTime;
	}


	/**
	 * Stores the application-specific data used for configuration and passes it
	 * on to the application to apply. 
	 * @param cdata String representing the data that are used for configuration.
	 */
	public final void setConfigData(Serializable cdata)
	{
		appSetConfigData(cdata);
	} //end setConfigData



	/**
	 * Returns a vector of Strings describing all monitorable parameters of this LP,
	 * each in the form "paramName, paramType".
	 * Handles all LogicalProcess parameters; calls {@link #appGetMonitorable} to
	 * get application-specific parameters.
	 *
	 * @return List containing all monitorable parameters of this LP.
	 */
	public final Vector<NameValuePair> getMonitorable()
	{
		return appGetMonitorable();
	}


	/**Returns a List of Strings describing all steerable parameters of this LP,
	 * each in the form "paramName, paramType".
	 * Handles all LogicalProcess parameters; calls {@link #appGetSteerable} to
	 * get application-specific parameters.
	 *
	 * @return List containing all steerable parameters of this LP.
	 */
	public final Vector<NameValuePair> getSteerable()
	{
		return appGetSteerable();
	}


	/**
	 * Performs generic initialization (if any) then invokes application-specific 
	 * initialization. Equivalent of GTW's IProc method.
	 * @return false if initialization failed for some reason, true if successful.
	 */
	final void initializeLP()
	{
		appInitializeLP();
	}


	/**
	 * Sets the flag for collecting simulation statistics.
	 *@param cf boolean indicating whether statistics should be collected
	 */
	synchronized void setStatFlag(boolean cf)
	{
		collectStat = cf;
		lpstat.PEid = PEid;
		lpstat.LPid = LPid;
		if (STRACE)
			System.out.println("LP " + LPid + " set stat flag to " + collectStat);
	}


	/**
	 * Invokes application-specific termination then does any needed generic
	 * cleanup. Equivalent to GTW's FProc method. 
	 */
	synchronized void finalizeLP()
	{
		//record endding time for this LP
		recordEndTime();
		System.out.println(LPid + " is finalizing...");
		appFinalizeLP();
	}


	@Override
	public String toString()
	{
		return ("LP " + PEid + " " + LPid + " " + APPid);
	}
} //end class LogicalProcess


/**
 * This class is the record of the procesing of one incoming message by
 * the LP. It contains the saved state before the message was processed,
 * the message itself, and a list of all messages the LP sent out while
 * processing this message.
 */
class ProcessingRecord
{
	/** The message which was processed by the LP. */
	public Message incomingMessage;

	/** The saved state before the message was processed. */
	public Serializable savedState;
	
	/** Saved state of the LP, used to restore after rollback. */
	public final int lastTime, currentTime;

	/** The list of messages which the LP sent while processing this message. */
	private LinkedList<Message> antiMessages;

	/** The irreversible actions which the LP requested while processing this event. */
	private LinkedList<Runnable> irreversibleActions;


	/**
	 * Creates a new processing record.
	 *
	 * @param msg the message which has been processed by the LP
	 * @param state the saved state before this message was processed
	 */
	public ProcessingRecord(Message msg, Serializable state, int lastTime, int currentTime)
	{
		this.lastTime = lastTime;
		this.currentTime = currentTime;
		incomingMessage = msg;
		savedState = state;
	}


	/** The anti-messages for the messages which the LP sent while processing this event. */
	public synchronized List<Message> getAntiMessages()
	{
		if (antiMessages == null)
			antiMessages = new LinkedList<Message>();

		return antiMessages;
	}


	/**
	 * Add a new irreversible actions to be executed.
	 * These actions cannot be performed until the GVT passes their time stamp,
	 * since they cannot be rolled back.
	 *
	 * @param the new action which should be queued up
	 */
	public synchronized void addIrreversibleAction(Runnable action)
	{
		if (irreversibleActions == null)
			irreversibleActions = new LinkedList<Runnable>();

		irreversibleActions.add(action);
	}


	/**
	 * Executes all irreversible actions which are were generated by the LP
	 * while it was processing this event.
	 */
	public void commitIrreversibleActions()
	{
		if (irreversibleActions == null)
			return;

		for (Runnable action : irreversibleActions)
			action.run();
	}
}
