package sassy;

import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

/**
 * This class represents the messages sent and received by LPs.
 * 
 * @author George Vulov <georgevulov@hotmail.com>
 * @author Yin Xiong
 */
public class Message implements Serializable, Comparable<Message>
{
	/**
	 * Constants representing categories of messages that can be sent.
	 * <ul>
	 *  <li>Regular simulation messages
	 *  <li>Simulation anti-messages (canceling a regular message)
	 *  <li>System monitoring requests
	 *  <li>System monitoring reports
	 *  <li>System steering requests
	 *  <li>System steering acknowledgements
	 * </ul>
	 */
	public static enum MsgType
	{
		SIM_REGULAR, SIM_ANTI, SYSTEM_MONITOR, SYSTEM_STEER, SYSTEM_REPORT, SYSTEM_STAT

	}

	/**
	 * The application-specified Comparator used to break ties between simultaneous events.
	 * If this is null or this comparator returns 0, our compare function still breaks
	 * ties in a repeatable manner.
	 */
	private transient static Comparator<Message> simultEventComparator;

	/** Type of the message, matching one of the constants defined in this class. */
	private MsgType msgType;

	/**(Simulation) time this message should occur.*/
	private int timeStamp;

	/** Message content. If content == null, the content is stored in serialized form. */
	private transient Serializable content;

	/** The serialized content. This is what is actually transferred over the netowork. */
	private byte[] serializedContent;

	/**Sender's {@link LogicalProcess#APPid APPid}, if sender is a
	 * {@link LogicalProcess}, or identifier, if sender is a monitoring/steering
	 * application. */
	private String senderAPPid;

	/**Sender's LP number*/
	private int senderLP;

	/**Sender's PE number*/
	private int senderPE;

	/**Receiver's {@link LogicalProcess#APPid APPid}, if receiver is a
	 * {@link LogicalProcess}, or identifier, if receiver is a monitoring/steering
	 * application. */
	private String receiverAPPid;

	/**Receiver's LP number*/
	private int receiverLP;

	/**Receiver's PE number*/
	private int receiverPE;

	/**
	 * The "age" of this message. This counter increases with each message an LP sends
	 * and is used for breaking ties with simultaneous messages.
	 */
	private long age;

	/** The receiver PE's RMI remote reference. */
	// changed to transient - 9/15/2010
	private transient Peer receiverRMIStub;

	public static final long serialVersionUID = 12345L;

	/**
	 * The ancestor messages of a message are those whose processing caused this
	 * message to be generated. If a message has the same time stamp as its 
	 * ancestor, we store the ancestor's information to make sure a message
	 * can't roll back its own ancestor.
	 * 
	 * key: the LP ID of the ancestor message
	 * value: the ages of the ancestor messages
	 */
	private Map<Integer, List<Long>> ancestorMessages;


	/**
	 * Sets the Comparator which will be invoked if two messages have an equal time stamp.
	 * It can be used by the application if simutaneous events should always be be processed
	 * in a certain manner. If this comparator is not set or it returns 0, the kernel uses
	 * other tie-breaking paramaters to ensure repeatability.
	 * 
	 * @param comparator the simultaneous events tie-breaking comparator
	 */
	static void setSimultEventComparator(Comparator<Message> comparator)
	{
		simultEventComparator = comparator;
	}


	/**
	 * Preferred constructor; constructs a Message object of undefined origin and
	 * destination.
	 * The sender and receiver info must be set later.
	 * NOTE: The age is only necessary for simulation messages; it is not used in
	 *		 monitoring/steering messages.
	 *
	 * @param time The local sim time this message occurs.
	 * @param typ MsgType of this Message.
	 * @param cont Object the content of the message.
	 * @param ancestor the message whose processing generated this message. This is null
	 *					for the very first message sent in the simulation
	 * @param age the age counter of the LP sending this message 
	 */
	public Message(int time, MsgType typ, Serializable cont, Message ancestor, long age)
	{
		timeStamp = time;
		msgType = typ;
		content = cont;
		senderLP = -1;
		senderPE = -1;
		receiverLP = -1;
		receiverPE = -1;
		this.age = age;

		//If the ancestor of this message has the same time stamp as the message, we need to 
		//store a reference to it
		if (ancestor != null && timeStamp == ancestor.timeStamp)
		{
			Map<Integer, List<Long>> higherAncestors = ancestor.ancestorMessages;

			//Copy all the higher ancestors.
			if (higherAncestors != null)
			{
				ancestorMessages = new HashMap<Integer, List<Long>>(higherAncestors.size() + 1);

				for (Entry<Integer, List<Long>> entry : higherAncestors.entrySet())
				{
					List<Long> agesCopy = new ArrayList<Long>(entry.getValue());
					ancestorMessages.put(entry.getKey(), agesCopy);
				}
			}
			else
			{
				ancestorMessages = new HashMap<Integer, List<Long>>(4);
			}

			//Add this ancestor to the list of ancestors
			if (!ancestorMessages.containsKey(ancestor.senderLP))
				ancestorMessages.put(ancestor.senderLP, new ArrayList<Long>());
			ancestorMessages.get(ancestor.senderLP).add(ancestor.age);
		}

		//Serialize the content
		if (content != null)
		{
			try
			{
				ByteArrayOutputStream out = new ByteArrayOutputStream();
				ObjectOutputStream serializer = new ObjectOutputStream(out);
				serializer.writeObject(content);
				serializer.close();
				serializedContent = out.toByteArray();
			}
			catch (IOException e)
			{
				throw new RuntimeException("Serialization of the message contents failed!", e);
			}
		}
	} //end constructor


	/**Constructs a Message object with preliminary (not complete) addressing.
	 * senderLP, senderPE, receiverLP, and receiverPE must be specified later!
	 *@param time The local sim time this message occurs.
	 *@param typ MsgType of this Message.
	 *@param cont Object the content of the message.
	 *@param sender String the APPid of the sender LP.
	 *@param receiver String the APPid of the receiver LP.
	 * @param ancestor the message whose processing generated this message. This is null
	 *					for the very first message sent in the simulation
	 * @param age the age counter of the LP sending this message 
	 */
	public Message(int time, MsgType typ, Serializable cont, String sender, String receiver, Message ancestor, long age)
	{
		this(time, typ, cont, ancestor, age);
		senderAPPid = sender;
		receiverAPPid = receiver;
	} //end constructor


	/** Returns a copy of the passed-in message with a negative sign. */
	Message makeAntiMsg()
	{
		Message antiMsg = new Message(timeStamp, Message.MsgType.SIM_ANTI, null, senderAPPid, receiverAPPid, null, age);
		antiMsg.senderLP = this.senderLP;
		antiMsg.senderPE = this.senderPE;
		antiMsg.receiverLP = this.receiverLP;
		antiMsg.receiverPE = this.receiverPE;
		antiMsg.content = content;
		antiMsg.serializedContent = serializedContent;
		if (ancestorMessages != null)
			antiMsg.ancestorMessages = new HashMap<Integer, List<Long>>(ancestorMessages);

		return antiMsg;
	}


	/**Returns the PE id of the receiver of the message.
	 */
	public int getReceiverPE()
	{
		return receiverPE;
	} //end getsReceiverPE


	/**Sets the sender's PE number.
	 *@param sPE int the PE where the sender resides.
	 */
	public void setSenderPE(int sPE)
	{
		senderPE = sPE;
	} //end setSenderPE


	/**Sets the sender's LP number.
	 *@param sLP int the LP id of the sender.
	 */
	public void setSenderLP(int sLP)
	{
		senderLP = sLP;
	} //end setSenderLP


	/**Sets the receiver's PE number.
	 *@param rPE int the PE where the receiver resides.
	 */
	public void setReceiverPE(int rPE)
	{
		receiverPE = rPE;
	} //end setReceiverPE


	/**Sets the receiver's LP number.
	 *@param rLP int the id of the receiver LP.
	 */
	public void setReceiverLP(int rLP)
	{
		receiverLP = rLP;
	} //end setReceiverLP


	/**Sets the sender APPid.*/
	public void setSenderAPPid(String said)
	{
		senderAPPid = said;
	} //end setsSenderName


	/**Sets the receiver APPid.*/
	public void setReceiverAPPid(String raid)
	{
		receiverAPPid = raid;
	} //end setsReceiverName


	/**gets the receiver LP.*/
	public int getReceiverLP()
	{
		return receiverLP;
	} //end getsReceiverLP


	/**gets the sender LP.*/
	public int getSenderLP()
	{
		return senderLP;
	} //end getsSenderLP


	/**Returns the sender name.*/
	public String getSenderAPPid()
	{
		return senderAPPid;
	} //end getSenderName


	/**Returns the receiver name.*/
	public String getReceiverAPPid()
	{
		return receiverAPPid;
	} //end getReceiverName


	/**Returns the type of the message.*/
	public MsgType getMsgType()
	{
		return msgType;
	}


	/**Returns the time stamp of the message.*/
	public int getTimeStamp()
	{
		return timeStamp;
	}


	/**Returns the content of the message.*/
	public synchronized Serializable getContent()
	{
		//De-serialize the contents
		if (content == null && serializedContent != null)
		{
			try
			{
				ByteArrayInputStream in = new ByteArrayInputStream(serializedContent);
				ObjectInputStream deserializer = new ObjectInputStream(in);
				content = (Serializable)deserializer.readObject();
			}
			catch (IOException e)
			{
				throw new RuntimeException("Deserialization of message contents failed!", e);
			}
			catch (ClassNotFoundException e)
			{
				throw new RuntimeException("Class needed for deserialization of message content not found!", e);
			}
		}

		return content;
	}


	/**@return string representation of the Message object. */
	public String toString()
	{
		return "Message of type " + msgType + " with timestamp " + timeStamp +
				" from (" + senderAPPid + "," + senderLP + "," + senderPE + ") to (" +
				receiverAPPid + "," + receiverLP + "," + receiverPE + ") with content " +
				content;
	}


	/**Compares two Messages' time stamps. If this Message has an earlier time
	 * stamp than the other, it is "less than" it, and a negative number is
	 * returned. If this message has a later time stamp, it is "greater" and a
	 * positive number is returned. If they have the same time stamp, zero is
	 * returned (marking them as "equal"). As such, note: this class has a natural
	 * ordering that is inconsistent with equals. 
	 *
	 * IMPORTANT ASSUMPTIONS:
	 * 1. A message should always compareTo() 0 with its anti-message
	 * 2. A message should always compareTo() 0 with itself
	 * 3. Two messages should not compareTo() 0 unless they're the same message!
	 */
	@Override
	public int compareTo(Message O)
	{
		//First, compare the timestamps
		if (timeStamp != O.timeStamp)
			return timeStamp - O.timeStamp;

		//If the two messages are actually the same, no need to invoke the tie breaking comparator
		if (senderLP == O.senderLP && age == O.age)
			return 0;


		//FIXME: Implementing dependency order like this guarantees that a message doesn't
		//roll back its ancestor, but it gives ordering to messages which are unrelated
		
		//This is the number of ancestors with the same timestamp as the message (i.e. 0-lookahead)
		int myNumAncestors = 0, hisNumAncestors = 0;
		if (ancestorMessages != null)
		{
			for (List<Long> ages : ancestorMessages.values())
				myNumAncestors += ages.size();
		}
		
		if (O.ancestorMessages != null)
		{
			for (List<Long> ages : O.ancestorMessages.values())
				hisNumAncestors += ages.size();
		}
		
		if (myNumAncestors != hisNumAncestors && getReceiverLP() == O.getReceiverLP())
			return myNumAncestors - hisNumAncestors;
		
		//FIXME: There are fallacies with this implementation of the dependency order
		//Check if one message is an ancestor of the other
//		if (ancestorMessages != null && ancestorMessages.containsKey(O.senderLP))
//		{
//			//The message O is our ancestor and thus precedes us
//			if (ancestorMessages.get(O.senderLP).contains(O.age))
//				return 1;
//		}
//		if (O.ancestorMessages != null && O.ancestorMessages.containsKey(senderLP))
//		{
//			//We're the ancestor of O and thus precede it
//			if (O.ancestorMessages.get(senderLP).contains(age))
//				return -1;
//		}

		//Ok, so we have two different messages with the same timestamp.
		//We invoke the application-specific tie-breaking comparator, if it is set
		if (simultEventComparator != null)
		{
			int res = simultEventComparator.compare(this, O);
			if (res != 0)
				return res;
		}

		//If the time stamps match, lesser LPid goes first
		if (senderLP < O.senderLP)
			return -1;
		else if (senderLP > O.senderLP)
			return 1;

		//If the timestamps and the LPid match, compare the age
		//If the ages are the same, then we return 0 since it's the same message
		return (int)(age - O.age);
	}

	
	@Override
	public boolean equals(Object o) {

		if(o == null || !(o instanceof Message))
			return false;
		Message other = (Message) o;
		if(senderLP == other.senderLP && age == other.age)
			return true;

		return contentEquals(other);
	}


	@Override
	public int hashCode() {
		assert false : "Hash Code not assigned";
		return 42;
	}

	
	/**
	 * Compares the contents of two messages. Returns true
	 * only if the two messages originate from the same LP, are destined for the
	 * same LP, and have the same time stamp and contents. Returns true if contents
	 * match, even if one message is a positive message and the other message is an
	 * anti-message.
	 * 
	 * @param other message which is being compared for equality
	 * @return true if the two messages are the same, false otherwise.
	 */
	public boolean contentEquals(Message other)
	{
		if ((senderLP != other.senderLP) || (timeStamp != other.timeStamp))
			return false;
		
		if (receiverAPPid == null || other.receiverAPPid == null)
		{
			//Compare the ID's of the destinations
			if (receiverLP != other.receiverLP)
				return false;
		}
		else
		{
			//Compare the names of the destinations
			if (!receiverAPPid.equalsIgnoreCase(other.receiverAPPid))
				return false;
		}
		
		return Arrays.equals(serializedContent, other.serializedContent);
	}

	public Peer getReceiverRMIStub()
	{
		return receiverRMIStub;
	}


	void setReceiverRMIStub(Peer receiverRMIStub)
	{
		this.receiverRMIStub = receiverRMIStub;
	}
}

