package sassy;

import java.io.*;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.*;
import java.rmi.*;
import java.rmi.server.*;
import java.net.*;

/**
 * This class represents a physical processor involved in the time warp
 * simulation engine. A PE could have multiple LPs running on it.
 *
 * @author Yin Xiong
 * @author George Vulov <georgevulov@hotmail.com>
 */
public class PE extends Thread implements Peer
{
	/**The flag for displaying debugging info for statistics collection*/
	public static final boolean STRACE = false;

	/**The flag for displaying debugging info for monitoring and steering. */
	public static final boolean MTRACE = false;

	/** If true, display trace information for debugging. */
	private static final boolean TRACE = false;

	/**The flag to begin the simulation*/
	volatile boolean beginToRun = false;

	/** The RMI remote reference used to communicate with this PE. */
	private Peer rmiPeerStub;

	/**The id of this PE.*/
	private int myID;

	/**The name of the machine where the PE is running*/
	private String myMachineName;

	/**The ip of the machine where the PE is running*/
	private String myIP;

	/** The global name server*/
	private GNSInterface gns;

	/** The LocalNameServer in charge of routing messages.*/
	private LocalNameServer localNameServer;

	/**
	 * The queue that holds the outgoing messages. Messages are placed into this
	 * by the PE from SchedulingLoop.outgoing. Then the PE sends out events
	 * from sendQ.
	 */
	private LinkedList <Message> sendQ = new LinkedList<Message>();

	/**The scheduler that schedules events for this PE.*/
	private SchedulingLoop schedulingLoop;

	/** True if the simulation has ended, false if it's running.*/
	private volatile boolean simulationEnded;

	/** True if the connection to the GNS is up. */
	private boolean gnsRunning;

	/** The name of the GNS server. */
	private String serverAddress;

	/**
	 * True if we're in GVT calculation mode; false otherwise.
	 * While we're in GVT calculation, we send marked acknowledgements
	 * for any messages we receive, so they can be considered in the GVT
	 * calculation of the sender. (Samadi's Algorithm)
	 */
	private volatile boolean isCalculatingGVT;

	/**
	 * The timestamp of the minimum marked acknowledgement we have received.
	 * Used in GVT calculation using Samadi's Algorithm.
	 */
	private int markedAckTS = Integer.MAX_VALUE;


	public PE(String sv)
	{
		super("sassy.PE");
		simulationEnded = false;
		gnsRunning = false;
		localNameServer = new LocalNameServer();

		try
		{
			InetAddress address = InetAddress.getLocalHost();
			myMachineName = address.getHostName();
			myIP = address.getHostAddress();
			serverAddress = sv;
		}
		catch (java.net.UnknownHostException e)
		{
			System.out.println("PE(String sv): "+e);
			e.printStackTrace();
		}
	} //end constructor

	/**Returns the statistics to GNS.
	  *@param lid int the id of the LP whose statistics is to be sent.
	  */
	@Override
	public LPStat getLPStat(int lid)
	{
		LogicalProcess lp = getLPByID(lid);
		return lp.lpstat;
	}

	/**Sets the beginToRun flag for this PE. */
	@Override
	public void beginSim()
	{
		beginToRun = true;
	}

	/**
	 * Connects this PE to the GNS server, receives LP information and create LPs
	 * accordingly.
	 */
	@SuppressWarnings("unchecked")
	private void connectToGNS()
	{
		//Register ourselves as a remote object
		try
		{
			rmiPeerStub = (Peer) UnicastRemoteObject.exportObject((Peer)this, 0);
			System.out.println("\nPE server started successfully on " + myMachineName + " " + myIP);
		}
		catch (RemoteException e)
		{
			System.err.println("Error initializing RMI for the PE:\n" + e);
			System.exit(1);
		}
		
		//Find the GNS server and connect to it
		try
		{
			Registry remoteRegistry = LocateRegistry.getRegistry(serverAddress);
			gns = (GNSInterface)remoteRegistry.lookup("GNS");
			GNS.PEConfig myConfig = gns.registerPE(myIP, myMachineName, rmiPeerStub);
			List <LPConfigInfo> lpConfigInfo = myConfig.LPConfigs;

			System.out.printf("%s connected to GNS=%s and received %d LPs.\n",
					  myMachineName, serverAddress, lpConfigInfo.size());
			gnsRunning = true;

			//Create all the LP's
			LPConfigInfo lpinfo;
			if (lpConfigInfo.size() > 0)
			{
				lpinfo = lpConfigInfo.get(0);
				myID = lpinfo.PEid;
				schedulingLoop = new SchedulingLoop(myID, this);
				PEInfo myPEInfo = new PEInfo(myID, myIP, myMachineName, this);

				//Create LPs & add their routing info to the LNS
				for (LPConfigInfo lp : lpConfigInfo)
				{
					schedulingLoop.createLP(lp);
					RoutingInfo lpRoutingInfo = new RoutingInfo(myPEInfo, lp.LPid, lp.getAPPid());
					localNameServer.addRoutingInfo(lpRoutingInfo);
				}

				//gets and sets the flag for collecting statistics
				boolean cf = gns.isCollectingStats();
				if (STRACE)
					System.out.println("PE " + myID + "stat flag is set to " + cf);
				schedulingLoop.setStatFlag(cf);
			}

			//Load the simultenous event comparator
			if (myConfig.simultEventComparator != null)
			{
				try
				{
					Object comparator = Class.forName(myConfig.simultEventComparator).newInstance();
					Message.setSimultEventComparator((Comparator <Message>)comparator);
				}
				catch (Exception ex)
				{
					System.err.printf("Could not load simultaneous event comparator \"%s\"\n", myConfig.simultEventComparator);
					ex.printStackTrace(System.err);
				}
			}
		}
		catch (RemoteException e)
		{
			System.err.printf("Error: Could not connect to GNS at %s.\n", serverAddress);
			System.exit(1);
		}
		catch (NotBoundException e)
		{
			System.err.printf("Error: Connected to RMI registry, but no GNS object found.\n");
			System.exit(1);
		}
	} //end connectToGNS


	private int gvtCounter = 0;
	private static int GVT_FREQUENCY = 300;
	
	/**
	  * Runs the simulation until it is ended. This loop takes messages in the "outgoing" queue,
	  * looks up their routing information, and dispatches them to the correct process executive.
	  */
	@Override
	public void run()
	{
		while (!beginToRun)
		{
			System.out.printf("PE %d: Waiting for all PE's to join...\n", getID());
			try
			{
				Thread.sleep(1000);
			}
			catch (InterruptedException ex)
			{
			}
		}
		System.out.printf("PE %d: All PE's ready, starting simulation\n", getID());
		
		initLPs();

		schedulingLoop.start();

		while (!simulationEnded)
		{
			try
			{
				//FIXME: This needs to be moved somewhere else!
				gvtCounter = (gvtCounter + 1) % GVT_FREQUENCY;
				if (gvtCounter == 0)
					gns.requestGVTCalculation();
			}
			catch (RemoteException ex)
			{
				System.out.println("(PE)run(): "+ex);
				ex.printStackTrace();
			}

			//send messages stored in sendQ
			rmiSendMessages();

			//While the outgoing queue is empty, keep waiting
			synchronized(schedulingLoop.outgoing)
			{
				while (schedulingLoop.outgoing.isEmpty() && !simulationEnded)
				{
					try
					{
						schedulingLoop.outgoing.wait();
					}
					catch (InterruptedException ex)
					{
						System.out.println("(PE)run(synchronized): "+ex);
					}
				}
			}

			//The queue is no longer empty, process events now
			if (!simulationEnded)
			{
				processEvents();
			}
		} //end while

		System.out.println(myMachineName + " finished running.");
		finalizePE();
	} //end run


	/**
	 * Stop the simulation globally by notifying the GNS that the simulation
	 * has terminated.
	 */
	private void finalizePE()
	{
		if (!gnsRunning)
			return ;

		try
		{
			if (gns.isSimulationRunning())
			{
				//System.out.println(myMachineName + " asked GNS to end the simulation.");
				gns.endSimulation();
			}
		}
		catch (RemoteException e)
		{
			//System.out.println("(PE)finalizedPE(): "+e);
			e.printStackTrace();
		}
	} //end finalizePE


	/**
	 * Sends messages by invoking the remote method receiveMsg(Message msg) on
	 * the receiver machine.
	 */
	private synchronized void rmiSendMessages()
	{
		while (!sendQ.isEmpty() && gnsRunning)
		{
			Message msg = sendQ.poll();
			int pid = msg.getReceiverPE();

			if (TRACE)
				System.out.printf("PE.java: rmiSendMesages(): Sending to PE %d, LP %s from LP %s.\n",
						  pid, msg.getReceiverAPPid(), msg.getSenderAPPid());

			Peer destinationPE = msg.getReceiverRMIStub();
			try
			{
				boolean receiverGVTCalculation = destinationPE.receiveMsg(msg);

				//If the receiver was in the process of calculating the GVT, we need to
				//counsider this meassage in our GVT calculation since the receiver won't
				if (receiverGVTCalculation)
					markedAckTS = Math.min(markedAckTS, msg.getTimeStamp());
			}
			catch (RemoteException e)
			{
				System.err.printf("Fatal error: RMI Communication to PE %d Failed.\n", pid);
				e.printStackTrace();
				System.exit(1);
			}
		} 
	}


	/**
	 * Processes events. It takes out each message in SchedulingLoop's outgoing
	 * queue, finds out the routing information by application-specified id of the
	 * destination or the LP id of the destination, consults the local name server
	 * or the global name server (in case local name server does not have the info)
	 * for the routing information. It then puts the message into the sendQ for
	 * sending out.
	 */
	private synchronized void processEvents()
	{
		while (!schedulingLoop.outgoing.isEmpty())
		{
			Message msg = schedulingLoop.outgoing.poll();

			String receiverAppId = msg.getReceiverAPPid();
			int receiverLpId = msg.getReceiverLP();

			//if receiver is gns, it must be an MS message, send the message to gns directly

			RoutingInfo rinfo = null;
			boolean byName = false;

			if (receiverAppId != null)
			{
				//if receiver is gns, it must be an MS message, send the message to gns directly
				if ( receiverAppId.equalsIgnoreCase("gns"))
				{
					try
					{
						gns.receive_MS_Msg(msg);
					}
					catch (Exception e)
					{
						e.printStackTrace();
					}
					return;
				}

				//check lns for receiver lp
				else
				{
					rinfo = localNameServer.getRoutingInfoByAid(receiverAppId);
					byName = true;
				}
			}
			else
			{
				rinfo = localNameServer.getRoutingInfoByLid(receiverLpId);
			}

			//if lns doesn't have the receiver info, check gns
			while (rinfo == null)
			{
				try
				{
					if (byName == true)
						rinfo = gns.getRoutingInfoByAid(receiverAppId);
					else
						rinfo = gns.getRoutingInfoByLid(receiverLpId);
				}
				catch (RemoteException e)
				{
					System.out.println("(PE)processEvent(while): "+e);
					e.printStackTrace();
				}

				//The GNS could not determine where to route this message; wait and try again
				if (rinfo == null)
				{
					System.err.printf("Could not determine where to send message destined for LP with ID %d, name '%s'\n", 
							receiverLpId, receiverAppId);
					System.exit(1);
				}
				else
				{
					//store rinfo in local name server for future use
					localNameServer.addRoutingInfo(rinfo);
				}		
			}

			//System.out.println("PE " + myID + " get routinfo: " + rinfo);
			if (rinfo != null)
			{
				msg.setReceiverPE(rinfo.getPEid());
				msg.setReceiverAPPid(rinfo.getAPPid());
				msg.setReceiverLP(rinfo.getLPid());
				msg.setReceiverRMIStub(rinfo.getRmiPeerStub());

				//if the receiver is not on this machine, put the msg in sendQ
				if (rinfo.getPEid() != myID)
					sendQ.add(msg);
				else
				{
					synchronized(schedulingLoop.incoming)
					{
						schedulingLoop.incoming.offer(msg);
						schedulingLoop.incoming.notify();
					}
				}
			} //end if rinfo is not null
		} //end for
	} //end ProcessEvents


	/** The main function for this PE.*/
	public static void main (String[] args) throws IOException
	{
		if (args.length != 1)
		{
			System.out.println("Usage: java PE [GNS server name]");
			System.exit(1);
		}

		PE pe = new PE(args[0]);
		pe.connectToGNS();
		pe.start();		
	}


	/**
	 * Calculates the local minimum timestamp. It is the minimum of the incoming queue,
	 * outgoing queue, simulation messages queues, sendQ and messages which have
	 * been sent to another PE but have received marked acknowledgments. (as described
	 * in Samadi's algorithm).
	 *
	 * @returns the local minimum timestamp of this PE
	 */
	private synchronized int calculateLocalMin()
	{
		//Note: It is very important that both this method and the methods
		//processEvents() and rmiSendMessages() are synchronized so
		//they cannot run synchronously. Else, messages can slip through the
		//cracks due to race issues
		
		//FIXME: Here we should synchronize on the scheduling loop, but that would cause deadlock issues
		//The whole synchronization of SASSY needs to be revisited.
		//synchronized (schedulingLoop)
		{
			int localMin = schedulingLoop.getMessageQueuesMinTS();

			//Take in account the unsent messages
			if (!sendQ.isEmpty())
			{
				Message front = Collections.min(sendQ);
				localMin = Math.min(front.getTimeStamp(), localMin);
			}
			
			//Take in account send messages with marked acknowledgements
			localMin = Math.min(markedAckTS, localMin);
			
			return localMin;
		}
	}


	/**
	 * Sets the flag to end the simulation running in this PE. This will
	 * in stop the simulation loop, finalize all the LP's, and then terminate
	 * the simulation globally by notifying the GNS.
	 */
	public synchronized void endSimulation()
	{
		if (!simulationEnded)
			finalizeLPs();
		simulationEnded = true;

		//If the PE is waiting for something in the outgoing queue to appear, tell it to stop
		synchronized(schedulingLoop.outgoing)
		{
			schedulingLoop.outgoing.notify();
		}

		//Stop the scheduling loop
		if (schedulingLoop.isRunning())
			schedulingLoop.endSimulation();

	}


	/**
	 * Returns true if the PE is still running, false otherwise.
	 *
	 * @return whether this PE has terminated or not
	 */
	public boolean isRunning()
	{
		return !simulationEnded;
	}


	/**
	 * Gets the LP by its id.
	 * @param lpid int the LP id.
	 */
	public LogicalProcess getLPByID(int lpid)
	{
		//System.out.println("in getLPByID,size:" + LPs.size());

		for (int i = 0;i < schedulingLoop.LPs.size();i++)
		{
			LogicalProcess lp = schedulingLoop.LPs.get(i);
			if (lp.getLID() == lpid)
				return lp;
		}
		return null;
	}


	/**
	 * Initialize LPs by invoking initializeLP method provided by subclasses
	 * of LogicalProcess.
	 */
	private void initLPs()
	{
		for (int i = 0;i < schedulingLoop.LPs.size();i++)
		{
			LogicalProcess sublp = schedulingLoop.LPs.get(i);
			sublp.initializeLP();
		}
	}


	/**
	 * Finalizes LPs by invoking finalizeLP method provided by subclasses of
	 * LogicalProcess.
	 */
	private void finalizeLPs()
	{
		//System.out.println("PE " + getID() + " finalize " + schedulingLoop.LPs.size());
		for (int i = 0;i < schedulingLoop.LPs.size();i++)
		{
			LogicalProcess sublp = schedulingLoop.LPs.get(i);
			sublp.finalizeLP();
		} //end for
	} //end finalizeLPs


	/**Returns the id of the PE.*/
	public int getID()
	{
		return myID;
	}


	/**
	 * Receives a message from a remote PE. Whether the receiving
	 * PE is in the process of GVT calculation.
	 * (We're using Samadi's algorithm for GVT calculation)
	 *
	 * @param msg a Message object that is to be sent to this PE.
	 * @returns true if the receiver has already reported its local min for GVT calculation
	 */
	@Override
	public boolean receiveMsg(Message message) throws RemoteException
	{
		if (TRACE)
			System.out.printf("PE.java: receiveMsg(%s)\n", message.getContent().toString());

		//Notify that there are new entries in the incoming queue
		synchronized(schedulingLoop.incoming)
		{
			schedulingLoop.incoming.offer(message);
			schedulingLoop.incoming.notify();
		}

		return isCalculatingGVT;
	}


	/**
	 * RMI method to shut down this PE. This should only be called by the
	 * GNS.
	 */
	@Override
	public void shutDown() throws RemoteException
	{
		endSimulation();
		gnsRunning = false;
		try
		{
			//System.out.println(myMachineName + " shutting down...");
			UnicastRemoteObject.unexportObject(this, true);
		}
		catch (Exception e)
		{
			System.out.println("(PE)shutDown(): "+e);
			e.printStackTrace();
		}
	}


	/**
	 * <p>RMI method which informs the PE that GVT calculation has begun. 
	 * The PE should calculate the local minimum of all unprocessed and 
	 * unacknowledged events it holds and return it.</p>
	 *
	 * <p>After this method is called, a PE should only send marked acknowledgments
	 * when receiving messages.</p>
	 *
	 * @returns the local minimum time stamp of the PE
	 */
	@Override
	public int GVTCalculationStarted() throws RemoteException
	{
		isCalculatingGVT = true;

		return calculateLocalMin();
	}


	/**
	 * RMI method which informs the PE that GVT calculation has ended and 
	 * passes it the new GVT value. The PE can now continue to send regular 
	 * acknowledgments when receiving messages and can perform fossil collection
	 * based on the new GVT value.
	 *
	 * @param gvt the newly calculated global virtual time
	 */
	@Override
	public void GVTCalculationEnded(int gvt) throws RemoteException
	{
		isCalculatingGVT = false;

		//Since GVT calculation has ended, we reset the timestamp
		//of marked acknowledgements. It was used to ensure that no message
		//was missed in the GVT calculation (Cf. Samadi's algorithm)
		markedAckTS = Integer.MAX_VALUE;

		//Let the scheduling loop know
		schedulingLoop.setGVT(gvt);
	}
}
