package sassy;

import sassy.monitoring_steering.MonitorNode;
import sassy.monitoring_steering.NameValuePair;
import sassy.monitoring_steering.SteerNode;
import sassy.monitoring_steering.ReportNode;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.*;
import java.lang.reflect.*;

/**
 * This class provides the simulation engine logical process scheduling loop. This
 * loop runs in its own thread in parallel with the communication thread
 * of the PE. After instantiation, this class should have one or more
 * LogicalProcesses added (by {@link #createLP createLP}) before the thread is
 * started.
 *
 * @author Yin Xiong
 * @author George Vulov <georgevulov@hotmail.com>
 * @author Glenn Matthews
 */
class SchedulingLoop extends Thread
{
	/**The flag for displaying debugging info */
	public final boolean TRACE = false;

	/**The flag for displaying debugging info for monitoring and steering. */
	public final boolean MTRACE = false;

	/**The flag for displaying debugging info for statistics collecting */
	public final boolean STRACE = false;

	/**The id of the parent PE of this scheduler.*/
	private int myID;

	/**The flag indicating whether the simulation is finished.*/
	private volatile boolean simulationEnded;

	/**The most recent GVT*/
	private volatile int GVT;

	/**True if elements with TS < GVT should be deleted (i.e. fossil collection should occur).*/
	private volatile boolean shouldCollectFossils;

	/**The flag for collecting statistics*/
	private boolean collectStat = false;

	/** Queue containing incoming {@link Message}s from the communication thread. */
	final Queue<Message> incoming = new ConcurrentLinkedQueue<Message>();

	/** Shared priority queue containing outgoing {@link Message}s to the
	 * communication thread. Note that the addressing of these Messages are
	 * not likely to be fully specified, since this class does not have access
	 * to a GlobalNameServer. In general, the communication thread should
	 * expect to have to use {@link Message#senderAPPid} and
	 * {@link Message#receiverAPPid} to look up the LPid and PE of the sender
	 * and receiver. The communication thread is also expected to recognize
	 * and handle the case where receiverAPPid ==
	 * {@link LogicalProcess#ALL_LPS}.
	 */
	final Queue<Message> outgoing = new ConcurrentLinkedQueue<Message>();;

	/** Local list storing all LogicalProcesses managed by this SchedulingLoop. */
	List<LogicalProcess> LPs;

	/** The PE that this SchedulingLoop is running. */
	private PE myPE;

	/** Queue storing positive simulation messages yet to be handled. */
	private PriorityQueue<Message> simMessages;

	/** Queue storing anti-messages not yet processed. */
	private PriorityQueue<Message> antiMessages;

	/** Local list storing sorted sets (one per LP) of current and future
	 * monitoring commands.
	 */
	private List<SortedSet<MonitorNode>> monitoring;

	/** Local list storing sorted sets (one per LP) of current and future
	 * steering commands.
	 */
	private List<SortedSet<SteerNode>> steering;


	/**
	 * Constructs a new instance. Sets the id of the PE to peid.
	 *
	 * @param pid int the id of the PE, i.e. the id of this process.
	 * @param f Flag indicating whether the simulation application program is finished.
	 */
	public SchedulingLoop(int pid, PE myPE)
	{
		super("sassy.SchedulingLoop");
		GVT = 0;
		simulationEnded = false;
		myID = pid;
		this.myPE = myPE;
		LPs = new ArrayList<LogicalProcess>();

		simMessages = new PriorityQueue<Message>();
		antiMessages = new PriorityQueue<Message>();
		monitoring = new ArrayList<SortedSet<MonitorNode>>();
		steering = new ArrayList<SortedSet<SteerNode>>();
	} //end constructor


	/**Sets the flag for collecting statistics
	 *@param cf boolean indicating whether statistics should be collected.
	 */
	public void setStatFlag(boolean cf)
	{
		collectStat = cf;
		if (collectStat)
		{
			if (STRACE)
				System.out.println("Scheduler " + myID + " set stat flag to " +
						collectStat);
			for (int i = 0; i < LPs.size(); i++)
			{
				LogicalProcess lp = LPs.get(i);
				lp.setStatFlag(collectStat);
			} //end for
		}
	} //end setStatFlag


	/**Applies steering and gets the response from the lP.
	 * @param vtype String represending the type of steering: V=single variable
	 *	  L=all steerable variable of the LP;  only V is implemented for now.
	 * @param vname String representing the name of the variable to be steered.
	 * @param vvalue String representing the new value of the variable to be steered.
	 * @param lp the LP whose variable is to be steered.
	 */
	private Vector<NameValuePair> getSteerResult(String vtype, String vname, String vvalue, LogicalProcess lp)
	{
		Class c = lp.getClass();
		Field f;
		Vector<NameValuePair> resultVector = new Vector<NameValuePair>();

		if (MTRACE)
			System.out.println("***** in getSteerResult ****: " + vname + " " + vvalue);

		try
		{
			if (vtype.equalsIgnoreCase("V"))
			{
				if (MTRACE)
					System.out.println("***** var type is V ****");

				f = c.getField(vname);
				//find out the data type of this variable
				String dataType = "";
				Vector<NameValuePair> nvs = lp.getSteerable();
				for (int i = 0; nvs != null && i < nvs.size(); i++)
				{
					NameValuePair nt = nvs.get(i);
					System.out.println("name,value: " + nt.varName + " " + nt.varValue);
					if (vname.equals(nt.varName))
					{
						dataType = nt.varValue;
						break;
					} //end if
				} //end for

				if (MTRACE)
					System.out.println("steer var data type: " + dataType);

				String name = "steer";
				String value = "done";

				if (dataType.equals("int"))
					f.setInt(lp, Integer.parseInt(vvalue));
				else if (dataType.equals("double"))
					f.setDouble(lp, Double.parseDouble(vvalue));
				else if (dataType.equals("String"))
					f.set(lp, vvalue);
				else
					value = "failed";

				System.out.println("steer: " + vname + " " + f.get(lp));
				NameValuePair nv = new NameValuePair(name, value);
				resultVector.add(nv);
			} //end if steer type =V
		}
		catch (Exception e)
		{
			System.out.println("getSteerResult " + e);
			e.printStackTrace();

		}

		return resultVector;
	} //end isSteerDone


	/** Applies any scheduled steering to the specified LP.
	 * @param index The storage index of the LP to steer. */
	private void applySteering(int index)
	{
		LogicalProcess lp = LPs.get(index);
		SortedSet<SteerNode> steer = steering.get(index);
		int lptime = lp.getCurrentTime();
		if (steer.isEmpty())
		{
			if (MTRACE)
				System.out.println("No assigned steering for LP " + lp.getAPPid());
			return;
		}
		Iterator<SteerNode> it = steer.iterator();
		while (it.hasNext())
		{
			SteerNode s = it.next();
			if (s.startTime > lptime)
				break; //set is sorted, so all remaining elements aren't valid

			Vector<NameValuePair> resultVector = getSteerResult(s.varType, s.varName, s.varValue, lp);
			ReportNode rep = new ReportNode(s.clientID, s.msgID, lptime, resultVector);

			//send steering message to GNS
			Message msg = new Message(lptime, Message.MsgType.SYSTEM_REPORT, rep, null, 0);
			msg.setSenderLP(lp.getLID());
			msg.setReceiverAPPid("gns");
			sendMessage(msg);

			//for now we only do one-time steering, so we can delete this steering command now
			it.remove();
		} //end while

	} //end applySteering()


	/**Gets the value of the variables of an monitored LP
	 * @param vtype String representing the variable type: V=single variable; L=all monitorable
	 * variables of an LP
	 * @param vname String representing the variable name: L=names of all monitorable variables
	 * @param lp the LP whose variable values are to be got
	 */
	private Vector<NameValuePair> getMonResult(String vtype, String vname, LogicalProcess lp)
	{
		Class c = lp.getClass();
		Field f;
		Object fieldValue;

		if (MTRACE)
			System.out.println("***** in getMonResult ****: " + vname);

		Vector<NameValuePair> resultVector = new Vector<NameValuePair>();
		try
		{
			if (vtype.equalsIgnoreCase("V"))
			{
				if (MTRACE)
					System.out.println("***** var type is V ****");

				f = c.getField(vname);
				fieldValue = f.get(lp);
				String sf = fieldValue.toString();
				if (MTRACE)
					System.out.println("fieldValue,sf: " + fieldValue + " " + sf);

				NameValuePair nv = new NameValuePair(vname, sf);
				resultVector.add(nv);
			} //end if
			else
			{
				if (MTRACE)
					System.out.println("***** var type is L ****");

				Vector<NameValuePair> nvs = lp.getMonitorable();
				for (int i = 0; nvs != null && i < nvs.size(); i++)
				{
					NameValuePair nt = nvs.get(i);
					String varName = nt.getVarName();
					f = c.getField(varName);
					fieldValue = f.get(lp);
					String sf = fieldValue.toString();
					NameValuePair nv = new NameValuePair(varName, sf);
					resultVector.add(nv);
				} //end for
			} //end else
		} //end try
		catch (Exception e)
		{
			System.out.println("getMonResult " + e);
			e.printStackTrace();

		}

		return resultVector;
	} //end getMonResult


	/** Applies any scheduled monitoring of a single LP.
	 * @param index The storage index of the LP to monitor. */
	private void applyMonitoring(int index)
	{
		LogicalProcess lp = LPs.get(index);
		SortedSet<MonitorNode> mon = monitoring.get(index);
		int lptime = lp.getCurrentTime();

		if (MTRACE)
		{
			System.out.println("mon size: " + mon.size());
		} //end if MTRACE

		if (mon.isEmpty())
			return;

		Iterator<MonitorNode> it = mon.iterator();
		while (it.hasNext())
		{
			MonitorNode m = it.next();
			if (MTRACE)
			{
				System.out.println("monitoring msg: " + m.stringValue());
			}
			//if monitoring is scheduled for the future, don't do moniroting
			if (m.startTime > lptime)
				break; //set is sorted, so all remaining elements aren't valid

			//Monitoring is applied BEFORE old commands are deleted
			Vector<NameValuePair> resultVector = getMonResult(m.varType, m.varName, lp);

			ReportNode rep = new ReportNode(m.clientID, m.msgID, lptime, resultVector);
			//send monitoring message to GNS
			Message msg = new Message(lptime, Message.MsgType.SYSTEM_REPORT, rep, null, 0);
			msg.setSenderLP(lp.getLID());
			msg.setReceiverAPPid("gns");
			sendMessage(msg);

			//NOW we can delete this command if it's obsolete.
			if (m.stopTime < lptime)
				it.remove();
		} //end while
	} //end applyMonitoring()


	/**
	 * Places the specified (partly-formatted) Message into the outgoing message
	 * queue for the communication thread to send out.
	 *
	 * @param msg Message to send, assumed to have correctly filled senderAPPid,
	 * senderLP, and receiverAPPid, but not necessarily senderPE, receiverLP,
	 * or receiverPE (which are not likely to be known by this class or by any
	 * LogicalProcess, but should be known by the messaging thread).
	 */
	public void sendMessage(Message msg)
	{
		synchronized(outgoing)
		{
			outgoing.offer(msg);

			//Notify the PE loop that something has been added to the outgoing queue
			outgoing.notify();
		}
	}


	/**
	 * Runs the scheduling loop until the simulation is stopped. This loop looks in the incoming
	 * queue and processes the events there. Then, it takes the message with the smallest
	 * time stamp and executes it.
	 */
	@Override
	public void run()
	{
		//Scheduling loop
		while (!simulationEnded)
		{
			//If fossil collection is scheduled, do it
			if (shouldCollectFossils)
			{
				for (LogicalProcess lp : LPs)
					lp.collectFossils();

				shouldCollectFossils = false;
			}

			//If we have neither sim messages nor anti-messages to process, we need to wait for some
			//incoming messages
			if (simMessages.isEmpty() && antiMessages.isEmpty())
			{
				synchronized(incoming)
				{
					while (incoming.isEmpty() && !simulationEnded)
					{
						try
						{
							incoming.wait();
						}
						catch (InterruptedException e)
						{
							System.out.println("run " + e);
						}
					}
				}
			}

			//Put the incoming messages into the appropriate queues, which are sorted by time stamp
			processIncoming();

			//Apply the most recent simulation message to the LP it's aimed at
			runOneLP();
		}
	} // end run()


	/**
	 * Takes the simulation message with the lowest time stamp and passes
	 * it to the appropriate LP for processing. Messages directed to a LogicalProcess not
	 * owned by this SchedulingLoop are assumed to have been misdelivered, and
	 * are placed in <code>outgoing</code>.
	 *
	 * @return The index in the local list of the LogicalProcess which just executed
	 */
	private synchronized int runOneLP()
	{
		Message msg;

		//Process anti-messages first to prevent the spread of incorrect computation
		if (antiMessages.size() > 0)
			msg = antiMessages.poll();
		else if (simMessages.size() > 0)
			msg = simMessages.poll();
		else
			return -1;

		int lpIndex = getLPIndexByAPPid(msg.getReceiverAPPid());

		if (lpIndex == -1)
		{
			//Send message back out in hope that someone else will know what to do with it
			sendMessage(msg);
			return -1;
		}

		//Apply scheduled monitoring/steering to that LP
		applyMonitoring(lpIndex);
		applySteering(lpIndex);

		//Pass the message into the LP in question
		LogicalProcess lp = LPs.get(lpIndex);

		if (msg.getMsgType() == Message.MsgType.SIM_REGULAR)
			lp.processPositiveMessage(msg);
		else if (msg.getMsgType() == Message.MsgType.SIM_ANTI)
			lp.processAntiMessage(msg);

		return lpIndex;
	} // end runOneLP()


	/**
	 * When an LP performs a rollback, it invokes this method to pass unprocessed messages
	 * back to the scheduling loop. Note that this method is package-private because it
	 * is to only be called from the SASSY kernel.
	 *
	 * @param msg the message which was just rolled back and is now unprocessed again.
	 */
	public void returnMessage(Message msg)
	{
		simMessages.add(msg);
	}


	/**
	 * Reads incoming messages from the communication thread and, based on their
	 * {@link Message.MsgType} places them in the appropriate data structures for
	 * later processing.
	 * <ul>
	 * <li>{@link Message.MsgType#SIM_REGULAR} - placed in {@link #simMessages}.
	 * <li>{@link Message.MsgType#SYSTEM_MONITOR} - placed in {@link #monitoring}.
	 * <li>{@link Message.MsgType#SYSTEM_STEER} - placed in {@link #steering}.
	 * </ul>
	 * Messages of unrecognized MsgType are discarded. Messages directed to
	 * LogicalProcesses not owned by this SchedulingLoop are also discarded.
	 */
	private synchronized void processIncoming()
	{
		if (incoming.isEmpty())
			return;

		Iterator<Message> iterator = incoming.iterator();
		while (iterator.hasNext())
		{
			Message m = iterator.next();
			switch (m.getMsgType())
			{
				case SIM_REGULAR:
					simMessages.offer(m);
					break;
				case SIM_ANTI:
					antiMessages.offer(m);
					break;
				case SYSTEM_MONITOR:
					MonitorNode mn;
					mn = (MonitorNode)m.getContent();
					int i = getLPIndexByAPPid(m.getReceiverAPPid());
					if (i >= 0)
						monitoring.get(i).add(mn);
					else
						System.out.println("Monitoring message was directed to LP \"" +
								m.getReceiverAPPid() + "\" - not present here!");
					break;
				case SYSTEM_STEER:
					SteerNode sn;
					sn = (SteerNode)m.getContent();
					i = getLPIndexByAPPid(m.getReceiverAPPid());
					if (i >= 0)
						steering.get(i).add(sn);
					else
						System.out.println("Steering message was directed to LP \"" +
								m.getReceiverAPPid() + "\" - not present here!");
					break;
				default:
					System.out.print("Message type " + m.getMsgType() +
							" is not recognized by this SchedulingLoop! ");
					System.out.println("\"" + m + "\" will be discarded.");
					break;
			}

			//It is important to remove the element from the incoming queue only AFTER processing it
			//so there is no chance of it being missed in GVT calculation
			iterator.remove();
		}
	} // end processIncoming()


	/**
	 * Calculates the minimum timestamp among all unprocessed but sorted
	 * simulation messages and anti-messages. To ensure correctness, this
	 * method should be called from a synchronized method in PE.
	 *
	 * @returns the min timestamp of the message and anti-message queue, 
	 *			or Integer.MAX_VALUE if the queues are empty
	 */
	public synchronized int getMessageQueuesMinTS()
	{
		//Note: It is very important that this method is synchronized
		//and so are the processIncoming() and runOneLP() methods.
		//This guarantees that these three methods won't run concurrently
		//and no messages will slip through the cracks.

		int localMin = Integer.MAX_VALUE;

		//Min of all queued simulation events
		if (simMessages.peek() != null)
			localMin = Math.min(simMessages.peek().getTimeStamp(), localMin);

		//Min of all anti-messages
		if (antiMessages.peek() != null)
			localMin = Math.min(antiMessages.peek().getTimeStamp(), localMin);

		//Min of incoming & outgoing queues. This is only safe if we're called from a
		//synchronized method in PE so that the incoming/outcoming queues aren't being
		//modified

		if (!incoming.isEmpty())
		{
			Message front = Collections.min(incoming);
			localMin = Math.min(front.getTimeStamp(), localMin);
		}

		if (!outgoing.isEmpty())
		{
			Message front = Collections.min(outgoing);
			localMin = Math.min(front.getTimeStamp(), localMin);
		}


		return localMin;
	}


	/**
	 * Stops this scheduling loop from running, which in term ends the PE and
	 * the whole simulation.
	 */
	public void endSimulation()
	{
		simulationEnded = true;

		//If the scheduling loop is waiting for incoming events, tell it to stop
		synchronized(incoming)
		{
			incoming.notify();
		}

		//Stop the PE
		if (myPE.isRunning())
			myPE.endSimulation();
	}


	/**
	 * Returns true if the scheduling loop is still running, false otherwise.
	 *
	 * @return whether this scheduling loop has terminated or not
	 */
	public boolean isRunning()
	{
		return !simulationEnded;
	}


	/**
	 * Returns the last known Global Virtual Time (GVT). The GVT
	 * guarantees that no messages with timestamp less than GVT will
	 * be received. Thus, events with TS < GVT will never be rolled back.
	 *
	 * @return the global virtual time of the simulation
	 */
	public int getGVT()
	{
		return GVT;
	}


	/**
	 * Update the Global Virtual Time (GVT) . The GVT
	 * guarantees that no messages with timestamp less than GVT will
	 * be received. Thus, events with TS < GVT will never be rolled back.
	 *
	 * @param newGVT the new global virtual time in the simulation
	 */
	public void setGVT(int newGVT)
	{
		GVT = newGVT;
		shouldCollectFossils = true;
		if (TRACE)
			System.out.printf("\nGVT = %d \n", newGVT);
	}


	/**
	 * Creates a new LogicalProcess of the specified type and adds it to the list
	 * of LPs managed by this run loop.
	 *
	 * @param LPConfigInfo the information about what kind of LP should be instantiated
	 * @see LogicalProcess
	 */
	public void createLP(LPConfigInfo lpInfo)
	{
		try
		{
			LogicalProcess LP = (LogicalProcess)(Class.forName(lpInfo.getLPClassName()).newInstance());
			LP.setPID(lpInfo.PEid);
			LP.setLID(lpInfo.LPid);
			LP.setAppID(lpInfo.getAPPid());
			LP.setName(lpInfo.getAPPname());
			LP.setConfigData(lpInfo.getLPData());
			LP.setScheduler(this);

			System.out.println("PE " + myID + " created  LP : " + lpInfo.LPid);
			LPs.add(LP);
			monitoring.add(new TreeSet<MonitorNode>());
			steering.add(new TreeSet<SteerNode>());
			//record the beginning time for this LP
			LP.recordBeginTime();
		}
		catch (ClassNotFoundException e)
		{
			System.err.printf("Error: Class \"%s\" not found; could not load PE.\n", lpInfo.getLPClassName());
		}
		catch (InstantiationException e)
		{
			System.err.printf("Error: Class \"%s\" could not be instantiatied.\n", lpInfo.getLPClassName());
			e.printStackTrace(System.err);
		}
		catch (IllegalAccessException e)
		{
			System.err.printf("Error: Class \"%s\" could not be instantiatied.\n", lpInfo.getLPClassName());
			e.printStackTrace(System.err);
		}
	}


	/** Finds the list index of a given LP (position associated with it in
	 * {@link #LPs}, {@link #monitoring}, and {@link #steering}) based on its
	 * APPid.
	 * @param appid The APPid to look for (case insensitive)
	 * @return The index of the first occurrence of this LP, or -1 if none found.
	 */
	public int getLPIndexByAPPid(String appid)
	{
		if (appid == null)
			return -1;
		for (int i = 0; i < LPs.size(); i++)
		{
			LogicalProcess lp = LPs.get(i);
			if (appid.equalsIgnoreCase(lp.getAPPid()))
				return i;
		}
		return -1;
	} //end getLPIndexByAPPid()


	/**Locates one of the LPs managed by this scheduling loop, based on its APPid.
	 * @param appid The APPid to look for (case insensitive)
	 * @return The first LP found with the given APPid, or null if none found.
	 */
	public LogicalProcess getLPByAPPid(String appid)
	{
		for (LogicalProcess lp : LPs)
		{
			if (appid.equalsIgnoreCase(lp.getAPPid()))
				return lp;
		}
		return null;
	} //end getLPByAPPid()
}
