package sassy.agent;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import sassy.*;

/**
 * Agents are autonomous beings that exist within a simulation.  They
 * continuously sense their environment, think about what to do, and then act.
 * Agents can send messages to one another, and move within their environment.
 *
 * @author Cole Sherer <koleslaw54@gmail.com>
 */
public abstract class Agent extends LogicalProcess  implements Serializable
{
	/** 
	 * This class is used to save the Agent state for rollback purposes.
	 * Make sure to update appropriately when modifying an agent's members.
	 */
	private static class SavedAgentState implements Serializable
	{
		public Map<String, Agent> worldAgents;

		public int senseThinkActDelay;

		public List<Sensor> sensors;
		
		public AgentAction action;

		/** The body state, obtained by calling Body.saveBody */
		public Serializable bodyState;

		/** The head state, obtained by calling Agent.saveHeadState */
		public Serializable headState;
	}
	
	/**
	 * All of the agents that can potentially be sensed by one of the agent's
	 * sensors.  In simulations with one IMLP, this is every agent in the
	 * simulation, otherwise it is all agents in the IMLPs that are subscribed
	 * to.
	 */
	private Map<String, Agent> worldAgents = new HashMap<String, Agent>();

	/** The amount of simulation time that passes between sense-think-act cycles.  */
	protected int senseThinkActDelay;

	/**
	 * The list of sensors that this agent will use to locate objects in
	 * the environment
	 */
	protected List<Sensor> sensors = new LinkedList<Sensor>();

	/** The agent's physical body within the simulation environment. */
	protected Body body;

	/**
	 * The action that the agent will take at the end of the current sense cycle
	 */
	protected AgentAction action;


	/**
	 * Create an Agent with no events scheduled.  The delay between
	 * sense-think-act cycles will be set to 1.
	 */
	public Agent() {
		super();
		senseThinkActDelay = 1;
	}


	/**
	 * Create an Agent with no events scheduled, and the given delay between
	 * sense-think-act cycles.
	 *
	 * @param delay the amount of time that passes between sense-think-act
	 * cycles
	 */
	public Agent(int delay) {
		super();
		senseThinkActDelay = delay;
	}


	/**
	 * Invoked by the Environment when the agent is first introduced into
	 * the simulation
	 */
	public abstract void init();

	
	/**
	 * Invoked by the Environment immediately after the agent is initialized.
	 * The application specific configuration data for this agent is passed in
	 * as a String.
	 *
	 * @param configData the configuration data passed in from a file
	 */
	public abstract void readConfig(Serializable configData);


	/**
	 * Sends a message to the provided agent
	 *
	 * @param content the message content
	 * @param timeInc the time increment assigned by the sender.  The send time
	 *					of the message is local time + timeInc
	 * @param targetId the id of the agent the message is intended for
	 */
	public final void sendMessage(Serializable content, int timeInc, String targetId) {
		AgentMessage msg = new AgentMessage(AgentMessage.MessageType.MESSAGE, content);
		sendSimMessage(msg, timeInc, targetId);
	}


	/**
	 * Invoked when a new message for this agent is received.
	 *
	 * @param content the message content
	 * @param senderId the id of the sending agent
	 */
	public abstract void rcvMessage(Object content, String senderId);

	/** 
	 * The agent must implement this method to support rollback. The
	 * object returned should contain all the state used in the agent thinking logic
	 * (no need to save body state or implementation details).
	 */
	public abstract Serializable saveHeadState();
	
	/**
	 * The agent must implement this method to support rollback. The object passed in
	 * to this method is obtained by calling saveHeadState(). This method should restore
	 * the thinking logic to its previous state.
	 */
	public abstract void restoreHeadState(Serializable savedHead);
	
	/**
	 * The agent will sense objects in its environment and return them.
	 * Calling this function multiple times performs the computation multiple times.
	 */
	public Map<String, Agent> sense()
	{
		Map<String,Agent> sensedAgents = new HashMap<String,Agent>();
		
		for(Sensor s : sensors) {
			for(String id : worldAgents.keySet()) {
				Agent agt = worldAgents.get(id);
				if(s.canSense(agt.getBody()))
					sensedAgents.put(agt.getAPPid(), agt);
			}
		}
		
		return sensedAgents;
	}


	/**
	 * The think cycle is called repeatedly after the agent senses its
	 * surroundings and immediately before acting.  All of the agent's actions
	 * should be determined in this method.
	 */
	public abstract void think();


	/**
	 * All action commands will be sent to the body
	 */
	public final void act() {
		if(action != null) {
			body.takeAction(action);
		}
		action = null;
	}


	/**
	 * Invoked by the kernel at the beginning of the simulation when the LP
	 * should initialize itself and send any initial messages.
	 */
	@Override
	public final void appInitializeLP() {
		init();
		sendSimMessage(new AgentMessage(AgentMessage.MessageType.ACT, null),
						senseThinkActDelay, getAPPid());
	}


	/**
	 * Invoked when a new message destined for this logical process
	 * is received.
	 * @param content The object associated with the new event.
	 * @param sender The application id of the LP the event came from.
	 */
	@Override
	public final void appRunLP(Object content, String sender) {
		AgentMessage msg = (AgentMessage)content;

		switch(msg.getType()) {
			case ACT:
				think();
				act();
				sendSimMessage(msg, senseThinkActDelay, sender);
				break;
			case MESSAGE:
				rcvMessage(msg.getContent(), sender);
				break;
			case UPDATE:
				Agent agt = (Agent)(msg.getContent());
				worldAgents.put(agt.getAPPid(), agt);
				break;
			default:
				break;
		}
	}


	/**
	 * Invoked by the kernel when the Agent is being initialized.
	 * The application-specific text from the configuration file is passed in.
	 * The Agent should parse it and set up its internal state accordingly.
	 *
	 * @param cdata String representing configuration data.
	 */
	@Override
	public final void appSetConfigData(Serializable cdata) {
		readConfig(cdata);
	}


	
	/** 
	 * Here we save the full agent state, so we can restore it later. 
	 * We call abstract methods in the body and the agent, so that subclasses have
	 * a chance to save their state.
	 */
	@Override
	public final Serializable appSaveLPState()
	{
		SavedAgentState savedState = new SavedAgentState();
		savedState.worldAgents = new HashMap<String, Agent>(worldAgents);
		savedState.senseThinkActDelay = senseThinkActDelay;
		//FIXME: Does the sensor class have state? Do we need to save it too?
		savedState.sensors = new ArrayList<Sensor>(sensors);
		
		if (action != null)
		{
			savedState.action = new AgentAction(action);
		}
		savedState.bodyState = body.saveBody();
		savedState.headState = saveHeadState();
		
		return savedState;
	}



	/** 
	 * Here we restore the full agent state from a saved snapshot.
	 * We have to restore the head, the body, and the implementation structures.
	 */
	@Override
	public final void appSetLPState(Serializable saved)
	{
		SavedAgentState savedAgent = (SavedAgentState)saved;
		worldAgents = savedAgent.worldAgents;
		senseThinkActDelay = savedAgent.senseThinkActDelay;
		sensors = savedAgent.sensors;
		action = savedAgent.action;
		body.restoreBody(savedAgent.bodyState);
		restoreHeadState(savedAgent.headState);
	}


	/**
	 * Invoked by the kernel at the end of the simulation.
	 */
	@Override
	public void appFinalizeLP() {}


	/**
	 * Get the agent's body
	 *
	 * @return the agent's body
	 */
	public Body getBody() {
		return body;
	}
}
