package sassy.agent;

import java.io.Serializable;
import java.util.LinkedList;
import java.util.List;

/**
 * The physical representation of an Agent within a simulation
 *
 * @author Cole Sherer <koleslaw54@gmail.com>
 */
public abstract class Body 
{
	/** The BodySegments that make up this body. */
	protected List<BodySegment> segments;


	/**
	 * Creates a Body with no body segments
	 */
	public Body() 
	{
		segments = new LinkedList<BodySegment>();
	}


	/**
	 * Creates a new Body that is a copy of the given Body
	 *
	 * @param body the Body to copy
	 */
	public Body(Body body) 
	{
		segments = new LinkedList<BodySegment>(body.getSegments());
	}


	/**
	 * The agent will handle the requested action in the best manner possible.
	 * If the agent is incapable of moving or rotating by the given amounts,
	 * it will make a best effort.
	 *
	 * @param action The action that the Agent specified in its "think" cycle
	 */
	public abstract void takeAction(AgentAction action);


	/**
	 * Add a BodySegment to this body
	 *
	 * @param segment the BodySegment to add
	 */
	public abstract void addSegment(BodySegment segment);

	/** 
	 * This function must be implemented by bodies in order to support rollback.
	 * It should return an object that holds the entire state of the body. 
	 */
	public abstract Serializable saveBody();
	
	/** 
	 * This function must be implemented in order to support rollback.
	 * It is given an object previously returned by Body.addSegment and it should
	 * restore the state of the body to that state.
	 */
	public abstract void restoreBody(Serializable savedState);

	/**
	 * Get the list of body segments for this body
	 *
	 * @return the list of body segments
	 */
	public List<BodySegment> getSegments() 
	{
		return segments;
	}
}
