package sassy.agent;

import java.io.Serializable;


/**
 * A Body designed to be used in a 2D environment
 * 
 * @author Cole Sherer <koleslaw54@gmail.com>
 */
public class Body2D extends Body 
{
	/** 
	 * Used in state saving, to support rollback. Make sure to update the state saving code
	 * if you modify the members of this class.
	 */
	private final static class SavedBodyState implements Serializable
	{
		public double x;
		public double y;
		public double theta;
		public double radius;
	}
	
	/**
	 * The location of this Body
	 */
	private double x;
	private double y;

	/**
	 * The direction the Body is facing (between 0 and 2π)
	 */
	private double theta;

	/**
	 * The maximum distance from the center of the Body to the outer edge of
	 * any of its segments
	 */
	private double radius;


	/**
	 * Create a Body centered at (0,0), facing right and with radius 0.0
	 */
	public Body2D() {
		x = 0.0;
		y = 0.0;
		theta = 0.0;
		radius = 0.0;
	}


	/**
	 * Create a Body centered at the given (x,y) location, facing the given
	 * direction
	 *
	 * @param x The x-location of the Body's center
	 * @param y The y-location of the Body's center
	 * @param theta The direction the Body is facing
	 */
	public Body2D(double x, double y, double theta) {
		this.x = x;
		this.y = y;
		this.theta = theta;
	}


	/**
	 * Get the x-coordinate of the Body's location
	 *
	 * @return the x-position
	 */
	public double getX() {
		return x;
	}


	/**
	 * Set the x-coordinate of the Body's location
	 *
	 * @param x the x-position
	 */
	public void setX(double x) {
		this.x = x;
	}


	/**
	 * Get the y-coordinate of the Body's location
	 *
	 * @return the y-position
	 */
	public double getY() {
		return y;
	}


	/**
	 * Set the y-coordinate of the Body's location
	 *
	 * @param y the y-position
	 */
	public void setY(double y) {
		this.y = y;
	}


	/**
	 * Get the direction this Body is facing
	 *
	 * @return the direction
	 */
	public double getTheta() {
		return theta;
	}


	/**
	 * Set the direction the Body is facing
	 *
	 * @param theta the direction
	 */
	public void setTheta(double theta) {
		this.theta = theta;
	}


	/**
	 * Get the radius of the Body
	 *
	 * @return the radius
	 */
	public double getRadius() {
		return radius;
	}


	/**
	 * Add a segment to the Body.
	 *
	 * @param segment the body segment
	 */
	@Override
	public void addSegment(BodySegment segment) {
		segments.add(segment);

		double newRadius = 0.0;

		if(segment instanceof CircleSegment) {
			CircleSegment s = (CircleSegment)segment;
			newRadius = Math.sqrt(Math.pow(s.getDeltaX(), 2) +
								  Math.pow(s.getDeltaY(), 2)) +
						  		  s.getRadius();
		}
		else if(segment instanceof RectangleSegment) {
			RectangleSegment r = (RectangleSegment) segment;
			newRadius = Math.sqrt(Math.pow(r.getDeltaX(), 2) +
								  Math.pow(r.getDeltaY(), 2)) +
						Math.sqrt(Math.pow(r.getWidth() / 2 , 2) +
								  Math.pow(r.getHeight() / 2, 2));
		}
		
		if(newRadius > radius) {
			radius = newRadius;
		}
	}


	/**
	 * Move the body the given amount in the direction it is facing
	 *
	 * @param speed the distance to move
	 */
	public void move(double speed) {
		x += Math.cos(theta) * speed;
		y += Math.sin(theta) * speed;
	}


	/**
	 * Rotate the body's heading by the given amount
	 *
	 * @param angle the amount to rotate by
	 */
	public void rotate(double angle) {
		theta += angle;
		if(theta > 2 * Math.PI)
			theta -= (2 * Math.PI);
	}


	/**
	 * Take the given AgentAction
	 *
	 * @param action the action to take
	 */
	@Override
	public void takeAction(AgentAction action) {
		rotate(action.getTheta());
		move(action.getSpeed());
	}


	@Override
	public Serializable saveBody()
	{
		SavedBodyState savedBody = new SavedBodyState();
		savedBody.x = x;
		savedBody.y = y;
		savedBody.theta = theta;
		savedBody.radius = radius;
		
		return savedBody;
	}


	@Override
	public void restoreBody(Serializable savedState)
	{
		SavedBodyState savedBody = (SavedBodyState)savedState;
		x = savedBody.x;
		y = savedBody.y;
		theta = savedBody.theta;
		radius = savedBody.radius;
	}
}
