package sassy.agent;

import java.io.Serializable;

/**
 * A Body designed to be used in a 3D environment
 *
 * @author Cole Sherer <koleslaw54@gmail.com>
 */
public class Body3D extends Body 
{
	/** 
	 * Used in state saving, to support rollback. Make sure to update the state saving code
	 * if you modify the members of this class.
	 */
	private final static class SavedBodyState implements Serializable
	{
		public double x;
		public double y;
		public double z;
		public double theta;
		public double phi;
		public double radius;
	}
	
	/**
	 * The (x,y,z) location of the Body's center
	 */
	private double x;
	private double y;
	private double z;

	/**
	 * The maximum distance from the center of the Body to the outer edge of
	 * any of its segments
	 */
	private double radius;

	/**
	 * The direction the body is facing  the xy-plane
	 */
	private double theta;

	/**
	 * The direction the body is facing between the xy-plane and the z-axis
	 */
	private double phi;


	/**
	 * Create a Body with a center at (0,0,0) lying in the xy-plane and facing
	 * the positive x direction.
	 */
	public Body3D() {
		x = 0.0;
		y = 0.0;
		z = 0.0;
		theta = 0.0;
		phi = 0.0;
	}


	/**
	 * Create a body at the given (x,y,z) location, facing the theta direction
	 * within the xy-plane, and facing the phi direction between the xy-plane
	 * and the z-axis
	 *
	 * @param x The x-location of the Body's center
	 * @param y The y-location of the Body's center
	 * @param z The z-location of the Body's center
	 * @param theta The direction the Body is facing in the xy-plane
	 * @param phi The direction the Body is facing between the xy-plane and the
	 * z-axis
	 */
	public Body3D(double x, double y, double z, double theta, double phi) {
		this.x = x;
		this.y = y;
		this.z = z;
		this.theta = theta;
		this.phi = phi;
	}


	/**
	 * Get the x-coordinate of the Body's center
	 *
	 * @return the x-coordinate
	 */
	public double getX() {
		return x;
	}


	/**
	 * Set the x-coordinate of the Body's center
	 *
	 * @param x the x-coordinate
	 */
	public void setX(double x) {
		this.x = x;
	}


	/**
	 * Get the y-coordinate of the Body's center
	 *
	 * @return the y-coordinate
	 */
	public double getY() {
		return y;
	}


	/**
	 * Set the y-coordinate of the Body's center
	 *
	 * @param y the y-coordinate
	 */
	public void setY(double y) {
		this.y = y;
	}


	/**
	 * Get the z-coordinate of the Body's center
	 *
	 * @return the z-coordinate
	 */
	public double getZ() {
		return z;
	}

	/**
	 * Set the x-coordinate of the Body's center
	 *
	 * @param z the z-coordinate
	 */
	public void setZ(double z) {
		this.z = z;
	}


	/**
	 * Get the direction the Body is facing in the xy-plane
	 *
	 * @return the direction
	 */
	public double getTheta() {
		return theta;
	}


	/**
	 * Set the direction the Body is facing in the xy-plane
	 *
	 * @return the direction
	 */
	public void setTheta(double theta) {
		this.theta = theta;
	}


	/**
	 * Get the direction the Body is facing between the xy-plane and the z-axis
	 *
	 * @return the direction
	 */
	public double getPhi() {
		return phi;
	}


	/**
	 * Set the direction the Body is facing between the xy-plane and the z-axis
	 *
	 * @return the direction
	 */
	public void setPhi(double phi) {
		this.phi = phi;
	}


	/**
	 * Get the Body's radius
	 *
	 * @return the radius
	 */
	public double getRadius() {
		return radius;
	}


	/**
	 * Add a BodySegment to the Body
	 *
	 * @param segment the BodySegment
	 */
	@Override
	public void addSegment(BodySegment segment) {
		segments.add(segment);

		double newRadius = 0.0;

		if(segment instanceof SphereSegment) {
			SphereSegment s = (SphereSegment)segment;
			newRadius = Math.sqrt(Math.pow(s.getDeltaX(), 2) +
								  Math.pow(s.getDeltaY(), 2) +
								  Math.pow(s.getDeltaZ(), 2)) + s.getRadius();
		}
		else if(segment instanceof BoxSegment) {
			BoxSegment s = (BoxSegment)segment;
			newRadius = Math.sqrt(Math.pow(s.getDeltaX(), 2) +
								  Math.pow(s.getDeltaY(), 2) +
								  Math.pow(s.getDeltaZ(), 2)) +
						Math.sqrt(Math.pow(s.getHeight() / 2, 2) +
								   Math.pow(s.getWidth() / 2, 2) +
								   Math.pow(s.getDepth() / 2, 2));
		}

		if(newRadius > radius) {
			radius = newRadius;
		}
	}


	/**
	 * Move the given distance in the current direction
	 *
	 * @param speed the distance to move
	 */
	public void move(double speed) {
		z += Math.sin(phi) * speed;

		double xyHypotenuse = Math.cos(phi) * speed;
		x += Math.cos(theta) * xyHypotenuse;
		y += Math.sin(theta) * xyHypotenuse;

	}


	/**
	 * Rotate by the given amounts in each plane
	 *
	 * @param theta the amount to rotate in the xy-plane
	 * @param phi the amount to rotate between the xy-plane and the z-axis
	 */
	public void rotate(double theta, double phi) {
		this.theta += theta;
		if(this.theta > 2 * Math.PI)
			this.theta -= (2 * Math.PI);

		this.phi += phi;
		if(this.phi > 2 * Math.PI)
			this.phi -= (2 * Math.PI);
	}


	/**
	 * Take the given AgentAction
	 *
	 * @param action the action to take
	 */
	@Override
	public void takeAction(AgentAction action) {
		rotate(action.getTheta(), action.getPhi());
		move(action.getSpeed());
	}


	@Override
	public Serializable saveBody()
	{
		SavedBodyState savedBody = new SavedBodyState();
		savedBody.x = x;
		savedBody.y = y;
		savedBody.z = z;
		savedBody.theta = theta;
		savedBody.phi = phi;
		savedBody.radius = radius;
		
		return savedBody;
	}


	@Override
	public void restoreBody(Serializable savedState)
	{
		SavedBodyState savedBody = (SavedBodyState)savedState;
		x = savedBody.x;
		y = savedBody.y;
		z = savedBody.z;
		theta = savedBody.theta;
		phi = savedBody.phi;
		radius = savedBody.radius;
	}
}
