package sassy.agent;

/**
 * A <code>BodySegment</code> that represents a standard rectangular prism.
 * This is designed to be used in 3D simulations.
 *
 * @author Cole Sherer <koleslaw54@gmail.com>
 */
public class BoxSegment extends BodySegment {

	/**
	 * The distance between the parent Body's location and the box's center
	 */
	private double deltaX;
	private double deltaY;
	private double deltaZ;

	private double width;
	private double height;
	private double depth;

	private double theta;
	private double phi;


	/**
	 * Create a BoxSegment with the same location as its parent Body, and
	 * dimensions of 1.0 x 1.0
	 */
	public BoxSegment() {
		deltaX = deltaY = deltaZ = 0.0;
		width = height = depth = 1.0;
		theta = phi = 0.0;
	}


	/**
	 * Create a BoxSegment that is the provided distance away from its
	 * parent Body with the given dimensions
	 *
	 * @param x The distance between the parent Body's x-coordinate and the
	 * segment's x-coordinate
	 * @param y The distance between the parent Body's y-coordinate and the
	 * segment's y x-coordinate
	 * @param w The width of the box
	 * @param h The height of the box
	 * @param t The rotation of the box in the (x,y) plane
	 * @param p The rotation of the box in the (y,z) plane
	 */
	public BoxSegment(double x, double y, double w, double h, double t,
					  double p) {
		deltaX = x;
		deltaY = y;
		width = w;
		height = h;
		theta = t;
		phi = p;
	}


	/**
	 * Set the distance between the parent Body's x-coordinate and the Segment's
	 * x-coordinate
	 *
	 * @param x the distance
	 */
	public void setDeltaX(double x) {
		deltaX = x;
	}


	/**
	 * Set the distance between the parent Body's y-coordinate and the Segment's
	 * y-coordinate
	 *
	 * @param y the distance
	 */
	public void setDeltaY(double y) {
		deltaY = y;
	}


	/**
	 * Set the distance between the parent Body's z-coordinate and the Segment's
	 * z-coordinate
	 *
	 * @param z the distance
	 */
	public void setDeltaZ(double z) {
		deltaZ = z;
	}


	/**
	 * Set the width of the BoxSegment
	 *
	 * @param w the width
	 */
	public void setWidth(double w) {
		width = w;
	}


	/**
	 * Set the height of the BoxSegment
	 *
	 * @param h the height
	 */
	public void setHeight(double h) {
		height = h;
	}


	/**
	 * Set the depth of the BoxSegment
	 *
	 * @param d the depth
	 */
	public void setDepth(double d) {
		depth = d;
	}


	/**
	 * Set the rotation of the BoxSegment
	 *
	 * @param a the angle of rotation
	 */
	public void setRotation(double t, double p) {
		theta = t;
		phi = p;
	}


	/**
	 * Get the distance between the parent Body's x-coordinate and the Segment's
	 * x-coordinate
	 *
	 * @return the distance
	 */
	public double getDeltaX() {
		return deltaX;
	}


	/**
	 * Get the distance between the parent Body's y-coordinate and the Segment's
	 * y-coordinate
	 *
	 * @return the distance
	 */
	public double getDeltaY() {
		return deltaY;
	}


	/**
	 * Get the distance between the parent Body's y-coordinate and the Segment's
	 * z-coordinate
	 *
	 * @return the distance
	 */
	public double getDeltaZ() {
		return deltaZ;
	}


	/**
	 * Get the width of the BoxSegment
	 *
	 * @return the width
	 */
	public double getWidth() {
		return width;
	}


	/**
	 * Get the height of the BoxSegment
	 *
	 * @return the height
	 */
	public double getHeight() {
		return height;
	}


	/**
	 * Get the height of the BoxSegment
	 *
	 * @return the height
	 */
	public double getDepth() {
		return depth;
	}


	/**
	 * Get the rotation of the BoxSegment
	 *
	 * @return the rotation
	 */
	public double getTheta() {
		return theta;
	}


	/**
	 * Get the rotation of the BoxSegment
	 *
	 * @return the rotation
	 */
	public double getPhi() {
		return phi;
	}
}