package sassy.agent;

import java.io.Serializable;
import java.util.HashMap;
import sassy.LogicalProcess;

/**
 * Grid based interest manager cells representing world state for agent based
 * simulator.  Each IMLP manages a publisher list and a subscriber list. The
 * interest manager will receive messages from publishers and reflect these
 * messages to subscribers.
 *
 * @author Cole Sherer <koleslaw54@gmail.com>
 * @author Tianhao He
 */
public class IMLP extends LogicalProcess implements Serializable
{

	private static final long serialVersionUID = 52323432L;

	// Whether we're currently debugging an IMLP or not
	private static final boolean DEBUG = true;

	/** subscriber list */
	private HashMap<String, Agent> subscriberList;

	/** publisher list */
	private HashMap<String, Agent> publisherList;

	/** the time increment for messages sent by this LP; the message will
	 *  be sent at currentTime+timeIncrement */
	private int timeIncrement = 0;
	

	public IMLP()
	{
		super(false);
		publisherList = new HashMap<String, Agent>();
		subscriberList = new HashMap<String, Agent>();
	}


	/**
	 * <p>Invoked by a subscribe message from the subscribing agent.
	 *    Adds the agent to the subscriber list</p>
	 * @param agt the agent who subscribed to this IMLP
	 */
	private void subscribe(Agent agt)
	{
		if(DEBUG) {
			printRollbackSafe(this.getName() + ": subscribed: " + agt.getAPPid()
							  + " : " + this.getCurrentTime());
		}


		// Add to subscriber list
		subscriberList.put(agt.getAPPid(), agt);
	}


	/**
	 * <p>Invoked by an unsubscribe message from the unsubscribe agent.
	 *    Removes the agent from the subscriber list</p>
	 * @param agt the agent who unsubscribed from this IMLP
	 */
	private void unsubscribe(Agent agt)
	{
		if(DEBUG) {
			printRollbackSafe(this.getName() + ": unsub: " + agt.getAPPid() +
							  " : " + this.getCurrentTime());
		}

		//remove from subscriber list
		if (subscriberList.containsKey(agt.getAPPid()))
		{
			subscriberList.remove(agt.getAPPid());
		}
		else if(DEBUG)
		{
			printRollbackSafe(this.getName() + ": the unsub ball is not in the"
					          + "sub list: " + agt.getAPPid() + " : " +
							  this.getCurrentTime());
		}
	}


	/**
	 * <p>Invoked by an UPDATE message from a publisher.  This message should
	 * be forwarded to all subscribing agents.</p>
	 * @param agt the agent who sends the update message
	 */
	private void update(Agent agt)
	{
		// Update our local copy of the environment state
		publisherList.put(agt.getAPPid(), agt);

		// Relay the update message to all subscribers
		for (String id : subscriberList.keySet())
		{
			if (!id.equals(agt.getAPPid()))
				sendSimMessage(new AgentMessage(AgentMessage.MessageType.UPDATE,
						       agt), timeIncrement, id);
		}
	}


	/**
	 * Invoked by an ENTER message from the agent who just entered this IMLP,
	 * the agent becomes a publisher of this IMLP.
	 * Adds the agent to the publisher list and inform subscribers by sending an
	 * UPDATE message.
	 * @param agt the agent enters this IMLP
	 */
	private void enter(Agent agt)
	{
		//Add the entry to the list of publishers
		publisherList.put(agt.getAPPid(), agt);

		// Send an update message to all subscribers
		for (String id : subscriberList.keySet())
		{
			if (!id.equals(agt.getAPPid()))
				sendSimMessage(new AgentMessage(AgentMessage.MessageType.UPDATE,
						       agt), timeIncrement, id);
		}
	}


	/**
	 * Invoked by a LEAVE message from an agent. Removes the agent from the
	 * publisher list.
	 * @param agt the agent leaving this IMLP
	 */
	private void leave(Agent agt)
	{
		if (!publisherList.containsKey(agt.getAPPid()))
		{
			if(DEBUG)
				printRollbackSafe(this.getName() + ": the leaving agent is not"
								  + "in the publisher list: " + agt.getAPPid() +
								  " : " + this.getCurrentTime());
			return;
		}

		publisherList.remove(agt.getAPPid());
	}


	@Override
	public void appFinalizeLP()
	{
	}


	@Override
	public void appInitializeLP()
	{
	}


	@Override
	public void appRunLP(Object content, String sender)
	{
		if(DEBUG) {
			printRollbackSafe(sender + " sent to IMLP: " + getAPPid());
		}

		// get message content and check the type of the message
		AgentMessage message = (AgentMessage)content;
		AgentMessage.MessageType type = message.getType();

		Agent agt = (Agent)(message.getContent());

		switch (type)
		{
			case UPDATE:
				update(agt);
				break;
			case SUBSCRIBE:
				subscribe(agt);
				break;
			case UNSUBSCRIBE:
				unsubscribe(agt);
				break;
			case ENTER:
				enter(agt);
				break;
			case LEAVE:
				leave(agt);
				break;
			default:
				if(DEBUG)
					printRollbackSafe("Unknown message type in IMLP");
				break;
		}
	}


	@Override
	public void appSetConfigData(Serializable cdata)
	{

	}


	@Override
	public Serializable appSaveLPState()
	{
		IMLP imlp = new IMLP();
		imlp.subscriberList = new HashMap<String, Agent>(subscriberList);
		imlp.publisherList = new HashMap<String, Agent>(publisherList);
		return imlp;
	}


	@Override
	public void appSetLPState(Serializable lp)
	{
		IMLP imlp = (IMLP)lp;
		this.publisherList = imlp.publisherList;
		this.subscriberList = imlp.subscriberList;
	}


	private void printRollbackSafe(final String msg)
	{
		submitIrreversibleAction(new Runnable()
		{
			@Override
			public void run()
			{
				System.out.println(msg);
			}
		});
	}
}
