package sassy.agent;

import java.util.LinkedList;
import java.util.List;

/**
 * A <code>Sensor</code> used to detect objects in a 2D simulation
 * 
 * @author Cole Sherer <koleslaw54@gmail.com>
 */
public class Sensor2D extends Sensor {

	private double x, y, radius, thetaStart, thetaEnd;


	/**
	 * Creates a Sensor located at (0,0), that can sense objects anywhere
	 * within a circle of radius 1.0
	 */
	public Sensor2D() {
		x = 0.0;
		y = 0.0;
		radius = 1.0;
		thetaStart = 0.0;
		thetaEnd = 2 * Math.PI;
	}


	/**
	 * Create a Sensor that is located at the given (x,y) location, that can
	 * sense objects that are within the given radius, between theta1 and theta2
	 *
	 * @param x The x-location of the Sensor
	 * @param y The y-location of the Sensor
	 * @param radius The distance from its center that the sensor can "sense"
	 * @param theta1 The starting direction of the sensing region
	 * @param theta2 The ending direction of the sensing region
	 */
	public Sensor2D(double x, double y, double radius, double theta1, double theta2) {
		this.x = x;
		this.y = y;
		this.radius = radius;
		this.thetaStart = theta1;
		this.thetaEnd = theta2;
	}

	
	/**
	 * Determines whether or not the Sensor can sense the given Body.  This
	 * method is automatically called during the "sense" phase of an Agent's
	 * sense-think-act cycle.
	 *
	 * TODO: Check this math.  The xDiff/yDiff may be subtracted in reverse order
	 *
	 * @param body the Body to test against
	 * @return true if the Body can be sensed, false otherwise
	 */
	@Override
	public boolean canSense(Body body) {
		Body2D agtBody = (Body2D)body;

		double xDiff = agtBody.getX() - x;
		double yDiff = agtBody.getY() - y;
		double radii = agtBody.getRadius() + radius;
		double theta = Math.atan2(xDiff, yDiff);

		// If the body is within our sensing range, try to sense
		// one of its segments
		if(xDiff * xDiff + yDiff * yDiff < radii * radii &&
			theta >= thetaStart && theta <= thetaEnd){

			List<BodySegment> segments = body.getSegments();

			for(BodySegment bs : segments) {

				if(bs instanceof CircleSegment) {
					CircleSegment s = (CircleSegment)bs;
					xDiff = agtBody.getX() + s.getDeltaX() - x;
					yDiff = agtBody.getY() + s.getDeltaY() - y;
					radii = radius + s.getRadius();
					theta = Math.atan2(xDiff, yDiff);

					if(xDiff * xDiff + yDiff * yDiff < radii * radii &&
					   theta >= thetaStart && theta <= thetaEnd)
						return true;
				}
				else if(bs instanceof RectangleSegment) {

					RectangleSegment r = (RectangleSegment)bs;
					double boxRadius = Math.sqrt(Math.pow(r.getHeight() / 2, 2)
											   + Math.pow(r.getWidth() / 2, 2));
					double cosDist = boxRadius * Math.cos(r.getRotation());
					double sinDist = boxRadius * Math.sin(r.getRotation());

					double bx = agtBody.getX() + r.getDeltaX();
					double by = agtBody.getY() + r.getDeltaY();
					double xs[] = {bx + cosDist, bx - sinDist, bx - cosDist,
								   bx + sinDist, bx + cosDist};
					double ys[] = {by + sinDist, by + cosDist, by - sinDist,
								   by - cosDist, by + sinDist};

					for(int i=0; i<4; i++) {
						xs[i] += bx;
						ys[i] += by;
						
						if(Math.abs((xs[i + 1] - xs[i]) * (ys[i] + y)
								- (xs[i] - x) * (ys[i+1] - ys[i])) /
								Math.sqrt(Math.pow((xs[i+1] - xs[i]), 2) +
								Math.pow((ys[i+1] - ys[i]), 2)) < radius) {

							// TODO: This only works for full circles
							return true;
						}
					}
				}
			}
		}

		return false;
	}
}
