package sassy.agent;

import java.util.LinkedList;
import java.util.List;

/**
 * A <code>Sensor</code> used to detect objects in a 3D simulation
 * 
 * @author Cole Sherer <koleslaw54@gmail.com>
 */
public class Sensor3D extends Sensor {

	private double x, y, z, radius, thetaStart, thetaEnd, phiStart, phiEnd;


	/**
	 * Create a Sensor that is located at the given (x,y,z) location, that can
	 * sense objects located within a sphere of radius 1.0
	 */
	public Sensor3D() {
		x = y = z = 0.0;
		radius = 1.0;
		thetaStart = phiStart = 0.0;
		thetaEnd = phiEnd = Math.PI * 2.0;
	}


	/**
	 * Create a Sensor that is located at the given (x,y,z) location, that can
	 * sense objects that are within the given radius, between thetaStart and
	 * thetaEnd on the xy-plane, and within phiStart and phiEnd between the
	 * xy-plane and the z-axis
	 *
	 * @param x The x-location of the Sensor
	 * @param y The y-location of the Sensor
	 * @param z The z-location of the Sensor
	 * @param radius The distance from its center that the sensor can "sense"
	 * @param thetaStart The starting direction of the sensing region in the
	 * xy-plane
	 * @param thetaEnd The ending direction of the sensing region in the
	 * xy-plane
	 * @param phiStart The starting direction of the sensing region between the
	 * xy-plane and the z-axis
	 * @param phiEnd The ending direction of the sensing region between the
	 * xy-plane and the z-axis
	 */
	public Sensor3D(double x, double y, double z, double radius,
					double thetaStart, double thetaEnd, double phiStart,
					double phiEnd) {

		this.x = x;
		this.y = y;
		this.z = z;
		this.radius = radius;
		this.thetaStart = thetaStart;
		this.thetaEnd = thetaEnd;
		this.phiStart = phiStart;
		this.phiEnd = phiEnd;
	}

	
	/**
	 * Determines whether or not the Sensor can sense the given Body.  This
	 * method is automatically called during the "sense" phase of an Agent's
	 * sense-think-act cycle.
	 *
	 * @param body the Body to test against
	 * @return true if the Body can be sensed, false otherwise
	 */
	@Override
	public boolean canSense(Body body) {
		Body3D agtBody = (Body3D)body;

		double xDiff = agtBody.getX() - x;
		double yDiff = agtBody.getY() - y;
		double zDiff = agtBody.getZ() - z;
		double xyLength = xDiff * xDiff + yDiff * yDiff;
		double zDiffSquared = zDiff * zDiff;
		double radii = agtBody.getRadius() + radius;
		
		double theta = Math.atan2(xDiff, yDiff);
		double phi = Math.atan2(xyLength, zDiffSquared);

		if(xyLength + zDiffSquared < radii * radii &&
			theta >= thetaStart && theta <= thetaEnd &&
			phi >= phiStart && phi <= phiEnd){

			List<BodySegment> segments = body.getSegments();

			for(BodySegment bs : segments) {

				if(bs instanceof SphereSegment) {
					SphereSegment s = (SphereSegment)bs;
					xDiff = agtBody.getX() + s.getDeltaX() - x;
					yDiff = agtBody.getY() + s.getDeltaY() - y;
					zDiff = agtBody.getZ() + s.getDeltaZ() - z;
					xyLength = xDiff * xDiff + yDiff * yDiff;
					zDiffSquared = zDiff * zDiff;
					radii = radius + s.getRadius();
					theta = Math.atan2(xDiff, yDiff);
					phi = Math.atan2(xyLength, zDiffSquared);

					if(xyLength + zDiffSquared < radii * radii &&
					   theta >= thetaStart && theta <= thetaEnd &&
					   phi >= phiStart && phi <= phiEnd) {
						return true;
					}
				}
				else if(bs instanceof BoxSegment) {

					// TODO: Calculate Sphere-Box Collision
					return true;
				}
			}
		}
		return false;
	}
}
