package sassy.monitoring_steering;

import java.io.*;

/** Data structure representing a single monitoring command.
 * <p>
 * This version only supports ongoing monitoring, not recurring monitoring. 
 * To perform one-time-only monitoring, set startTime = stopTime.
 *
 * @version 2.0 (02/2007)
 * @author Yin xiong
 * @author Glenn Matthews
 */

public class MonitorNode implements Serializable, Comparable<MonitorNode>
{

	/**The flag for displaying debugging messages for monitoring and steering*/
	public boolean MTRACE = false;

	/** The system-assigned unique ID of the client who requested the monitoring*/
	public String clientID;

	/**The client-assigned id of this monitoring message*/
	public String msgID;

	/**The type of the variable to be monitored: V=single variable, L=all monitorable for an LP*/
	public String varType;

	/** The APPid of the target LP to be monitored*/
	public int targetID;

	/** The name of the variable to be monitored. */
	public String varName;

	/** The simulation time at which this monitoring command becomes relevant.
	  * 0 = now*/
	public int startTime;

	/** The simulation time span this monitoring command will last;0=to the end*/
	public int timeSpan;

	/** The simulation time when the monitoring action stops;10000=end*/
	public int stopTime;

	/** Basic constructor. */
	public MonitorNode(String cid, String mid, int tid, String vtype, int btime,
			   int tspan, String vname)
	{
		clientID = cid;
		msgID = mid;
		targetID = tid;
		varType = vtype;
		varName = vname;
		startTime = btime;
		timeSpan = tspan;
		if (timeSpan == 0)
			stopTime = 10000;
		else
			stopTime = startTime + timeSpan;

		if (MTRACE)
		{
			System.out.println("\nin MonitorNode constructor 1: " );
			System.out.println("client id" + clientID);
			System.out.println("message id" + msgID);
			System.out.println("target id: " + targetID);
			System.out.println("variable type " + varType);
			System.out.println("variable name: " + varName);
			System.out.println("start time: " + startTime);
			System.out.println("time span: " + timeSpan);
		} //end debugging

	} //end basic constructor

	/**Returns the class members and their values as a string*/
	public String stringValue()
	{
		String s = "clientID: " + clientID + "\n" +
			   "msgID: " + msgID + "\n" +
			   "targetID: " + targetID + "\n" +
			   "varType: " + varType + "\n" +
			   "varName: " + varName + "\n" +
			   "startTime: " + startTime + "\n" +
			   "timeSpan: " + timeSpan + "\n" +
			   "stopTime: " + stopTime + "\n";
		return s;
	} //end stringValue()

	/* GETTERS */
	public String getClientID()
	{
		return clientID;
	}
	public String getMsgID()
	{
		return msgID;
	}
	public int getTargetID()
	{
		return targetID;
	}
	public String getVarType()
	{
		return varType;
	}
	public String getVarName()
	{
		return varName;
	}
	public int getStartTime()
	{
		return startTime;
	}
	public int getTimeSpan()
	{
		return timeSpan;
	}
	public int getStopTime()
	{
		return stopTime;
	}

	/* SETTERS */
	void setStartTime(int t)
	{
		startTime = t;
	}
	void setStopTime(int t)
	{
		timeSpan = t;
	}
	void setVarName(String s)
	{
		varName = s;
	}
	void setVarType(String s)
	{
		varType = s;
	}
	void setClientID(String i)
	{
		clientID = i;
	}

	/** Compares two MonitorNodes. Ordering is as follows:
	 * <ol><li>by {@link #startTime}
	 * <li>by {@link #stopTime}
	 * <li>by {@link #clientID}
	 * <li>by {@link #varType}
	 * </ol>
	 */
	public int compareTo(MonitorNode o)
	{

		int i = 0;

		if (startTime < o.startTime)
			return -1;
		else if (startTime > o.startTime)
			return 1;
		else
		{
			if (stopTime < o.stopTime)
				return -1;
			else if (stopTime > o.stopTime)
				return 1;
			else
			{

				i = clientID.compareTo(o.clientID);
				if (i != 0)
					return i;
				else
				{
					i = varType.compareTo(o.varType);
					if (i <= 0)
						return -1;
					else
						return 1;
				} //end clientID
			} //end stopTime
		} //end startTime
	} //compareTo

	/** Checks for equality. A MonitorNode is never equal to an Object that is not
		* another MonitorNode; two MonitorNodes are equal if all their class members are equal
	*/
	public boolean equals(Object o)
	{
		if (!(o instanceof MonitorNode))
			return false;
		MonitorNode m = (MonitorNode)o;
		return (startTime == m.startTime &&
			timeSpan == m.timeSpan &&
			varType.equals(m.varType) &&
			varName.equals(m.varName) &&
			clientID.equals(m.clientID));
	} //equals

	
	public int hashCode() {
		return startTime ^ timeSpan ^ varType.length() ^ varName.length() ^ clientID.length();
	}

} //MonitorNode
