package sassy.monitoring_steering;

/**This class provides a server socket for the monitoring and steering client to connect to. This
 *server parses the incoming message, consults the name servers to find out which LP is the
 *destination, then send the message to it.
 *@author Yin Xiong
 *@version 2.0(02/2007)
 */
import sassy.*;
import sassy.monitoring_steering.SteerNode;
import sassy.monitoring_steering.ReportNode;
import java.net.*;
import java.io.*;
import java.util.*;

public class SimSocket extends Thread
{

	/**the flag for display monitoring/steering debugging messages*/
	private static final boolean MTRACE = true;

	/** the port number this socket server will listen at; default to be 1500*/
	private int port;

	/** the server socket that receives connection from MS clients*/
	private ServerSocket server_socket;

	/** the reader that reads MS commands from the socket*/
	private BufferedReader input;

	/** the writer that writes MS report to the socket*/
	private PrintWriter output;

	/** the GNS that owns this socket server*/
	private GNS gns;

	/** the flag indicating whether the simulation is ended*/
	private volatile boolean endOfSim = false;

	/** the id assigned to an MS client*/
	private String clientID = "1";

	/** the socket client*/
	private Socket socket = null;

	/**Constructor
	 *@param g GNS instance
	 */
	public SimSocket(GNS g)
	{
		gns = g;
	}

	/**Stops this socket server
	 */
	public void stopMS()
	{
		endOfSim = true;
		try
		{
			server_socket.close();
		}
		catch (IOException ex)
		{ }
	}

	/** Writes the MS report to the socket
	 *@param rnode a ReportNode instance contaiting MS report information.
	 */
	public void sendReport(ReportNode rnode)
	{
		//send monitoring/steering results
		if (rnode != null)
		{
			String report = rnode.stringValue();
			if (MTRACE)
				System.out.println("socket server write report: " + report);
			output.println(report);
			output.flush();
		} //end if
	} //end sendReport

	/**Opens a socket and waits for the monitoring and steering client to connect with.
	 *The socket remains open until the simulation runs to the end or the client requests so.
	 */
	public void run()
	{
		endOfSim = false;
		port = 1500;

		try
		{
			server_socket = new ServerSocket(port);
			System.out.println("\nServer waiting for client on port " + server_socket.getLocalPort());
		} //end try

		catch (IOException e)
		{
			System.out.println(e);
		}

		// server infinite loop
		while (!endOfSim)
		{
			try
			{
				socket = server_socket.accept();

				if (MTRACE)
					System.out.println("\nNew connection accepted " + socket.getInetAddress() +
							   ":" + socket.getPort());
				input = new BufferedReader(new InputStreamReader(socket.getInputStream()));
				output = new PrintWriter(socket.getOutputStream(), true);

				String message = input.readLine();
				while (message != null)
				{
					//construct a MS message
					Message msg = getMsMessage(message);
					if (msg != null)
						gns.sendMSmessage(msg);
					message = input.readLine();
				} //end if input string is not null

			} //end try

			catch (IOException e)
			{
				System.out.println("Monitoring/Steering Socket Closed.");
			}

			// connection closed by client
			try
			{
				if (socket != null)
					socket.close();
				System.out.println("Connection closed by client");
			}
			catch (IOException e)
			{
				System.out.println(e);
			}
		} //end while not the end of simulation
	} //end run

	/**Processes the incoming MS command and makes it into a MonitorNode or a SteerNode.
	 *@param msg a String representation of the MS command.
	 */
	Message getMsMessage(String msg)
	{
		Message message = null;
		StringTokenizer st = new StringTokenizer(msg, "\t");
		int numTokens = st.countTokens();
		String msgID = st.nextToken();
		//System.out.println("number of tokens: " + numTokens);

		if (!msgID.equals("*"))
		{
			String msgType, varType, targetID, varName, bt, ts;
			String varValue = null;
			int beginTime, timeSpan;
			targetID = st.nextToken();
			targetID = targetID.trim();
			msgType = st.nextToken();
			if ( (msgType.equalsIgnoreCase("M") && numTokens != 7) ||
				(msgType.equalsIgnoreCase("S") && numTokens != 8) )
			{
				System.out.println("Malformated M/S commands!");
				System.exit(1);
			}

			varType = st.nextToken();
			bt = st.nextToken();
			ts = st.nextToken();
			beginTime = Integer.parseInt(bt);
			timeSpan = Integer.parseInt(ts);
			varName = st.nextToken();
			if (msgType.equalsIgnoreCase("S"))
				varValue = st.nextToken();

			RoutingInfo rinfo = null;

			try
			{
				rinfo = gns.getRoutingInfoByAid(targetID);
				while (rinfo == null)
				{
					rinfo = gns.getRoutingInfoByAid(targetID);
				} //end while
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}

			if(rinfo != null) {
				int pid = rinfo.getPEid();
				int lid = rinfo.getLPid();
				if (MTRACE)
				{
					System.out.println("get pid lid: " + + pid + " " + lid);
					System.out.println("client id: " + clientID);
					System.out.println("message ID: " + msgID);
					System.out.println("target ID: " + lid);
					System.out.println("message type: " + msgType);
					System.out.println("variabel type: " + varType);
					System.out.println("begin time: " + beginTime);
					System.out.println("time span: " + timeSpan);
					System.out.println("variable Name: " + varName);
					System.out.println("variable value: " + varValue);
				} //end if MTRACE

				if (msgType.equalsIgnoreCase("M"))
				{

					MonitorNode mNode = new MonitorNode(clientID, msgID, lid,
								    varType, beginTime, timeSpan, varName);
					message = new Message(beginTime, Message.MsgType.SYSTEM_MONITOR, mNode, null, 0);

				} //end if it's a monotoring command

				//if it's steering command
				else
				{

					SteerNode sNode = new SteerNode(clientID, msgID, lid, varType, beginTime, timeSpan,
								varName, varValue);
					message = new Message(beginTime, Message.MsgType.SYSTEM_STEER, sNode, null, 0);
				} //end if it's steering command

				message.setSenderAPPid("MS");
				message.setReceiverAPPid(targetID);
				message.setReceiverPE(pid);
				message.setReceiverLP(lid);
			}
		} // end if it's not comment

		return message;

	} //end getMessage

} //end class





