package sassy.monitoring_steering;

import java.io.*;

/** Data structure representing a single steering command.
 * <p>
 * This version only supports ongoing steering, not recurring steering. 
 * To perform one-time-only steering, set timeSpan=0.
 *
 * @version 2.0 (02/2007)
 * @author Yin xiong
 * @author Glenn Matthews
 */

public class SteerNode implements Serializable, Comparable<SteerNode>
{

	/**The flag for displaying debugging messages for steering and steering*/
	public boolean MTRACE = false;

	/** The system-assigned unique ID of the client who requested the steering*/
	public String clientID;

	/**The client-assigned id of this steering message*/
	public String msgID;

	/**The type of the variable to be monitored: V=single variable, L=all steerable for an LP*/
	public String varType;

	/** The APPid of the target LP to be steered*/
	public int targetID;

	/** The name of the variable to be steered. */
	public String varName;

	/** The value of the variable to be set at*/
	public String varValue;

	/** The simulation time at which this steering command becomes relevant.
	  * 0 = now*/
	public int startTime;

	/** The simulation time span this steering command will last;0=to the end*/
	public int timeSpan;

	/** The simulation time when the steering action stops;10000=end*/
	public int stopTime;

	/** Basic constructor. */
	public SteerNode(String cid, String mid, int tid, String vtype, int btime, int tspan,
			 String vname, String vvalue)
	{
		clientID = cid;
		msgID = mid;
		targetID = tid;
		varType = vtype;
		varName = vname;
		varValue = vvalue;
		startTime = btime;
		timeSpan = tspan;
		if (timeSpan == 0)
			stopTime = 10000;
		else
			stopTime = startTime + timeSpan;

		if (MTRACE)
		{
			System.out.println("\nin SteerNode constructor 1: " );
			System.out.println("client id" + clientID);
			System.out.println("message id" + msgID);
			System.out.println("target id: " + targetID);
			System.out.println("variable type " + varType);
			System.out.println("start time: " + startTime);
			System.out.println("time span: " + timeSpan);
			System.out.println("variable name: " + varName);
			System.out.println("variable value: " + varValue);
		} //end debugging

	} //end basic constructor

	/**Returns the class members and their values as a string*/
	String stringValue()
	{
		String s = "clientID: " + clientID + "\n" +
			   "msgID: " + msgID + "\n" +
			   "targetID: " + targetID + "\n" +
			   "varType: " + varType + "\n" +
			   "startTime: " + startTime + "\n" +
			   "timeSpan: " + timeSpan + "\n" +
			   "stopTime: " + stopTime + "\n" +
			   "varName: " + varName + "\n" +
			   "varValue: " + varValue + "\n";
		return s;
	} //end stringValue()

	/* GETTERS */
	public String getClientID()
	{
		return clientID;
	}
	public String getMsgID()
	{
		return msgID;
	}
	public int getTargetID()
	{
		return targetID;
	}
	public String getVarType()
	{
		return varType;
	}
	public String getVarName()
	{
		return varName;
	}
	public int getStartTime()
	{
		return startTime;
	}
	public int getTimeSpan()
	{
		return timeSpan;
	}
	public int getStopTime()
	{
		return stopTime;
	}

	/* SETTERS */
	void setStartTime(int t)
	{
		startTime = t;
	}
	void setStopTime(int t)
	{
		timeSpan = t;
	}
	void setVarName(String s)
	{
		varName = s;
	}
	void setVarType(String s)
	{
		varType = s;
	}
	void setClientID(String i)
	{
		clientID = i;
	}

	/** Compares two SteerNodes. Ordering is as follows:
	 * <ol><li>by {@link #startTime}
	 * <li>by {@link #stopTime}
	 * <li>by {@link #clientID}
	 * <li>by {@link #varType}
	 * </ol>
	 */
	public int compareTo(SteerNode o)
	{

		int i = 0;

		if (startTime < o.startTime)
			return -1;
		else if (startTime > o.startTime)
			return 1;
		else
		{
			if (stopTime < o.stopTime)
				return -1;
			else if (stopTime > o.stopTime)
				return 1;
			else
			{

				i = clientID.compareTo(o.clientID);
				if (i != 0)
					return i;
				else
				{
					i = varType.compareTo(o.varType);
					if (i <= 0)
						return -1;
					else
						return 1;
				} //end clientID
			} //end stopTime
		} //end startTime
	} //compareTo

	/** Checks for equality. A SteerNode is never equal to an Object that is not
		* another SteerNode; two SteerNodes are equal if all their class members are equal
	*/
	public boolean equals(Object o)
	{
		if (!(o instanceof SteerNode))
			return false;
		SteerNode m = (SteerNode)o;
		return (startTime == m.startTime &&
			timeSpan == m.timeSpan &&
			varType.equals(m.varType) &&
			varName.equals(m.varName) &&
			clientID.equals(m.clientID));
	} //equals

	public int hashCode() {
		return startTime ^ timeSpan ^ varType.length() ^ varName.length() ^ clientID.length();
	}

} //SteerNode
